/* ---------------------- **
**      Atomic locks      **
** ---------------------- */

#ifdef i386
#include "x86_locks.h"
#endif /* i386 */

#ifdef sparc
#include "sparc_locks.h"
#endif /* sparc */



/* ---------------------------- **
**      Locking operations      **
** ---------------------------- */

#ifdef YAPOR
#define INIT_LOCK(LOCK_VAR)    ((LOCK_VAR) = 0)
#define LOCK(LOCK_VAR)         do {                                     \
                                 if (TRY_LOCK(&(LOCK_VAR))) break;      \
			         SIMICS_SERVICE(START_COUNTER, 11);     \
		                 while (IS_LOCKED(LOCK_VAR)) continue;  \
			         SIMICS_SERVICE(STOP_COUNTER, 11);      \
		               } while (1)
#define UNLOCK(LOCK_VAR)       ((LOCK_VAR) = 0)
#define IS_LOCKED(LOCK_VAR)    ((LOCK_VAR) != 0)
#define IS_UNLOCKED(LOCK_VAR)  ((LOCK_VAR) == 0)
#else
#define INIT_LOCK(LOCK_VAR)
#define LOCK(LOCK_VAR)
#define UNLOCK(LOCK_VAR)
#define IS_LOCKED(LOCK_VAR)
#define IS_UNLOCKED(LOCK_VAR)
#endif /* YAPOR */



/* ------------------------------------- **
**      Alloc and free data structs      **
** ------------------------------------- */

#ifdef YAPOR_STATISTICS
#define UPDATE_OR_STATS(STAT, VALUE)  \
        LOCK(GLOBAL_STATS_lock);      \
        STAT += VALUE;                \
        UNLOCK(GLOBAL_STATS_lock)
#else
#define UPDATE_OR_STATS(STAT, VALUE)
#endif /* YAPOR_STATISTICS */ 

#ifdef TABLING_STATISTICS
#define UPDATE_TAB_STATS(STAT, VALUE)  \
        LOCK(GLOBAL_STATS_lock);       \
        STAT += VALUE;                 \
        UNLOCK(GLOBAL_STATS_lock)
#else
#define UPDATE_TAB_STATS(STAT, VALUE)
#endif /* TABLING_STATISTICS */ 

#ifdef OPTYAP_STATISTICS
#define UPDATE_OPT_STATS(STAT, VALUE)  \
        LOCK(GLOBAL_STATS_lock);       \
        STAT += VALUE;                 \
        UNLOCK(GLOBAL_STATS_lock)
#else
#define UPDATE_OPT_STATS(STAT, VALUE)
#endif /* OPTYAP_STATISTICS */ 

#ifdef YAPOR
#define ALLOC_OR_FRAME(fr)                                           \
        LOCK(GLOBAL_LOCKS_next_free_or_fr);                          \
        if (GLOBAL_next_free_or_fr == NULL) alloc_or_frames();       \
        fr = GLOBAL_next_free_or_fr;                                 \
        GLOBAL_next_free_or_fr = OrFr_next(GLOBAL_next_free_or_fr);  \
        UNLOCK(GLOBAL_LOCKS_next_free_or_fr);                        \
        UPDATE_OR_STATS(GLOBAL_STATS_used_or_fr, 1)

#define FREE_OR_FRAME(fr)                             \
        LOCK(GLOBAL_LOCKS_next_free_or_fr);           \
        OrFr_next(fr) = GLOBAL_next_free_or_fr;       \
	GLOBAL_next_free_or_fr = fr;                  \
	UNLOCK(GLOBAL_LOCKS_next_free_or_fr);         \
        UPDATE_OR_STATS(GLOBAL_STATS_used_or_fr, -1)

#define ALLOC_ANSWER_FRAME(fr)                                          \
        LOCK(GLOBAL_LOCKS_next_free_ans_fr);                            \
        if (GLOBAL_next_free_ans_fr == NULL) alloc_answer_frames();     \
        fr = GLOBAL_next_free_ans_fr;                                   \
        GLOBAL_next_free_ans_fr = AnsFr_next(GLOBAL_next_free_ans_fr);  \
        UNLOCK(GLOBAL_LOCKS_next_free_ans_fr);                          \
        UPDATE_OR_STATS(GLOBAL_STATS_used_ans_fr, 1)

#define FREE_ANSWER_FRAME(fr)                          \
        LOCK(GLOBAL_LOCKS_next_free_ans_fr);           \
        AnsFr_next(fr) = GLOBAL_next_free_ans_fr;      \
	GLOBAL_next_free_ans_fr = fr;                  \
	UNLOCK(GLOBAL_LOCKS_next_free_ans_fr);         \
        UPDATE_OR_STATS(GLOBAL_STATS_used_ans_fr, -1)

#define ALLOC_SOLUTION_FRAME(fr)                                        \
        LOCK(GLOBAL_LOCKS_next_free_sol_fr);                            \
        if (GLOBAL_next_free_sol_fr == NULL) alloc_solution_frames();   \
        fr = GLOBAL_next_free_sol_fr;                                   \
        GLOBAL_next_free_sol_fr = SolFr_next(GLOBAL_next_free_sol_fr);  \
        UNLOCK(GLOBAL_LOCKS_next_free_sol_fr);                          \
        UPDATE_OR_STATS(GLOBAL_STATS_used_sol_fr, 1)

#define FREE_SOLUTION_FRAME(fr)                        \
	LOCK(GLOBAL_LOCKS_next_free_sol_fr);           \
	SolFr_next(fr) = GLOBAL_next_free_sol_fr;      \
        GLOBAL_next_free_sol_fr = fr;                  \
        UNLOCK(GLOBAL_LOCKS_next_free_sol_fr);         \
        UPDATE_OR_STATS(GLOBAL_STATS_used_sol_fr, -1)
#endif /* YAPOR */

#ifdef TABLING
#define ALLOC_TABLE_ENTRY(te)                                              \
        LOCK(GLOBAL_LOCKS_next_free_tab_ent);                              \
        if (GLOBAL_next_free_tab_ent == NULL) alloc_table_entries();       \
        te = GLOBAL_next_free_tab_ent;                                     \
        GLOBAL_next_free_tab_ent = TabEnt_next(GLOBAL_next_free_tab_ent);  \
        UNLOCK(GLOBAL_LOCKS_next_free_tab_ent);                            \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_tab_ent, 1)

#define FREE_TABLE_ENTRY(te)                             \
	LOCK(GLOBAL_LOCKS_next_free_tab_ent);            \
	TabEnt_next(te) = GLOBAL_next_free_tab_ent;      \
        GLOBAL_next_free_tab_ent = te;                   \
        UNLOCK(GLOBAL_LOCKS_next_free_tab_ent);          \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_tab_ent, -1)

#define ALLOC_SUBGOAL_TRIE_NODE(tr)                                        \
        LOCK(GLOBAL_LOCKS_next_free_sg_node);                              \
        if (GLOBAL_next_free_sg_node == NULL) alloc_subgoal_trie_nodes();  \
        tr = GLOBAL_next_free_sg_node;                                     \
        GLOBAL_next_free_sg_node = TrNode_next(GLOBAL_next_free_sg_node);  \
        UNLOCK(GLOBAL_LOCKS_next_free_sg_node);                            \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_sg_node, 1)

#define FREE_SUBGOAL_TRIE_NODE(tr)                       \
	LOCK(GLOBAL_LOCKS_next_free_sg_node);            \
	TrNode_next(tr) = GLOBAL_next_free_sg_node;      \
        GLOBAL_next_free_sg_node = tr;                   \
        UNLOCK(GLOBAL_LOCKS_next_free_sg_node);          \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_sg_node, -1)

#define ALLOC_SUBGOAL_FRAME(fr)                                      \
        LOCK(GLOBAL_LOCKS_next_free_sg_fr);                          \
        if (GLOBAL_next_free_sg_fr == NULL) alloc_subgoal_frames();  \
        fr = GLOBAL_next_free_sg_fr;                                 \
        GLOBAL_next_free_sg_fr = SgFr_next(GLOBAL_next_free_sg_fr);  \
        UNLOCK(GLOBAL_LOCKS_next_free_sg_fr);                        \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_sg_fr, 1)

#define FREE_SUBGOAL_FRAME(fr)                         \
	LOCK(GLOBAL_LOCKS_next_free_sg_fr);            \
	SgFr_next(fr) = GLOBAL_next_free_sg_fr;        \
        GLOBAL_next_free_sg_fr = fr;                   \
        UNLOCK(GLOBAL_LOCKS_next_free_sg_fr);          \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_sg_fr, -1)

#define ALLOC_ANSWER_TRIE_NODE(tr)                                           \
        LOCK(GLOBAL_LOCKS_next_free_ans_node);                               \
        if (GLOBAL_next_free_ans_node == NULL) alloc_answer_trie_nodes();    \
        tr = GLOBAL_next_free_ans_node;                                      \
        GLOBAL_next_free_ans_node = TrNode_next(GLOBAL_next_free_ans_node);  \
        UNLOCK(GLOBAL_LOCKS_next_free_ans_node);                             \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_ans_node, 1)

#define FREE_ANSWER_TRIE_NODE(tr)                         \
	LOCK(GLOBAL_LOCKS_next_free_ans_node);            \
	TrNode_next(tr) = GLOBAL_next_free_ans_node;      \
        GLOBAL_next_free_ans_node = tr;                   \
        UNLOCK(GLOBAL_LOCKS_next_free_ans_node);          \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_ans_node, -1)

#define ALLOC_DEPENDENCY_FRAME(fr)                                       \
        LOCK(GLOBAL_LOCKS_next_free_dep_fr);                             \
        if (GLOBAL_next_free_dep_fr == NULL) alloc_dependency_frames();  \
        fr = GLOBAL_next_free_dep_fr;                                    \
        GLOBAL_next_free_dep_fr = DepFr_next(GLOBAL_next_free_dep_fr);   \
        UNLOCK(GLOBAL_LOCKS_next_free_dep_fr);                           \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_dep_fr, 1)

#define FREE_DEPENDENCY_FRAME(fr)                        \
	LOCK(GLOBAL_LOCKS_next_free_dep_fr);             \
	DepFr_next(fr) = GLOBAL_next_free_dep_fr;        \
        GLOBAL_next_free_dep_fr = fr;                    \
        UNLOCK(GLOBAL_LOCKS_next_free_dep_fr);           \
        UPDATE_TAB_STATS(GLOBAL_STATS_used_dep_fr, -1)

#ifdef YAPOR
#define ALLOC_SUSPENSION_FRAME(fr)                                         \
        LOCK(GLOBAL_LOCKS_next_free_susp_fr);                              \
        if (GLOBAL_next_free_susp_fr == NULL) alloc_suspension_frames();   \
        fr = GLOBAL_next_free_susp_fr;                                     \
        GLOBAL_next_free_susp_fr = SuspFr_next(GLOBAL_next_free_susp_fr);  \
        UNLOCK(GLOBAL_LOCKS_next_free_susp_fr);                            \
        UPDATE_OPT_STATS(GLOBAL_STATS_used_susp_fr, 1)

#define FREE_SUSPENSION_FRAME(fr)                        \
	LOCK(GLOBAL_LOCKS_next_free_susp_fr);            \
	SuspFr_next(fr) = GLOBAL_next_free_susp_fr;      \
        GLOBAL_next_free_susp_fr = fr;                   \
        UNLOCK(GLOBAL_LOCKS_next_free_susp_fr);          \
        UPDATE_OPT_STATS(GLOBAL_STATS_used_susp_fr, -1)
#endif /* YAPOR */
#endif /* TABLING */



/* ------------------------------------- **
**      Bitmap tests and operations      **
** ------------------------------------- */

#define BITMAP_empty(b)		       ((b) == 0)
#define BITMAP_member(b,n)	       (((b) & (1<<(n))) != 0)
#define BITMAP_alone(b,n)	       ((b) == (1<<(n)))
#define BITMAP_subset(b1,b2)	       (((b1) & (b2)) == b2)
#define BITMAP_same(b1,b2)             ((b1) == (b2))

#define BITMAP_clear(b)	               ((b) = 0)
#define BITMAP_and(b1,b2)              ((b1) &= (b2))
#define BITMAP_minus(b1,b2)            ((b1) &= ~(b2))
#define BITMAP_insert(b,n)	       ((b) |= (1<<(n)))
#define BITMAP_delete(b,n)	       ((b) &= (~(1<<(n))))
#define BITMAP_copy(b1,b2)	       ((b1) = (b2))
#define BITMAP_intersection(b1,b2,b3)  ((b1) = ((b2) & (b3)))
#define BITMAP_difference(b1,b2,b3)    ((b1) = ((b2) & (~(b3))))



/* ----------------------------- **
**      Miscelaneous macros      **
** ----------------------------- */

#define H_BASE   ((CELL *) GlobalBase)
#define B_BASE   ((choiceptr) LocalBase)
#define TR_BASE  ((tr_fr_ptr) TrailBase)

#ifdef YAPOR
#define SEQUENTIAL_MODE   0
#define PARALLEL_MODE     1
#define EXECUTING_MODE    0
#define SUSPENDED_MODE    1
#define PERFORMANCE_OFF           0x0
#define PERFORMANCE_ON            0x1
#define PERFORMANCE_IN_EXECUTION  0x2
#define YAMOP_LTT  u.ld.ltt
#define YAMOP_CUT  u.ld.cut
#define YAMOP_SEQ  u.ld.seq
#endif /* YAPOR */



/* ---------------------------------- **
**      Message and debug macros      **
** ---------------------------------- */

#define INFORMATION_MESSAGE(MESG, ARGS...)  information_message(MESG, ##ARGS)

#ifdef YAPOR_ERRORS
#define YAPOR_ERROR_MESSAGE(MESG, ARGS...)  error_message(MESG, ##ARGS)
#else
#define YAPOR_ERROR_MESSAGE(MESG, ARGS...)
#endif /* YAPOR_ERRORS */

#ifdef TABLING_ERRORS
#define TABLING_ERROR_MESSAGE(MESG, ARGS...)  error_message(MESG, ##ARGS)
#else
#define TABLING_ERROR_MESSAGE(MESG, ARGS...)
#endif /* TABLING_ERRORS */

#ifdef OPTYAP_ERRORS
#define OPTYAP_ERROR_MESSAGE(MESG, ARGS...)  error_message(MESG, ##ARGS)
#else
#define OPTYAP_ERROR_MESSAGE(MESG, ARGS...)
#endif /* OPTYAP_ERRORS */

#ifdef YAPOR_DEBUG
#define YAPOR_DEBUG_MESSAGE(MESG, ARGS...)  debug_message(MESG, ##ARGS)
#else
#define YAPOR_DEBUG_MESSAGE(MESG, ARGS...)
#endif /* YAPOR_DEBUG */

#ifdef TABLING_DEBUG
#define TABLING_DEBUG_MESSAGE(MESG, ARGS...)  debug_message(MESG, ##ARGS)
#define TABLING_INSTR_DEBUG_MESSAGE(MESG, PRED_ENTRY, ARGS) tabling_instr_debug_message(MESG, PRED_ENTRY, ARGS)
#else
#define TABLING_DEBUG_MESSAGE(MESG, ARGS...)
#define TABLING_INSTR_DEBUG_MESSAGE(MESG, PRED_ENTRY, ARGS)
#endif /* TABLING_DEBUG */

#ifdef OPTYAP_DEBUG
#define OPTYAP_DEBUG_MESSAGE(MESG, ARGS...)  debug_message(MESG, ##ARGS)
#else
#define OPTYAP_DEBUG_MESSAGE(MESG, ARGS...)
#endif /* OPTYAP_DEBUG */



/* ----------------------- **
**      SimICS macros      **
** ----------------------- */

/*
**  Counter  0. Total time
**  Counter  1. Prolog
**  Counter  2. P Share
**  Counter  3. Scheduler
**  Counter  4. Cut request
**  Counter  5. End operations
**  Counter  6. Cut shared (Counter 1 or 3 or 5)
**  Counter  7. Number of requests to share work (Counter 3)
**  Counter  8. Number of refused requests (Counter 3)
**  Counter  9. Number of tasks (Counter 1)
**  Counter 10. Number of calls (Counter 1)
**  Counter 11. Number of failed TRY_LOCK's
*/

#define START_COUNTER            1
#define STOP_COUNTER             2

#define TOTAL_TIME               0
#define PROLOG                   1
#define SHARE                    2
#define SCHEDULER                3
#define CUT_REQUEST              4
#define END_OPERATIONS           5
#define CUT_SHARED               6
#define ONE_MORE_REQUEST         7
#define ONE_MORE_REFUSED_REQUEST 8
#define ONE_MORE_TASK            9
#define ONE_MORE_CALL            10

#ifdef SIMICS
#define SIMICS_ATOMIC_SERVICE(COUNTER)           \
        SIMICS_SERVICE(START_COUNTER, COUNTER);  \
        SIMICS_SERVICE(STOP_COUNTER, COUNTER)
#define SIMICS_SERVICE(COMMAND, COUNTER)         \
        do {                                     \
          asm volatile ("sethi %0, %%g0" :       \
          /* no outputs */ :                     \
          "g" ((COMMAND << 16) | COUNTER));      \
	} while(0)
#else
#define SIMICS_ATOMIC_SERVICE(COUNTER)
#define SIMICS_SERVICE(COMMAND, COUNTER)
#endif /* SIMICS */
