/* --------
   Includes
   -------- */

#include "Yap.h"
#ifdef ACOW
#include "Yatom.h"
#include "Heap.h"
#include "or.macros.h"
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>



/* ---------------------------
   Local functions declaration
   --------------------------- */

static void share_private_nodes(int worker_q);


/* ----------------
   Global functions
   ---------------- */

void make_root_choice_point(void) {
  if (worker_id == 0) {
    LOCAL_top_cp = GLOBAL_root_cp = OrFr_node(GLOBAL_root_or_fr) = B;
  } else {
    B = LOCAL_top_cp = GLOBAL_root_cp;
    B->cp_tr = TR = ((choiceptr) (worker_offset(0) + (CELL)(B)))->cp_tr;
  }
  B->cp_h = H0;
  B->cp_ap = GETWORK;
  B->cp_or_fr = GLOBAL_root_or_fr;
  B->cp_lub = 0;
  LOCAL_top_or_fr = GLOBAL_root_or_fr;
  LOCAL_load = -DELAYED_RELEASE_LOAD;
  LOCAL_cut_cp = NULL;
  GLOBAL_branch_ltt(worker_id, 0) = 0;
  GLOBAL_branch_cut(worker_id, 0) = 0;
  return;
}


void free_root_choice_point(void) {
  B = LOCAL_top_cp->cp_b;
  LOCAL_top_cp = B_BASE;
  return;
}


static inline
void PUT_ME_BUSY(int worker_num) {
  LOCK(GLOBAL_LOCKS_bm_idle_workers);
  BITMAP_delete(GLOBAL_bm_idle_workers, worker_num);
  UNLOCK(GLOBAL_LOCKS_bm_idle_workers);
  return;
}

int p_share_work(void) {
  int worker_q = LOCAL_request;
  int son;

  if (! BITMAP_member(OrFr_members(REMOTE_top_or_fr(worker_q)), worker_id) ||
      B == REMOTE_top_cp(worker_q) ||
      (LOCAL_load <= 0 && OrFr_nearest_livenode(LOCAL_top_or_fr) == NULL)) {
    /* refuse sharing request */
    REMOTE_reply_signal(worker_q) = refused;
    LOCAL_request = MAX_WORKERS;
    SCH_put_me_out_of_access_work(worker_id);
    return 1;
  }
  /* sharing request accepted */
  LOCAL_reply_signal = no_reply;
  REMOTE_reply_signal(worker_q) = accepted;
  share_private_nodes(worker_q);
  if ((son = fork()) == 0) {
    int worker_p;
    worker_p  = worker_id;   /* parent is giving worker */
    worker_id = worker_q;    /* child becomes requesting worker */
    LOCAL = REMOTE[worker_id];

    LOCAL_reply_signal = ready;
    LOCK_WORKER(worker_p);
    REMOTE_request(worker_p) = MAX_WORKERS;
    REMOTE_reply_signal(worker_p) = ready;
    UNLOCK_WORKER(worker_p);
    SCH_put_us_access_work(worker_id, worker_p);

    /* update local variables */
    LOCAL_top_cp = B;
    LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
    LOCAL_load = -DELAYED_RELEASE_LOAD;
    PUT_ME_BUSY(worker_id);

    return 0;
  }
  worker_pid(worker_q) = son;
  while (LOCAL_reply_signal == no_reply) continue;
  return 1;
}


int q_share_work(int worker_p) {
  LOCK_OR_FRAME(LOCAL_top_or_fr);
  /* looking for cut messages */
  if (LOCAL_cut_cp != NULL) {
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    CUT_move_up_to_cut_cp();
    CUT_clear_cut_cp();
    return 0;
  }
  if (! BITMAP_member(OrFr_members(LOCAL_top_or_fr), worker_p)) {
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    return 0;
  }
  GLOBAL_branch_cut(worker_id, OrFr_depth(LOCAL_top_or_fr)) = 
    GLOBAL_branch_cut(worker_p, OrFr_depth(LOCAL_top_or_fr));
  GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr)) = 
    GLOBAL_branch_ltt(worker_p, OrFr_depth(LOCAL_top_or_fr));

  /* restart pending cuts */
  if (OrFr_pend_cut_cp(LOCAL_top_or_fr) != NULL && no_workers_left_pending_cut()) {
    BITMAP_delete(OrFr_members(LOCAL_top_or_fr), worker_id);
    CUT_continue_pending_cut();
    return 0;
  }
  UNLOCK_OR_FRAME(LOCAL_top_or_fr);

  /* make sharing request */
  LOCK_WORKER(worker_p);
  if (BITMAP_member(GLOBAL_bm_idle_workers, worker_p) || REMOTE_request(worker_p) != MAX_WORKERS) {
    /* worker p is idle or has another request */
    UNLOCK_WORKER(worker_p);
    return 0;
  }
  LOCAL_reply_signal = no_reply;
  REMOTE_request(worker_p) = worker_id;
  UNLOCK_WORKER(worker_p);
  SIMICS_ATOMIC_SERVICE(ONE_MORE_REQUEST);

  /* wait for an answer */
  while (LOCAL_reply_signal == no_reply) continue;
  if (LOCAL_reply_signal == refused) {
    /* sharing request refused */
    SIMICS_ATOMIC_SERVICE(ONE_MORE_REFUSED_REQUEST);
    LOCAL_reply_signal = no_reply;
    return 0;
  }
  /* exit this process */
  exit(0);
}

/* ---------------
   Local functions
   --------------- */


static void share_private_nodes(int worker_q) {
  choiceptr AuxB;
  or_fr_ptr or_frame;

  AuxB = B;
  {
    int i, depth, leftmost;
    or_fr_ptr previous_or_frame;
    /* initialize auxiliary variables */
    previous_or_frame = NULL;
    depth = OrFr_depth(LOCAL_top_or_fr);
    /* searching for leftmostness */
    leftmost = 1;
    for (i = 0; i < number_workers; i++) {
      if (i != worker_id && 
          BITMAP_member(OrFr_members(LOCAL_top_or_fr), i) &&
          GLOBAL_branch_ltt(i, depth) > GLOBAL_branch_ltt(worker_id, depth)) {
        leftmost = 0;
        break;
      }
    }
    /* sharing loop */
    while (AuxB != LOCAL_top_cp) {
      depth++;
      ALLOC_OR_FRAME(or_frame);
      INIT_LOCK(OrFr_lock(or_frame));
      OrFr_node(or_frame) = AuxB;
      OrFr_alternative(or_frame) = AuxB->cp_ap;
      OrFr_pend_cut_cp(or_frame) = NULL;
      OrFr_solutions(or_frame) = NULL;
      BITMAP_clear(OrFr_members(or_frame));
      BITMAP_insert(OrFr_members(or_frame), worker_id);
      BITMAP_insert(OrFr_members(or_frame), worker_q);
      if (leftmost)
        OrFr_nearest_leftnode(or_frame) = OrFr_nearest_leftnode(LOCAL_top_or_fr);
      else
        OrFr_nearest_leftnode(or_frame) = LOCAL_top_or_fr;
      if (AuxB->cp_ap && AuxB->cp_ap->YAMOP_SEQ) {
        AuxB->cp_ap = GETWORK_SEQ;
      } else {
        AuxB->cp_ap = GETWORK;
      }
      AuxB->cp_or_fr = or_frame;
      AuxB->cp_lub = 0;
      AuxB = AuxB->cp_b;
      if (previous_or_frame) {
        OrFr_nearest_livenode(previous_or_frame) = OrFr_next(previous_or_frame) = or_frame;
      }
      previous_or_frame = or_frame;
    }
    /* initialize last or-frame pointer */
    or_frame = AuxB->cp_or_fr;
    if (previous_or_frame) {
      OrFr_nearest_livenode(previous_or_frame) = OrFr_next(previous_or_frame) = or_frame;
    }
    /* update depth */
    if (depth >= MAX_DEPTH)
      abort_optyap("maximum depth exceded (%d/%d) (share_private_nodes)", MAX_DEPTH, depth - MAX_DEPTH);
    or_frame = B->cp_or_fr;
    while (or_frame != LOCAL_top_or_fr) {
      yamop *alt;
      if ((alt = OrFr_alternative(or_frame)) != NULL) {
        GLOBAL_branch_ltt(worker_id, depth) = GLOBAL_branch_ltt(worker_q, depth) = alt->YAMOP_LTT + 1;
        GLOBAL_branch_cut(worker_id, depth) = GLOBAL_branch_cut(worker_q, depth) = alt->YAMOP_CUT;
      } else {
        GLOBAL_branch_ltt(worker_id, depth) = GLOBAL_branch_ltt(worker_q, depth) = 1;
        GLOBAL_branch_cut(worker_id, depth) = GLOBAL_branch_cut(worker_q, depth) = 0;
      }
      OrFr_depth(or_frame) = depth--;
      or_frame = OrFr_next_on_stack(or_frame);
    }
    /* update old shared nodes */
    while (or_frame != REMOTE_top_or_fr(worker_q)) {
      LOCK_OR_FRAME(or_frame);
      BITMAP_insert(OrFr_members(or_frame), worker_q);
      GLOBAL_branch_cut(worker_q, OrFr_depth(or_frame)) = GLOBAL_branch_cut(worker_id, OrFr_depth(or_frame));
      GLOBAL_branch_ltt(worker_q, OrFr_depth(or_frame)) = GLOBAL_branch_ltt(worker_id, OrFr_depth(or_frame));
      UNLOCK_OR_FRAME(or_frame);
      or_frame = OrFr_next_on_stack(or_frame);
    }
    /* looking for cut messages */
    if (LOCAL_cut_cp != NULL) { 
      LOCK_WORKER(worker_id);
      CUT_make_cut_request(worker_q, LOCAL_cut_cp);
      UNLOCK_WORKER(worker_id);
    }
    /* update top shared node */
    LOCAL_top_cp = B;
    LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
    LOCAL_load = -DELAYED_RELEASE_LOAD;
  }
  return;
}
#endif /* ACOW */
