/* ------------------ **
**      Includes      **
** ------------------ */

#include "Yap.h"
#ifdef ENV_COPY
#include "Heap.h"
#include "or.macros.h"
#ifdef TABLING
#include "tab.macros.h"
#endif /* TABLING */



/* ---------------------- **
**      Local macros      **
** ---------------------- */

#define COMPUTE_SEGMENTS_TO_COPY_TO(Q)                                   \
        REMOTE_start_global_copy(Q) = (CELL) (REMOTE_top_cp(Q)->cp_h);   \
        REMOTE_end_global_copy(Q)   = (CELL) (B->cp_h);                  \
        REMOTE_start_local_copy(Q)  = (CELL) (B);                        \
        REMOTE_end_local_copy(Q)    = (CELL) (REMOTE_top_cp(Q));         \
        REMOTE_start_trail_copy(Q)  = (CELL) (REMOTE_top_cp(Q)->cp_tr);  \
        REMOTE_end_trail_copy(Q)    = (CELL) (TR)

#define P_COPY_GLOBAL_TO(Q)                                                         \
        memcpy((void *) (worker_offset(Q) + REMOTE_start_global_copy(Q)),           \
               (void *) REMOTE_start_global_copy(Q),                                \
               (size_t) (REMOTE_end_global_copy(Q) - REMOTE_start_global_copy(Q)))
#define Q_COPY_GLOBAL_FROM(P)                                                       \
        memcpy((void *) LOCAL_start_global_copy,                                    \
               (void *) (worker_offset(P) + LOCAL_start_global_copy),               \
               (size_t) (LOCAL_end_global_copy - LOCAL_start_global_copy))

#define P_COPY_LOCAL_TO(Q)                                                          \
        memcpy((void *) (worker_offset(Q) + REMOTE_start_local_copy(Q)),            \
               (void *) REMOTE_start_local_copy(Q),                                 \
               (size_t) (REMOTE_end_local_copy(Q) - REMOTE_start_local_copy(Q)))
#define Q_COPY_LOCAL_FROM(P)                                                        \
        memcpy((void *) LOCAL_start_local_copy,                                     \
               (void *) (worker_offset(P) + LOCAL_start_local_copy),                \
               (size_t) (LOCAL_end_local_copy - LOCAL_start_local_copy))

#define P_COPY_TRAIL_TO(Q)                                                          \
        memcpy((void *) (worker_offset(Q) + REMOTE_start_trail_copy(Q)),            \
               (void *) REMOTE_start_trail_copy(Q),                                 \
               (size_t) (REMOTE_end_trail_copy(Q) - REMOTE_start_trail_copy(Q)))
#define Q_COPY_TRAIL_FROM(P)                                                        \
        memcpy((void *) LOCAL_start_trail_copy,                                     \
               (void *) (worker_offset(P) + LOCAL_start_trail_copy),                \
               (size_t) (LOCAL_end_trail_copy - LOCAL_start_trail_copy))



/* ------------------------------------- **
**      Local functions declaration      **
** ------------------------------------- */

static void share_private_nodes(int worker_q);



/* -------------------------- **
**      Global functions      **
** -------------------------- */

void make_root_choice_point(void) {
  if (worker_id == 0) {
    LOCAL_top_cp = GLOBAL_root_cp = OrFr_node(GLOBAL_root_or_fr) = B;
  } else {
    B = LOCAL_top_cp = GLOBAL_root_cp;
    B->cp_tr = TR = ((choiceptr) (worker_offset(0) + (CELL)(B)))->cp_tr;
  }
  B->cp_h = H0;
  B->cp_ap = GETWORK;
  B->cp_or_fr = GLOBAL_root_or_fr;
  B->cp_lub = 0;
  LOCAL_top_or_fr = GLOBAL_root_or_fr;
  LOCAL_load = -DELAYED_RELEASE_LOAD;
  LOCAL_cut_cp = NULL;
  GLOBAL_branch_ltt(worker_id, 0) = 0;
  GLOBAL_branch_cut(worker_id, 0) = 0;
#ifdef TABLING
  /* maybe not necessary */
  LOCAL_top_cp_on_stack = LOCAL_top_cp;
  LOCAL_top_collected_susp_fr = NULL;
  LOCAL_top_dep_fr = GLOBAL_root_dep_fr; 
  LOCAL_top_sg_fr = NULL; 
  H_FZ = H_BASE;
  B_FZ = B_BASE;
  TR_FZ = TR_BASE;
#endif /* TABLING */
  return;
}


void free_root_choice_point(void) {
  B = LOCAL_top_cp->cp_b;
  LOCAL_top_cp = B_BASE;
#ifdef TABLING
  LOCAL_top_cp_on_stack = B_BASE;
#endif /* TABLING */
  return;
}


void p_share_work(void) {
  int worker_q = LOCAL_request;

  if (! BITMAP_member(OrFr_members(REMOTE_top_or_fr(worker_q)), worker_id) ||
      B == REMOTE_top_cp(worker_q) ||
      (LOCAL_load <= 0 && OrFr_nearest_livenode(LOCAL_top_or_fr) == NULL)) {
    /* refuse sharing request */
    REMOTE_reply_signal(worker_q) = refused;
    LOCAL_request = MAX_WORKERS;
    SCH_put_me_out_of_access_work(worker_id);
    return;
  }
  /* sharing request accepted */
  COMPUTE_SEGMENTS_TO_COPY_TO(worker_q);
  LOCAL_reply_signal = no_reply;
  REMOTE_q_fase_signal(worker_q) = Q_idle;
  REMOTE_p_fase_signal(worker_q) = P_idle;
  REMOTE_reply_signal(worker_q) = accepted;
  share_private_nodes(worker_q);
  REMOTE_reply_signal(worker_q) = local_ok;
  /* copy local stack ? */
  LOCK(REMOTE_lock_signals(worker_q));
  if (REMOTE_q_fase_signal(worker_q) < local) {
    REMOTE_p_fase_signal(worker_q) = local;              
    UNLOCK(REMOTE_lock_signals(worker_q));
    P_COPY_LOCAL_TO(worker_q);
  } else {
    UNLOCK(REMOTE_lock_signals(worker_q));
    goto wait_install;
  }
  /* copy global stack ? */
  LOCK(REMOTE_lock_signals(worker_q));
  if (REMOTE_q_fase_signal(worker_q) < global) {
    REMOTE_p_fase_signal(worker_q) = global;
    UNLOCK(REMOTE_lock_signals(worker_q));
    P_COPY_GLOBAL_TO(worker_q);
  } else {
    UNLOCK(REMOTE_lock_signals(worker_q));
    goto wait_install;
  }
  /* copy trail stack ? */
  LOCK(REMOTE_lock_signals(worker_q));
  if (REMOTE_q_fase_signal(worker_q) < trail) {
    REMOTE_p_fase_signal(worker_q) = trail;
    UNLOCK(REMOTE_lock_signals(worker_q));
    P_COPY_TRAIL_TO(worker_q);
  } else UNLOCK(REMOTE_lock_signals(worker_q));

wait_install:
  REMOTE_reply_signal(worker_q) = ready;
  while (LOCAL_reply_signal < install) continue;
  return;
}


int q_share_work(int worker_p) {
  register tr_fr_ptr aux_tr;
  register CELL aux_cell;

  LOCK_OR_FRAME(LOCAL_top_or_fr);
  /* looking for cut messages */
  if (LOCAL_cut_cp != NULL) {
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    CUT_move_up_to_cut_cp();
    CUT_clear_cut_cp();
    return FALSE;
  }
  if (! BITMAP_member(OrFr_members(LOCAL_top_or_fr), worker_p)) {
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    return FALSE;
  }
  GLOBAL_branch_cut(worker_id, OrFr_depth(LOCAL_top_or_fr)) = 
    GLOBAL_branch_cut(worker_p, OrFr_depth(LOCAL_top_or_fr));
  GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr)) = 
    GLOBAL_branch_ltt(worker_p, OrFr_depth(LOCAL_top_or_fr));

  /* restart pending cuts */
  if (OrFr_pend_cut_cp(LOCAL_top_or_fr) != NULL && no_workers_left_pending_cut()) {
    BITMAP_delete(OrFr_members(LOCAL_top_or_fr), worker_id);
    CUT_continue_pending_cut();
    return FALSE;
  }
  UNLOCK_OR_FRAME(LOCAL_top_or_fr);

  /* unbinding variables */
  aux_tr = LOCAL_top_cp->cp_tr;
  while (aux_tr != TR) {
    aux_cell = TrailTerm(--TR);
    /* check for global or local variables */
    if (IsVarTerm(aux_cell)) {
      RESET_VARIABLE(aux_cell);
    }
#ifdef TABLING
    else {
      /* avoid frozen segments */
      aux_cell = (CELL) RepPair(aux_cell);
      if ((ADDR) aux_cell >= TrailBase) {
        TR = (tr_fr_ptr) aux_cell;
      }
    }
#endif TABLING
  }

#ifdef OPTYAP_ERRORS
  if (LOCAL_top_cp_on_stack != LOCAL_top_cp)
    OPTYAP_ERROR_MESSAGE("LOCAL_top_cp and LOCAL_top_cp_on_stack different (q_share_work)");
#endif /* OPTYAP_ERRORS */

  /* make sharing request */
  LOCK_WORKER(worker_p);
  if (BITMAP_member(GLOBAL_bm_idle_workers, worker_p) || REMOTE_request(worker_p) != MAX_WORKERS) {
    /* worker p is idle or has another request */
    UNLOCK_WORKER(worker_p);
    return FALSE;
  }
  LOCAL_reply_signal = no_reply;
  REMOTE_request(worker_p) = worker_id;
  UNLOCK_WORKER(worker_p);
  SIMICS_ATOMIC_SERVICE(ONE_MORE_REQUEST);

  /* wait for an answer */
  while (LOCAL_reply_signal == no_reply) continue;
  if (LOCAL_reply_signal == refused) {
    /* sharing request refused */
    SIMICS_ATOMIC_SERVICE(ONE_MORE_REFUSED_REQUEST);
    LOCAL_reply_signal = no_reply;
    return FALSE;
  }

  /* copy trail stack ? */
  LOCK(LOCAL_lock_signals);
  if (LOCAL_p_fase_signal > trail) {
    LOCAL_q_fase_signal = trail;
    UNLOCK(LOCAL_lock_signals);
    Q_COPY_TRAIL_FROM(worker_p);
  } else {
    UNLOCK(LOCAL_lock_signals);
    goto wait_ready;
  }

  /* copy global stack ? */
  LOCK(LOCAL_lock_signals);
  if (LOCAL_p_fase_signal > global) {
    LOCAL_q_fase_signal = global;
    UNLOCK(LOCAL_lock_signals);
    Q_COPY_GLOBAL_FROM(worker_p);
  } else {
    UNLOCK(LOCAL_lock_signals);
    goto wait_ready;
  }

  /* wait for local_ok signal */
  while (LOCAL_reply_signal < local_ok) continue;
  /* copy local stack ? */
  LOCK(LOCAL_lock_signals);
  if (LOCAL_p_fase_signal > local) {
    LOCAL_q_fase_signal = local;
    UNLOCK(LOCAL_lock_signals);
    Q_COPY_LOCAL_FROM(worker_p);
  } else UNLOCK(LOCAL_lock_signals);

wait_ready:
  REMOTE_reply_signal(worker_p) = install;
  while (LOCAL_reply_signal < ready) continue;

  /* install fase */
  /* At this point, TR and LOCAL_top_cp->cp_tr are equal */
  aux_tr = ((choiceptr) LOCAL_start_local_copy)->cp_tr;
  while (TR != aux_tr) {
    aux_cell = TrailTerm(--aux_tr);
    if (IsVarTerm(aux_cell)) {
      if (((CELL *)aux_cell >= H0  &&  (CELL *)aux_cell < LOCAL_top_cp->cp_h) ||
          (EQUAL_OR_YOUNGER_CP(LOCAL_top_cp, (choiceptr)aux_cell)  &&  (ADDR)aux_cell <= LocalBase))
        *((CELL *) aux_cell) = *((CELL *) (worker_offset(worker_p) + aux_cell));
    }
#ifdef TABLING
    else {
      /* avoid frozen segments */
      aux_cell = (CELL) RepPair(aux_cell);
      if ((ADDR) aux_cell >= TrailBase){
        aux_tr = (tr_fr_ptr) aux_cell;
      }
    }
#endif TABLING
  }

  LOCK_WORKER(worker_p);
  REMOTE_reply_signal(worker_p) = ready;
  REMOTE_request(worker_p) = MAX_WORKERS;
  UNLOCK_WORKER(worker_p);
  SCH_put_us_access_work(worker_id, worker_p);

  /* update local variables */
  TR = (tr_fr_ptr) LOCAL_end_trail_copy;
#ifdef TABLING
  H_FZ = DepFr_h_fz(LOCAL_top_dep_fr);
  B_FZ = DepFr_b_fz(LOCAL_top_dep_fr);
  TR_FZ = DepFr_tr_fz(LOCAL_top_dep_fr);
  LOCAL_top_cp_on_stack = 
#endif /* TABLING */
  LOCAL_top_cp = (choiceptr) LOCAL_start_local_copy;
  LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
  LOCAL_load = -DELAYED_RELEASE_LOAD;
  return TRUE;
}



/* ------------------------- **
**      Local functions      **
** ------------------------- */

static void share_private_nodes(int worker_q) {
  choiceptr AuxB;
  or_fr_ptr or_frame;

#ifdef OPTYAP_ERRORS
  if (YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack)) {
    OPTYAP_ERROR_MESSAGE("YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack) (share_private_nodes)");
  } else {
    AuxB = B;
    while (AuxB != LOCAL_top_cp) {
      if (YOUNGER_CP(LOCAL_top_cp, AuxB)) {
        OPTYAP_ERROR_MESSAGE("LOCAL_top_cp not in branch (share_private_nodes)");
        break;
      }
      if (EQUAL_OR_YOUNGER_CP(LOCAL_top_cp_on_stack, AuxB)) {
        OPTYAP_ERROR_MESSAGE("shared frozen segments in branch (share_private_nodes)");
        break;
      }
      AuxB = AuxB->cp_b;
    }
  }
#endif /* OPTYAP_ERRORS */

  AuxB = B;
#ifdef TABLING
  /* check if the branch is already shared */
  if (EQUAL_OR_YOUNGER_CP(LOCAL_top_cp_on_stack, AuxB)) {
    sg_fr_ptr sg_frame;
    dep_fr_ptr dep_frame;



printf("++++++++++++++++++++++++++++++++++++++++++++++\n");



#ifdef OPTYAP_ERRORS
    or_frame = LOCAL_top_or_fr;
    while (or_frame != REMOTE_top_or_fr(worker_q)) {
      if (YOUNGER_CP(OrFr_node(REMOTE_top_or_fr(worker_q)), OrFr_node(or_frame))) {
        OPTYAP_ERROR_MESSAGE("YOUNGER_CP(OrFr_node(REMOTE_top_or_fr(worker_q)), OrFr_node(or_frame)) (share_private_nodes)");
        break; 
      }
      or_frame = OrFr_next_on_stack(or_frame);
    }
#endif /* OPTYAP_ERRORS */
    /* update old shared nodes */
    or_frame = LOCAL_top_or_fr;
    while (or_frame != REMOTE_top_or_fr(worker_q)) {
      LOCK_OR_FRAME(or_frame);
      if (BITMAP_member(OrFr_hidden_members(or_frame), worker_id))
        BITMAP_insert(OrFr_hidden_members(or_frame), worker_q);
      else
        BITMAP_insert(OrFr_members(or_frame), worker_q);
      GLOBAL_branch_cut(worker_q, OrFr_depth(or_frame)) = GLOBAL_branch_cut(worker_id, OrFr_depth(or_frame));
      GLOBAL_branch_ltt(worker_q, OrFr_depth(or_frame)) = GLOBAL_branch_ltt(worker_id, OrFr_depth(or_frame));
      UNLOCK_OR_FRAME(or_frame);
      or_frame = OrFr_next_on_stack(or_frame);
    }

    /* update worker Q top subgoal frame */
    sg_frame = LOCAL_top_sg_fr;
    while (sg_frame && YOUNGER_CP(SgFr_gen_cp(sg_frame), GEN_CP(AuxB))) {
      sg_frame = SgFr_next(sg_frame);
    }
    REMOTE_top_sg_fr(worker_q) = sg_frame;

    /* update worker Q top dependency frame */
    dep_frame = LOCAL_top_dep_fr;
    while (YOUNGER_CP(DepFr_cons_cp(dep_frame), CONS_CP(AuxB))) {
      dep_frame = DepFr_next(dep_frame);
    }
    REMOTE_top_dep_fr(worker_q) = dep_frame;

    /* looking for cut messages */
    if (LOCAL_cut_cp != NULL) { 
      LOCK_WORKER(worker_id);
      CUT_make_cut_request(worker_q, LOCAL_cut_cp);
      UNLOCK_WORKER(worker_id);
    }

    /* update local load */
    LOCAL_load = -DELAYED_RELEASE_LOAD;
  } else
#endif /* TABLING */
  {
    int i, depth, leftmost;
    or_fr_ptr previous_or_frame;
#ifdef TABLING
    choiceptr next_node_on_branch;
    cons_cp_ptr cons_cp;
    dep_fr_ptr dep_frame;
    sg_fr_ptr sg_frame;
    CELL *stack, *stack_base, *stack_top;

    /* find top dependency frame above current choice point */
    dep_frame = LOCAL_top_dep_fr;
    while (EQUAL_OR_YOUNGER_CP(DepFr_cons_cp(dep_frame), CONS_CP(AuxB))) {
      dep_frame = DepFr_next(dep_frame);
    }
    /* initialize tabling auxiliary variables */ 
    cons_cp = DepFr_cons_cp(dep_frame);
    next_node_on_branch = NULL;
    stack_top = (CELL *)TrailTop;
    stack_base = stack = AuxSp;
#endif /* TABLING */

    /* initialize auxiliary variables */
    previous_or_frame = NULL;
    depth = OrFr_depth(LOCAL_top_or_fr);
    /* searching for leftmostness */
    leftmost = 1;
    for (i = 0; i < number_workers; i++) {
      if (i != worker_id && 
          BITMAP_member(OrFr_members(LOCAL_top_or_fr), i) &&
          GLOBAL_branch_ltt(i, depth) > GLOBAL_branch_ltt(worker_id, depth)) {
        leftmost = 0;
        break;
      }
    }

    /* sharing loop */
#ifdef TABLING
    while (YOUNGER_CP(AuxB, LOCAL_top_cp_on_stack)) {
#else
    while (AuxB != LOCAL_top_cp) {
#endif /* TABLING */
#ifdef OPTYAP_ERRORS
      if (next_node_on_branch) {
        choiceptr aux_cp;
        aux_cp = B;
        while (aux_cp != next_node_on_branch) {
          if (AuxB == aux_cp)
            OPTYAP_ERROR_MESSAGE("AuxB on branch (share_private_nodes)");
          if (YOUNGER_CP(next_node_on_branch, aux_cp)) {
            OPTYAP_ERROR_MESSAGE("next_node_on_branch not in branch (share_private_nodes)");
            break;
  	  }
          aux_cp = aux_cp->cp_b;
        }
      } else {
        choiceptr aux_cp;
        aux_cp = B;
        while (aux_cp != AuxB) {
          if (YOUNGER_CP(AuxB, aux_cp)) {
            OPTYAP_ERROR_MESSAGE("AuxB not in branch (share_private_nodes)");
            break;
	  }
          aux_cp = aux_cp->cp_b;
        }
      }
#endif /* OPTYAP_ERRORS */

      depth++;
      ALLOC_OR_FRAME(or_frame);
      INIT_LOCK(OrFr_lock(or_frame));
      OrFr_node(or_frame) = AuxB;
      OrFr_alternative(or_frame) = AuxB->cp_ap;
      OrFr_pend_cut_cp(or_frame) = NULL;
      OrFr_solutions(or_frame) = NULL;
      BITMAP_clear(OrFr_members(or_frame));
#ifdef TABLING
      OrFr_suspensions(or_frame) = NULL;
      BITMAP_clear(OrFr_hidden_members(or_frame));
      /* if next_node_on_branch is not NULL 
         this means that the current node is frozen */
      if (next_node_on_branch) {
        /* as the current node belongs to a frozen branch 
           we mark the workers as hidden */
        BITMAP_insert(OrFr_hidden_members(or_frame), worker_id);
        BITMAP_insert(OrFr_hidden_members(or_frame), worker_q);
        OrFr_nearest_leftnode(or_frame) = NULL;
      } else
#endif /* TABLING */
      {
        BITMAP_insert(OrFr_members(or_frame), worker_id);
        BITMAP_insert(OrFr_members(or_frame), worker_q);
        if (leftmost)
          OrFr_nearest_leftnode(or_frame) = OrFr_nearest_leftnode(LOCAL_top_or_fr);
        else
          OrFr_nearest_leftnode(or_frame) = LOCAL_top_or_fr;
      }

#ifdef YAPOR_ERRORS
      if (AuxB->cp_ap == GETWORK || AuxB->cp_ap == GETWORK_SEQ)
        YAPOR_ERROR_MESSAGE("choicepoint already shared (share_private_nodes)");
#endif /* YAPOR_ERRORS */
      if (AuxB->cp_ap && AuxB->cp_ap->YAMOP_SEQ) {
        AuxB->cp_ap = GETWORK_SEQ;
      } else {
        AuxB->cp_ap = GETWORK;
      }
      AuxB->cp_or_fr = or_frame;
      AuxB->cp_lub = 0;
      AuxB = AuxB->cp_b;

#ifdef TABLING
      /* check for frozen stack segments */
      if (YOUNGER_CP(cons_cp, CONS_CP(AuxB))) {
        if (! next_node_on_branch)
          next_node_on_branch = AuxB;
        STACK_PUSH(or_frame, stack, stack_top);
        STACK_PUSH(AuxB, stack, stack_top);
        AuxB = (choiceptr) cons_cp;
        dep_frame = DepFr_next(dep_frame);
        cons_cp = DepFr_cons_cp(dep_frame);
      } else if (cons_cp == CONS_CP(AuxB)) {
        dep_frame = DepFr_next(dep_frame);
        cons_cp = DepFr_cons_cp(dep_frame);
      }
      if (next_node_on_branch == AuxB)
        next_node_on_branch = NULL;
#endif /* TABLING */
#ifdef OPTYAP_ERRORS
      if (next_node_on_branch && YOUNGER_CP(next_node_on_branch, AuxB))
        OPTYAP_ERROR_MESSAGE("frozen node greater than next_node_on_branch (share_private_nodes)");
#endif /* OPTYAP_ERRORS */

      if (previous_or_frame) {
#ifdef TABLING
        OrFr_next_on_stack(previous_or_frame) =
#endif /* TABLING */
        OrFr_nearest_livenode(previous_or_frame) = OrFr_next(previous_or_frame) = or_frame;
      }
      previous_or_frame = or_frame;
    }

    /* initialize last or-frame pointer */
    or_frame = AuxB->cp_or_fr;
    if (previous_or_frame) {
#ifdef TABLING
      OrFr_next_on_stack(previous_or_frame) =
#endif /* TABLING */
      OrFr_nearest_livenode(previous_or_frame) = OrFr_next(previous_or_frame) = or_frame;
    }

#ifdef TABLING
    /* update or-frames stored in auxiliary stack */
    while (STACK_NOT_EMPTY(stack, stack_base)) {
      next_node_on_branch = (choiceptr) STACK_POP(stack);
      or_frame = (or_fr_ptr) STACK_POP(stack);
      OrFr_nearest_livenode(or_frame) = OrFr_next(or_frame) = next_node_on_branch->cp_or_fr;
    }
#endif /* TABLING */

    /* update depth */
    if (depth >= MAX_DEPTH)
      abort_optyap("maximum depth exceded (%d/%d) (share_private_nodes)", MAX_DEPTH, depth - MAX_DEPTH);
    or_frame = B->cp_or_fr;
#ifdef TABLING
    previous_or_frame = LOCAL_top_cp_on_stack->cp_or_fr;
    while (or_frame != previous_or_frame) {
#else
    while (or_frame != LOCAL_top_or_fr) {
#endif /* TABLING */
      yamop *alt;
      if ((alt = OrFr_alternative(or_frame)) != NULL) {
        GLOBAL_branch_ltt(worker_id, depth) = GLOBAL_branch_ltt(worker_q, depth) = alt->YAMOP_LTT + 1;
        GLOBAL_branch_cut(worker_id, depth) = GLOBAL_branch_cut(worker_q, depth) = alt->YAMOP_CUT;
      } else {
        GLOBAL_branch_ltt(worker_id, depth) = GLOBAL_branch_ltt(worker_q, depth) = 1;
        GLOBAL_branch_cut(worker_id, depth) = GLOBAL_branch_cut(worker_q, depth) = 0;
      }
#ifdef TABLING
      if (OrFr_nearest_leftnode(or_frame) == NULL && BITMAP_empty(OrFr_members(or_frame)))
        OrFr_nearest_leftnode(or_frame) = OrFr_next(or_frame);
#endif /* TABLING */
      OrFr_depth(or_frame) = depth--;
      or_frame = OrFr_next_on_stack(or_frame);
    }
#ifdef YAPOR_ERRORS
    if (depth != OrFr_depth(LOCAL_top_or_fr))
      YAPOR_ERROR_MESSAGE("incorrect depth value (share_private_nodes)");
#endif /* YAPOR_ERRORS */
#ifdef OPTYAP_ERRORS
    or_frame = B->cp_or_fr;
    while (or_frame != LOCAL_top_cp_on_stack->cp_or_fr) {
      AuxB = OrFr_node(or_frame);
      if (OrFr_next(or_frame) != AuxB->cp_b->cp_or_fr)
        OPTYAP_ERROR_MESSAGE("OrFr_next not in branch (share_private_nodes)");
      if (OrFr_nearest_livenode(or_frame) != AuxB->cp_b->cp_or_fr)
        OPTYAP_ERROR_MESSAGE("OrFr_nearest_livenode not in branch (share_private_nodes)");
      or_frame = OrFr_next_on_stack(or_frame);
    }
    or_frame = B->cp_or_fr;
    while (or_frame != LOCAL_top_cp_on_stack->cp_or_fr) {
      or_fr_ptr nearest_leftnode;
      nearest_leftnode = OrFr_nearest_leftnode(or_frame);
      if (nearest_leftnode == NULL) {
        if (BITMAP_empty(OrFr_members(or_frame)))
          OPTYAP_ERROR_MESSAGE("unexpected NULL value in OrFr_nearest_leftnode (share_private_nodes)");
      } else {
        AuxB = OrFr_node(or_frame);
        while (OrFr_node(nearest_leftnode) != AuxB) {
          if (YOUNGER_CP(OrFr_node(nearest_leftnode), AuxB)) {
            OPTYAP_ERROR_MESSAGE("OrFr_nearest_leftnode not in branch (share_private_nodes)");
            break;
          }
          AuxB = AuxB->cp_b;
        }
      }
      or_frame = OrFr_next_on_stack(or_frame);
    }
#endif /* OPTYAP_ERRORS */

    /* update old shared nodes */
    while (or_frame != REMOTE_top_or_fr(worker_q)) {
      LOCK_OR_FRAME(or_frame);
#ifdef TABLING
      if (BITMAP_member(OrFr_hidden_members(or_frame), worker_id))
        BITMAP_insert(OrFr_hidden_members(or_frame), worker_q);
      else
#endif /* TABLING */
        BITMAP_insert(OrFr_members(or_frame), worker_q);
      GLOBAL_branch_cut(worker_q, OrFr_depth(or_frame)) = GLOBAL_branch_cut(worker_id, OrFr_depth(or_frame));
      GLOBAL_branch_ltt(worker_q, OrFr_depth(or_frame)) = GLOBAL_branch_ltt(worker_id, OrFr_depth(or_frame));
      UNLOCK_OR_FRAME(or_frame);
      or_frame = OrFr_next_on_stack(or_frame);
    }

#ifdef TABLING
    /* update subgoal frames in the maintained private branches */
    sg_frame = LOCAL_top_sg_fr;
    while (sg_frame && YOUNGER_CP(SgFr_gen_cp(sg_frame), GEN_CP(B))) {
      choiceptr top_cp_on_branch;
      top_cp_on_branch = NORM_CP(SgFr_gen_cp(sg_frame));
      while (YOUNGER_CP(top_cp_on_branch, B)) {
        top_cp_on_branch = top_cp_on_branch->cp_b;
      }
      SgFr_gen_top_or_fr(sg_frame) = top_cp_on_branch->cp_or_fr;
      sg_frame = SgFr_next(sg_frame);
    }
    /* update worker Q top subgoal frame */
    REMOTE_top_sg_fr(worker_q) = sg_frame;
    /* update subgoal frames in the recently shared branches */
    while (sg_frame && YOUNGER_CP(SgFr_gen_cp(sg_frame), GEN_CP(LOCAL_top_cp_on_stack))) {
      SgFr_gen_top_or_fr(sg_frame) = SgFr_gen_cp(sg_frame)->gcp_or_fr;
      sg_frame = SgFr_next(sg_frame);
    }

    /* update dependency frames in the maintained private branches */
    dep_frame = LOCAL_top_dep_fr;
    while (YOUNGER_CP(DepFr_cons_cp(dep_frame), CONS_CP(B))) {
      choiceptr top_cp_on_branch;
      top_cp_on_branch = NORM_CP(DepFr_cons_cp(dep_frame));
      while (YOUNGER_CP(top_cp_on_branch, B)) {
        top_cp_on_branch = top_cp_on_branch->cp_b;
      }
      DepFr_top_or_fr(dep_frame) = top_cp_on_branch->cp_or_fr;
      dep_frame = DepFr_next(dep_frame);
    }
    /* update worker Q top dependency frame */
    REMOTE_top_dep_fr(worker_q) = dep_frame;
    /* update dependency frames in the recently shared branches */
    while (YOUNGER_CP(DepFr_cons_cp(dep_frame), CONS_CP(LOCAL_top_cp_on_stack))) {
      DepFr_top_or_fr(dep_frame) = DepFr_cons_cp(dep_frame)->ccp_or_fr;
      dep_frame = DepFr_next(dep_frame);
    }
#endif /* TABLING */
#ifdef OPTYAP_ERRORS
    dep_frame = LOCAL_top_dep_fr;
    while(dep_frame != GLOBAL_root_dep_fr) {
      choiceptr top_cp_on_branch;
      top_cp_on_branch = NORM_CP(DepFr_cons_cp(dep_frame));
      while (YOUNGER_CP(top_cp_on_branch, B)) {
        top_cp_on_branch = top_cp_on_branch->cp_b;
      }
      if (top_cp_on_branch->cp_or_fr != DepFr_top_or_fr(dep_frame))
        OPTYAP_ERROR_MESSAGE("Error on DepFr_top_or_fr (share_private_nodes)");
      dep_frame = DepFr_next(dep_frame);
    }
#endif /* OPTYAP_ERRORS */

    /* looking for cut messages */
    if (LOCAL_cut_cp != NULL) { 
      LOCK_WORKER(worker_id);
      CUT_make_cut_request(worker_q, LOCAL_cut_cp);
      UNLOCK_WORKER(worker_id);
    }

    /* update top shared node */
#ifdef TABLING
    LOCAL_top_cp_on_stack = 
#endif /* TABLING */
    LOCAL_top_cp = B;
    LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
    LOCAL_load = -DELAYED_RELEASE_LOAD;
  }


    /*
printf("   SHARING WORK %d->%d top_cp %x\n", worker_id, worker_q, LOCAL_top_cp);
    */


  return;
}
#endif /* ENV_COPY */
