/* -------------------- **
**      Prototypes      **
** -------------------- */

STD_PROTO(static inline void PUT_ME_IN_EXECUTING, (int));
STD_PROTO(static inline void PUT_ME_OUT_EXECUTING, (int));
STD_PROTO(static inline void PUT_ME_IN_FINISHED, (int));

STD_PROTO(static inline void SCH_update_local_or_tops, (void));
STD_PROTO(static inline int  SCH_any_request, (void));
STD_PROTO(static inline void SCH_put_us_access_work, (int,int));
STD_PROTO(static inline void SCH_put_me_out_of_access_work, (int));
STD_PROTO(static inline void SCH_set_load, (choiceptr));
STD_PROTO(static inline void SCH_new_alternative, (yamop *,yamop *));
#if defined(ENV_COPY) || defined(SBA)
STD_PROTO(static inline void SCH_check, (void));
#endif /* ENV_COPY || SBA */

STD_PROTO(static inline int CUT_have_cut_request, (void));
STD_PROTO(static inline int CUT_shared, (choiceptr));
STD_PROTO(static inline int CUT_wait_until_leftmost, (void));
STD_PROTO(static inline choiceptr CUT_leftmost_until, (void));
STD_PROTO(static inline void CUT_put_new_solution, (or_fr_ptr, ans_fr_ptr));
STD_PROTO(static inline void CUT_put_solutions_together, (sol_fr_ptr));
STD_PROTO(static inline void CUT_put_inherted_solutions, (sol_fr_ptr, or_fr_ptr));
STD_PROTO(static inline void CUT_move_up_to_cut_cp, (void));
STD_PROTO(static inline sol_fr_ptr CUT_put_not_cut_solutions_together, (sol_fr_ptr));
STD_PROTO(static inline void CUT_free_pending_solutions, (sol_fr_ptr));
STD_PROTO(static inline void CUT_make_cut_request, (int, choiceptr));
STD_PROTO(static inline void CUT_clear_cut_cp, (void));
STD_PROTO(static inline void CUT_continue_pending_cut, (void));



/* ---------------------- **
**      Engine Stuff      **
** ---------------------- */

static inline 
void PUT_ME_IN_EXECUTING(int w) {
  LOCK(GLOBAL_LOCKS_bm_executing_workers);
  BITMAP_insert(GLOBAL_bm_executing_workers, w);
  UNLOCK(GLOBAL_LOCKS_bm_executing_workers);
}

static inline 
void PUT_ME_OUT_EXECUTING(int w) {
  LOCK(GLOBAL_LOCKS_bm_executing_workers);
  BITMAP_delete(GLOBAL_bm_executing_workers, w);
  UNLOCK(GLOBAL_LOCKS_bm_executing_workers);
}

static inline 
void PUT_ME_IN_FINISHED(int w) {
  LOCK(GLOBAL_LOCKS_bm_finished_workers);
  BITMAP_insert(GLOBAL_bm_finished_workers, w);
  UNLOCK(GLOBAL_LOCKS_bm_finished_workers);
}


#define LOCK_OR_FRAME(fr)    LOCK(OrFr_lock(fr))
#define UNLOCK_OR_FRAME(fr)  UNLOCK(OrFr_lock(fr))


#define LOCK_WORKER(w)    LOCK(REMOTE_lock(w))
#define UNLOCK_WORKER(w)  UNLOCK(REMOTE_lock(w))



/* ------------------------- **
**      Scheduler Stuff      **
** ------------------------- */

static inline 
void SCH_update_local_or_tops(void) {
  LOCAL_top_cp = LOCAL_top_cp->cp_b;
  LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
  return;
}


static inline 
int SCH_any_request(void) {                                                
  return (LOCAL_request != MAX_WORKERS && LOCAL_reply_signal != install);
}


static inline 
void SCH_put_us_access_work(int q, int p) {
  LOCK(GLOBAL_LOCKS_bm_requestable_workers);
  BITMAP_insert(GLOBAL_bm_requestable_workers, q);
  BITMAP_insert(GLOBAL_bm_requestable_workers, p);
  UNLOCK(GLOBAL_LOCKS_bm_requestable_workers);
  return;
}


static inline 
void SCH_put_me_out_of_access_work(int p) {
  LOCK(GLOBAL_LOCKS_bm_requestable_workers);
  BITMAP_delete(GLOBAL_bm_requestable_workers, p);
  UNLOCK(GLOBAL_LOCKS_bm_requestable_workers);
  return;
}


static inline 
void SCH_set_load(choiceptr curb) {
  /*
  curb->cp_lub = curb->cp_b->cp_lub + curb->cp_b->cp_ap->YAMOP_LTT;
  if (LOCAL_load > 0) LOCAL_load = curb->cp_lub + curb->cp_ap->YAMOP_LTT;
  */
  if (!curb->cp_b->cp_ap->YAMOP_SEQ)
    curb->cp_lub = curb->cp_b->cp_lub + curb->cp_b->cp_ap->YAMOP_LTT;
  else
    curb->cp_lub = curb->cp_b->cp_lub;

  if (LOCAL_load >= 0) {
    if (!curb->cp_ap->YAMOP_SEQ)
      LOCAL_load = curb->cp_lub + curb->cp_ap->YAMOP_LTT;
    else
      LOCAL_load = curb->cp_lub;
  }
  return;
}


static inline 
void SCH_new_alternative(yamop *curpc, yamop *new) {
  OrFr_alternative(LOCAL_top_or_fr) = new;
  GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr)) = curpc->YAMOP_LTT;
  GLOBAL_branch_cut(worker_id, OrFr_depth(LOCAL_top_or_fr)) = curpc->YAMOP_CUT;
  UNLOCK_OR_FRAME(LOCAL_top_or_fr);
  return;
}


#if defined(ENV_COPY) || defined(SBA)
static inline 
void SCH_check(void) {
  if (LOCAL_load < 0)
    LOCAL_load++;

  LOCK_WORKER(worker_id);
  if (SCH_any_request()) {
    UNLOCK_WORKER(worker_id);
    p_share_work();
  } else {
    UNLOCK_WORKER(worker_id);
  }
  return;
}
#else /* ACOW */
/* cannot be made into function 
   because it jumps to a label */
#define SCH_check()                  \
        if (LOCAL_load < 0)          \
          LOCAL_load++;              \
                                     \
        LOCK_WORKER(worker_id);      \
        if (SCH_any_request()) {     \
          UNLOCK_WORKER(worker_id);  \
          if (p_share_work() == 0)   \
            goto shared_fail;        \
        } else {                     \
          UNLOCK_WORKER(worker_id);  \
        }
#endif /* ENV_COPY || SBA */


/* cannot be made into function because it 
   touches local variables HBREG and CPREG */
#define SCH_last_alternative(curpc, CP_PTR)    \
        H = HBREG = PROTECT_FROZEN_H(CP_PTR);  \
        CPREG = CP_PTR->cp_cp;		       \
        ENV = CP_PTR->cp_env;                  \
        SCH_new_alternative(curpc, NULL)


#define SCH_shared_cp()  (LOCAL_top_cp == B)


#define SCHEDULER_GET_WORK()  \
        if (get_work())       \
          goto shared_fail;   \
        else                  \
          goto shared_end



/* ------------------- **
**      Cut Stuff      **
** ------------------- */

static inline
int CUT_have_cut_request(void) {
  if (LOCAL_cut_cp != NULL) {
    SIMICS_SERVICE(STOP_COUNTER, 1);
    SIMICS_SERVICE(START_COUNTER, 4);
    LOCK_WORKER(worker_id);
    if (SCH_any_request()) {
      int q = LOCAL_request;
      REMOTE_reply_signal(q) = refused;
      LOCAL_request = MAX_WORKERS;
      SCH_put_me_out_of_access_work(worker_id);
    }
    UNLOCK_WORKER(worker_id);
    CUT_move_up_to_cut_cp();
    CUT_clear_cut_cp();
    SIMICS_SERVICE(STOP_COUNTER, 4);
    SIMICS_SERVICE(START_COUNTER, 1);
    return 1;
  }
  return 0;
}


static inline 
int CUT_shared(choiceptr cut_choicepoint) {
  if (YOUNGER_CP(LOCAL_top_cp, cut_choicepoint)) {
    if (CUT_have_cut_request())     
      return 0;
    SIMICS_SERVICE(START_COUNTER, 6);
    cut_shared(cut_choicepoint, NULL);
    SIMICS_SERVICE(STOP_COUNTER, 6);
  } else B = cut_choicepoint;
  return 1;
}


static inline
int CUT_wait_until_leftmost(void) {
  int i, depth;
  bitmap may_be_leftmost;
  or_fr_ptr check_node;

  if (BITMAP_empty(GLOBAL_bm_executing_workers))
    return 1; /* there is no parallel execution */

  check_node = LOCAL_top_or_fr;
  while (check_node) {
ciclo:
    if (CUT_have_cut_request())
      return 0;
    LOCK_WORKER(worker_id);
    if (SCH_any_request()) {
      UNLOCK_WORKER(worker_id);
#if defined(ENV_COPY) || defined(SBA)
      p_share_work();
#else /* ACOW */
      if (p_share_work() == 0)
        return 0;  /* goto shared_fail; */
#endif /* ENV_COPY || SBA */
    } else UNLOCK_WORKER(worker_id);

    depth = OrFr_depth(check_node);
    BITMAP_copy(may_be_leftmost, OrFr_members(check_node));
    BITMAP_delete(may_be_leftmost, worker_id);
    for (i = 0; i < number_workers; i++) {
      if (BITMAP_member(may_be_leftmost, i) && 
          GLOBAL_branch_ltt(i, depth) > GLOBAL_branch_ltt(worker_id, depth)) { 
        goto ciclo;
      }
    }
    check_node = OrFr_nearest_leftnode(check_node);
  }
  return 1;
}


static inline
choiceptr CUT_leftmost_until(void) {
  int i, depth, leftmost;
  bitmap may_be_leftmost;
  or_fr_ptr until_node;

  leftmost = 1;
  BITMAP_clear(may_be_leftmost);
  BITMAP_insert(may_be_leftmost, worker_id);
  until_node = LOCAL_top_or_fr;
  while (until_node != NULL) {
    depth = OrFr_depth(until_node);
    BITMAP_difference(may_be_leftmost, OrFr_members(until_node), may_be_leftmost);
    for (i = 0; i < number_workers; i++) {
      if (BITMAP_member(may_be_leftmost, i) &&
	  GLOBAL_branch_ltt(i, depth) > GLOBAL_branch_ltt(worker_id, depth)) {
	leftmost = 0;
	break;
      }
    }
    if (!leftmost) break;
    BITMAP_copy(may_be_leftmost, OrFr_members(until_node));
    until_node = OrFr_nearest_leftnode(until_node);
  }
  if (until_node != LOCAL_top_or_fr) {
    or_fr_ptr aux_fr = LOCAL_top_or_fr;
    while (OrFr_nearest_leftnode(aux_fr) != until_node) {
      LOCK_OR_FRAME(aux_fr);
      OrFr_nearest_leftnode(aux_fr) = until_node;
      UNLOCK_OR_FRAME(aux_fr);
      aux_fr = OrFr_next(aux_fr);
    }
    if (until_node == NULL)
      until_node = GLOBAL_root_or_fr;
  }
  return OrFr_node(until_node);
}


static inline 
void CUT_put_new_solution(or_fr_ptr frame, ans_fr_ptr actual_answer) {
  sol_fr_ptr *aux_sol;
  sol_fr_ptr alloc_sol;
  int ltt;

  ltt = GLOBAL_branch_ltt(worker_id, OrFr_depth(frame));
  aux_sol = & OrFr_solutions(frame);
  while (*aux_sol != NULL) {
    if (SolFr_ltt(*aux_sol) >= ltt) break;
    aux_sol = & SolFr_next(*aux_sol);
  }
  if (*aux_sol != NULL && SolFr_ltt(*aux_sol) == ltt) {
    AnsFr_next(SolFr_last(*aux_sol)) = actual_answer;
    SolFr_last(*aux_sol) = actual_answer;
  } else {
    ALLOC_SOLUTION_FRAME(alloc_sol);
    SolFr_next(alloc_sol) = *aux_sol;
    SolFr_ltt(alloc_sol) = ltt;
    SolFr_first(alloc_sol) = actual_answer;
    SolFr_last(alloc_sol) = actual_answer;
    *aux_sol = alloc_sol;
  }
  return;
}


static inline 
void CUT_put_solutions_together(sol_fr_ptr sol) {
  sol_fr_ptr aux_sol;
  while ((aux_sol = SolFr_next(sol)) != NULL) {
    AnsFr_next(SolFr_last(sol)) = SolFr_first(aux_sol);
    SolFr_last(sol) = SolFr_last(aux_sol);
    SolFr_next(sol) = SolFr_next(aux_sol);
    FREE_SOLUTION_FRAME(aux_sol);
  }
  return;
}


static inline 
void CUT_put_inherted_solutions(sol_fr_ptr sol, or_fr_ptr frame) {
  sol_fr_ptr *aux_sol;
  int ltt;

  ltt = GLOBAL_branch_ltt(worker_id, OrFr_depth(frame));
  aux_sol = & OrFr_solutions(frame);
  while (*aux_sol != NULL) {
    if (SolFr_ltt(*aux_sol) >= ltt) break;
    aux_sol = & SolFr_next(*aux_sol);
  }
  if (*aux_sol != NULL && SolFr_ltt(*aux_sol) == ltt) {
    AnsFr_next(SolFr_last(*aux_sol)) = SolFr_first(sol);
    SolFr_last(*aux_sol) = SolFr_last(sol);
    FREE_SOLUTION_FRAME(sol);
  } else {
    SolFr_next(sol) = *aux_sol;
    SolFr_ltt(sol) = ltt;
    *aux_sol = sol;
  }
  return;
}


static inline 
void CUT_move_up_to_cut_cp(void) {
  while (YOUNGER_CP(LOCAL_top_cp, LOCAL_cut_cp)) {
    LOCK_OR_FRAME(LOCAL_top_or_fr);
    BITMAP_delete(OrFr_members(LOCAL_top_or_fr), worker_id);
    if (BITMAP_empty(OrFr_members(LOCAL_top_or_fr))) {
      CUT_free_pending_solutions(OrFr_solutions(LOCAL_top_or_fr));
      FREE_OR_FRAME(LOCAL_top_or_fr);
    } else UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    SCH_update_local_or_tops();
  }
  return;
}


static inline 
sol_fr_ptr CUT_put_not_cut_solutions_together(sol_fr_ptr sol) {
  sol_fr_ptr aux_sol1, aux_sol2;

  aux_sol1 = OrFr_solutions(LOCAL_top_or_fr);
  OrFr_solutions(LOCAL_top_or_fr) = NULL;
  while (aux_sol1 != NULL &&
	 SolFr_ltt(aux_sol1) < GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr))) {
    ans_fr_ptr aux_answer1, aux_answer2;
    aux_answer1 = SolFr_first(aux_sol1);
    while (aux_answer1 != NULL) {
      aux_answer2 = aux_answer1;
      aux_answer1 = AnsFr_next(aux_answer1);
      FREE_ANSWER_FRAME(aux_answer2);
    }
    aux_sol2 = aux_sol1;
    aux_sol1 = SolFr_next(aux_sol1);
    FREE_SOLUTION_FRAME(aux_sol2);
  }
  if (aux_sol1 != NULL) {
    while ((aux_sol2 = SolFr_next(aux_sol1)) != NULL) {
      AnsFr_next(SolFr_last(aux_sol1)) = SolFr_first(aux_sol2);
      SolFr_last(aux_sol1) = SolFr_last(aux_sol2);
      SolFr_next(aux_sol1) = SolFr_next(aux_sol2);
      FREE_SOLUTION_FRAME(aux_sol2);
    }
    if (sol != NULL) {
      AnsFr_next(SolFr_last(sol)) = SolFr_first(aux_sol1);
      SolFr_last(sol) = SolFr_last(aux_sol1);
      SolFr_next(sol) = SolFr_next(aux_sol1);
      FREE_SOLUTION_FRAME(aux_sol1);
    } else sol = aux_sol1;
  }
  return (sol);
}


static inline 
void CUT_free_pending_solutions(sol_fr_ptr sol) {
  while (sol != NULL) {
    sol_fr_ptr aux_sol;
    ans_fr_ptr aux1_answer, aux2_answer;
    aux1_answer = SolFr_first(sol);
    while (aux1_answer != NULL) {
      aux2_answer = aux1_answer;
      aux1_answer = AnsFr_next(aux1_answer);
      FREE_ANSWER_FRAME(aux2_answer);
    }
    aux_sol = sol;
    sol = SolFr_next(sol);
    FREE_SOLUTION_FRAME(aux_sol);
  }
  return;
}


static inline 
void CUT_make_cut_request(int prune_worker, choiceptr prune_cp) {
  LOCK_WORKER(prune_worker);
  if (YOUNGER_CP(REMOTE_cut_cp(prune_worker), prune_cp))
    REMOTE_cut_cp(prune_worker) = prune_cp;
  UNLOCK_WORKER(prune_worker);
  return;
}


static inline 
void CUT_clear_cut_cp(void) {
  LOCK_WORKER(worker_id); 
  if (LOCAL_cut_cp != NULL && EQUAL_OR_YOUNGER_CP(LOCAL_cut_cp, LOCAL_top_cp))
    LOCAL_cut_cp = NULL;
  UNLOCK_WORKER(worker_id);
  return;
}


static inline 
void CUT_continue_pending_cut(void) {
  sol_fr_ptr solution;
  choiceptr pending_cut_cp;  

  pending_cut_cp = OrFr_pend_cut_cp(LOCAL_top_or_fr);
  SIMICS_SERVICE(START_COUNTER, 6);
  GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr)) = OrFr_pend_cut_ltt(LOCAL_top_or_fr);
  solution = CUT_put_not_cut_solutions_together(NULL);
  if (BITMAP_empty(OrFr_members(LOCAL_top_or_fr))) {
    FREE_OR_FRAME(LOCAL_top_or_fr);
  } else UNLOCK_OR_FRAME(LOCAL_top_or_fr);
  SCH_update_local_or_tops();
  cut_shared(pending_cut_cp, solution);
  SIMICS_SERVICE(STOP_COUNTER, 6);
  return;
}
