/* --------
   Includes
   -------- */

#include "Yap.h"
#ifdef SBA
#include "Yatom.h"
#include "Heap.h"
#include "or.macros.h"
#include <stdio.h>



/* ---------------------------
   Local functions declaration
   --------------------------- */

static void share_private_nodes(int worker_q);


/* ----------------
   Global functions
   ---------------- */

#define COMPUTE_SEGMENTS_TO_COPY_TO(Q)                                   \
        REMOTE_end_local_copy(Q)      = (CELL) (B);                      \
        REMOTE_start_local_copy(Q)    = (CELL) (REMOTE_top_cp(Q))


void make_root_choice_point(void) {
  if (worker_id == 0) {
    LOCAL_top_cp = GLOBAL_root_cp = OrFr_node(GLOBAL_root_or_fr) = B;
    B->cp_h = H0;
    B->cp_ap = GETWORK;
    B->cp_or_fr = GLOBAL_root_or_fr;
    B->cp_lub = 0;
  } else {
    B = LOCAL_top_cp = GLOBAL_root_cp;
    TR = B->cp_tr;
  }
  LOCAL_top_or_fr = GLOBAL_root_or_fr;
  LOCAL_load = -DELAYED_RELEASE_LOAD;
  LOCAL_cut_cp = NULL;
  GLOBAL_branch_ltt(worker_id, 0) = 0;
  GLOBAL_branch_cut(worker_id, 0) = 0;
  H_FZ = H_BASE;
  B_FZ = B_BASE;
  TR_FZ = TR_BASE;
  return;
}


void free_root_choice_point(void) {
  B = LOCAL_top_cp->cp_b;
  LOCAL_top_cp = B_BASE;
  H_FZ = H_BASE;
  B_FZ = B_BASE;
  TR_FZ = TR_BASE;
  return;
}


void p_share_work(void) {
  int worker_q = LOCAL_request;

  if (! BITMAP_member(OrFr_members(REMOTE_top_or_fr(worker_q)), worker_id) ||
      B == REMOTE_top_cp(worker_q) ||
      (LOCAL_load <= 0 && OrFr_nearest_livenode(LOCAL_top_or_fr) == NULL)) {
    /* refuse sharing request */
    REMOTE_reply_signal(worker_q) = refused;
    LOCAL_request = MAX_WORKERS;
    SCH_put_me_out_of_access_work(worker_id);
    return;
  }
  /* sharing request accepted */
  COMPUTE_SEGMENTS_TO_COPY_TO(worker_q);
  LOCAL_reply_signal = no_reply;
  REMOTE_reply_signal(worker_q) = accepted;
  share_private_nodes(worker_q);
  REMOTE_reply_signal(worker_q) = ready;
  while (LOCAL_reply_signal < ready) continue;
  return;
}


int q_share_work(int worker_p) {
  register tr_fr_ptr aux_tr;
  register CELL aux_cell;

  LOCK_OR_FRAME(LOCAL_top_or_fr);
  /* looking for cut messages */
  if (LOCAL_cut_cp != NULL) {
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    CUT_move_up_to_cut_cp();
    CUT_clear_cut_cp();
    return 0;
  }
  if (! BITMAP_member(OrFr_members(LOCAL_top_or_fr), worker_p)) {
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    return 0;
  }
  GLOBAL_branch_cut(worker_id, OrFr_depth(LOCAL_top_or_fr)) = 
    GLOBAL_branch_cut(worker_p, OrFr_depth(LOCAL_top_or_fr));
  GLOBAL_branch_ltt(worker_id, OrFr_depth(LOCAL_top_or_fr)) = 
    GLOBAL_branch_ltt(worker_p, OrFr_depth(LOCAL_top_or_fr));

  /* restart pending cuts */
  if (OrFr_pend_cut_cp(LOCAL_top_or_fr) != NULL && no_workers_left_pending_cut()) {
    BITMAP_delete(OrFr_members(LOCAL_top_or_fr), worker_id);
    CUT_continue_pending_cut();
    return 0;
  }
  UNLOCK_OR_FRAME(LOCAL_top_or_fr);

  /* unbinding variables */
  aux_tr = LOCAL_top_cp->cp_tr;
  while (aux_tr != TR) {
    aux_cell = TrailTerm(--TR);
    /* check for global or local variables */
    if (IsVarTerm(aux_cell)) {
      /* clean up the trail when we backtrack */
      if (Unsigned((Int)(aux_cell)-(Int)(H_FZ)) >
	  Unsigned((Int)(B_FZ)-(Int)(H_FZ))) {
	RESET_VARIABLE(STACK_TO_SBA(aux_cell));
      } else {
	RESET_VARIABLE(aux_cell);
      }
    }
    else {
      /* avoid frozen segments */
      if (IsPairTerm(aux_cell) && (ADDR) RepPair(aux_cell) > HeapTop) {
	TR = (tr_fr_ptr) RepPair(aux_cell);
      }
    }
  }

  /* make sharing request */
  LOCK_WORKER(worker_p);
  if (BITMAP_member(GLOBAL_bm_idle_workers, worker_p) || REMOTE_request(worker_p) != MAX_WORKERS) {
    /* worker p is idle or has another request */
    UNLOCK_WORKER(worker_p);
    return 0;
  }
  LOCAL_reply_signal = no_reply;
  REMOTE_request(worker_p) = worker_id;
  UNLOCK_WORKER(worker_p);
  SIMICS_ATOMIC_SERVICE(ONE_MORE_REQUEST);

  /* wait for an answer */
  while (LOCAL_reply_signal == no_reply) continue;
  if (LOCAL_reply_signal == refused) {
    /* sharing request refused */
    SIMICS_ATOMIC_SERVICE(ONE_MORE_REFUSED_REQUEST);
    LOCAL_reply_signal = no_reply;
    return 0;
  }
  while (LOCAL_reply_signal < ready) continue;

  /* install fase */
  /* At this point, TR and LOCAL_top_cp->cp_tr are equal */
  TR = ((choiceptr)LOCAL_start_local_copy)->cp_tr;
  aux_tr = ((choiceptr) LOCAL_end_local_copy)->cp_tr;
  while (TR != aux_tr) {
    aux_cell = TrailTerm(--aux_tr);
    if (IsVarTerm(aux_cell)) {
      CELL *ptr = STACK_TO_SBA(aux_cell);
      *ptr = TrailVal(aux_tr);
    } else if ((ADDR) RepPair(aux_cell) >= HeapTop) {
    /* avoid frozen segments */
      aux_tr = (tr_fr_ptr) RepPair(aux_cell);
    }
  }

  LOCK_WORKER(worker_p);
  REMOTE_reply_signal(worker_p) = ready;
  REMOTE_request(worker_p) = MAX_WORKERS;
  UNLOCK_WORKER(worker_p);
  SCH_put_us_access_work(worker_id, worker_p);

  /* update local variables */
  LOCAL_top_cp = (choiceptr) LOCAL_end_local_copy;
  LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
  LOCAL_load = -DELAYED_RELEASE_LOAD;
  TR = LOCAL_top_cp->cp_tr;

  return 1;
}

/* ---------------
   Local functions
   --------------- */

static void share_private_nodes(int worker_q) {
  choiceptr AuxB;
  or_fr_ptr or_frame;

  AuxB = B;
  {
    int i, depth, leftmost;
    or_fr_ptr previous_or_frame;
    /* initialize auxiliary variables */
    previous_or_frame = NULL;
    depth = OrFr_depth(LOCAL_top_or_fr);
    /* searching for leftmostness */
    leftmost = 1;
    for (i = 0; i < number_workers; i++) {
      if (i != worker_id && 
          BITMAP_member(OrFr_members(LOCAL_top_or_fr), i) &&
          GLOBAL_branch_ltt(i, depth) > GLOBAL_branch_ltt(worker_id, depth)) {
        leftmost = 0;
        break;
      }
    }
    /* sharing loop */
    while (AuxB != LOCAL_top_cp) {
      depth++;
      ALLOC_OR_FRAME(or_frame);
      INIT_LOCK(OrFr_lock(or_frame));
      OrFr_node(or_frame) = AuxB;
      OrFr_alternative(or_frame) = AuxB->cp_ap;
      OrFr_pend_cut_cp(or_frame) = NULL;
      OrFr_solutions(or_frame) = NULL;
      BITMAP_clear(OrFr_members(or_frame));
      BITMAP_insert(OrFr_members(or_frame), worker_id);
      BITMAP_insert(OrFr_members(or_frame), worker_q);
      if (leftmost)
        OrFr_nearest_leftnode(or_frame) = OrFr_nearest_leftnode(LOCAL_top_or_fr);
      else
        OrFr_nearest_leftnode(or_frame) = LOCAL_top_or_fr;
      if (AuxB->cp_ap && AuxB->cp_ap->YAMOP_SEQ) {
        AuxB->cp_ap = GETWORK_SEQ;
      } else {
        AuxB->cp_ap = GETWORK;
      }
      AuxB->cp_or_fr = or_frame;
      AuxB->cp_lub = 0;
      AuxB = AuxB->cp_b;
      if (previous_or_frame) {
        OrFr_nearest_livenode(previous_or_frame) = OrFr_next(previous_or_frame) = or_frame;
      }
      previous_or_frame = or_frame;
    }
    /* initialize last or-frame pointer */
    or_frame = AuxB->cp_or_fr;
    if (previous_or_frame) {
      OrFr_nearest_livenode(previous_or_frame) = OrFr_next(previous_or_frame) = or_frame;
    }
    /* update depth */
    if (depth >= MAX_DEPTH)
      abort_optyap("maximum depth exceded (%d/%d) (share_private_nodes)", MAX_DEPTH, depth - MAX_DEPTH);
    or_frame = B->cp_or_fr;

    while (or_frame != LOCAL_top_or_fr) {
      yamop *alt;
      if ((alt = OrFr_alternative(or_frame)) != NULL) {
        GLOBAL_branch_ltt(worker_id, depth) = GLOBAL_branch_ltt(worker_q, depth) = alt->YAMOP_LTT + 1;
        GLOBAL_branch_cut(worker_id, depth) = GLOBAL_branch_cut(worker_q, depth) = alt->YAMOP_CUT;
      } else {
        GLOBAL_branch_ltt(worker_id, depth) = GLOBAL_branch_ltt(worker_q, depth) = 1;
        GLOBAL_branch_cut(worker_id, depth) = GLOBAL_branch_cut(worker_q, depth) = 0;
      }
      OrFr_depth(or_frame) = depth--;
      or_frame = OrFr_next_on_stack(or_frame);
    }

    /* update old shared nodes */
    while (or_frame != REMOTE_top_or_fr(worker_q)) {
      LOCK_OR_FRAME(or_frame);
      BITMAP_insert(OrFr_members(or_frame), worker_q);
      GLOBAL_branch_cut(worker_q, OrFr_depth(or_frame)) = GLOBAL_branch_cut(worker_id, OrFr_depth(or_frame));
      GLOBAL_branch_ltt(worker_q, OrFr_depth(or_frame)) = GLOBAL_branch_ltt(worker_id, OrFr_depth(or_frame));
      UNLOCK_OR_FRAME(or_frame);
      or_frame = OrFr_next_on_stack(or_frame);
    }
    /* update frozen registers */
    B_FZ  = B;
    H_FZ  = B->cp_h;
    TR_FZ = B->cp_tr;
    /* move conditional bindings to BA */
    {
      tr_fr_ptr top, tr_ptr;
      top = LOCAL_top_cp->cp_tr;
      tr_ptr = TR;
      while (tr_ptr != top) {
        CELL aux_cell = TrailTerm(--tr_ptr);
        if (IsVarTerm(aux_cell) && ((CELL *)aux_cell < H_FZ || (choiceptr)aux_cell > B_FZ) ) {
	  CELL *ptr = STACK_TO_SBA(aux_cell);
  	  *ptr = TrailVal(tr_ptr);
	  *(CELL *)aux_cell = (CELL)ptr;
        } else if (IsPairTerm(aux_cell) && (ADDR) RepPair(aux_cell) > HeapTop) {
	  /* avoid frozen segments */
  	  aux_cell = (CELL) RepPair(aux_cell);
	  tr_ptr = (tr_fr_ptr) aux_cell;
        }
      }
    }
    /* looking for cut messages */
    if (LOCAL_cut_cp != NULL) { 
      LOCK_WORKER(worker_id);
      CUT_make_cut_request(worker_q, LOCAL_cut_cp);
      UNLOCK_WORKER(worker_id);
    }
    /* update top shared node */
    LOCAL_top_cp = B;
    LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
    LOCAL_load = -DELAYED_RELEASE_LOAD;
  }
}

#endif /* SBA */
