/* ------------------------------------------------ **
**      Tabling instructions: auxiliary macros      **
** ------------------------------------------------ */

#define store_new_subgoal(PTR, ARITY, AP, SG_FR, BOTTOM_SG_NODE)            \
        { register CELL *pt_args;                                           \
          register gen_cp_ptr gcp;                                          \
                                                                            \
          /* store args */                                                  \
          pt_args = XREGS + (ARITY);                                        \
	  while (pt_args > XREGS) {                                         \
            register CELL aux_arg = pt_args[0];                             \
            --PTR;                                                          \
            --pt_args;                                                      \
            *PTR = aux_arg;                                                 \
	  }                                                                 \
          /* initialize generator choice point pointer */                   \
          gcp = --GEN_CP(PTR);                                              \
          /* store subgoal frame and adjust top subgoal frame pointer */    \
          new_subgoal_frame(SG_FR, BOTTOM_SG_NODE, gcp, LOCAL_top_sg_fr);   \
          LOCAL_top_sg_fr = SG_FR;                                          \
          /* store generator choice point */                                \
          HBREG = H;                                                        \
          gcp->gcp_tr = TR;            	                                    \
          gcp->gcp_ap = (yamop *)(AP);                                      \
          gcp->gcp_h  = H;                                                  \
          gcp->gcp_b  = B;                                                  \
          gcp->gcp_env = ENV;                                               \
          gcp->gcp_cp = CPREG;                                              \
          gcp->gcp_sg_fr = SG_FR;                                           \
          set_cut((CELL *)gcp, B);                                          \
          B = NORM_CP(gcp);                                                 \
          YAPOR_SET_LOAD(B);                                                \
        }


#define restore_generator_node(PTR, ARITY, AP)          \
        { register CELL *pt_args, *x_args;              \
          register gen_cp_ptr gcp = (gen_cp_ptr)PTR;    \
                                                        \
          /* restore generator choice point */          \
          H = HBREG = PROTECT_FROZEN_H(NORM_CP(gcp));   \
          CPREG = gcp->gcp_cp;                          \
          ENV = gcp->gcp_env;                           \
          YAPOR_update_alternative(PREG, (yamop *) AP)  \
          gcp->gcp_ap = (yamop *) AP;                   \
          /* restore args */                            \
          pt_args = (CELL *)(gcp + 1) + ARITY;          \
          x_args = XREGS + 1 + ARITY;                   \
          while (x_args > XREGS + 1) {                  \
            register CELL x = pt_args[-1];              \
            --x_args;                                   \
            --pt_args;                                  \
            *x_args = x;                                \
	  }                                             \
        }


#define store_consumer_node(PTR, SG_FR)                                           \
        { cons_cp_ptr ccp;                                                        \
          dep_fr_ptr new_dep_fr;                                                  \
          choiceptr leader_cp;                                                    \
          int leader_dep_on_stack;                                                \
                                                                                  \
          /* find leader node */                                                  \
          find_leader_node(SG_FR, leader_cp, leader_dep_on_stack);                \
	  /* initialize ccp */                                                    \
          ccp = --CONS_CP(PTR);                                                   \
          /* adjust freeze registers */                                           \
          H_FZ = H;                                                               \
          B_FZ = NORM_CP(ccp);                 	                                  \
          TR_FZ = TR;                                                             \
          /* store dependency frame and adjust top dependency frame pointer */    \
          new_dependency_frame(new_dep_fr, leader_dep_on_stack, LOCAL_top_or_fr,  \
                               leader_cp, ccp, SG_FR, LOCAL_top_dep_fr);          \
          LOCAL_top_dep_fr = new_dep_fr;                                          \
          /* store consumer choice point */                                       \
          HBREG = H;                                                              \
          ccp->ccp_tr = TR;         	                                          \
          ccp->ccp_ap = ANSWER_RETURN;                                            \
          ccp->ccp_h  = H;                                                        \
          ccp->ccp_b  = B;                                                        \
          ccp->ccp_env= ENV;                                                      \
          ccp->ccp_cp = CPREG;                                                    \
          ccp->ccp_dep_fr = LOCAL_top_dep_fr;                                     \
          set_cut(((CELL *)ccp), B);                                              \
          B = NORM_CP(ccp);                                                       \
          YAPOR_SET_LOAD(B);                                                      \
        }


#define allocate_environment(PTR)  \
        PTR[E_CP] = (CELL) CPREG;  \
        PTR[E_E] = (CELL) ENV;     \
        PTR[E_B] = (CELL) B;       \
        ENV = PTR



/* ------------------------------ **
**      Tabling instructions      **
** ------------------------------ */  

  PBOp(table_try_me_single, ld)
    sg_node_ptr sg_node;
    sg_fr_ptr sg_fr;
    CELL *Yaddr;

    TABLING_INSTR_DEBUG_MESSAGE("table_try_me_single", PREG->u.ld.p, XREGS+1);
    Yaddr = Y;
    check_trail();
    sg_node = subgoal_search(PREG->u.ld.s, TabEnt_subgoal_trie(PREG->u.ld.te), &Yaddr);
    Y = Yaddr;
    /* the child field of the returned node points to the subgoal frame */
    LOCK(TrNode_lock(sg_node));
    if (TrNode_child(sg_node) == NULL) {
      /* new subgoal */
      store_new_subgoal(Y, PREG->u.ld.s, COMPLETION, sg_fr, sg_node);
      TABLING_DEBUG_MESSAGE("          -> new subgoal (B %x sg_fr %x)",
                            (unsigned int)B, (unsigned int)sg_fr);
      TrNode_child(sg_node) = (sg_node_ptr)sg_fr;
      UNLOCK(TrNode_lock(sg_node));
      PREG = NEXTOP(PREG, ld);
      PREFETCH_OP(PREG);
      allocate_environment(Y);
      GONext();
    }
    /* subgoal not new */
    UNLOCK(TrNode_lock(sg_node));
    sg_fr = (sg_fr_ptr) TrNode_child(sg_node);
    if (SgFr_completed(sg_fr)) {
      ans_node_ptr ans_node;
      /* subgoal completed */
      TABLING_DEBUG_MESSAGE("          -> subgoal completed (sg_fr %x)",
                            (unsigned int)sg_fr);
      ans_node = SgFr_answer_trie(sg_fr);
      if (IS_ANSWER_LEAF_NODE(ans_node)) {
        /* yes answer --> procceed */
        PREG = (yamop *)CPREG;
        PREFETCH_OP(PREG);
        Y = ENV;
        GONext();
      } else if (TrNode_child(ans_node) == NULL) {
        /* no answers --> fail */
        goto fail;
      }
#ifdef TABLING_DEBUG
      {
        char answer_str[1000];
        int answer_arity[1000];
        TABLING_DEBUG_MESSAGE("          -> show answer trie");
        answer_arity[0] = 0;
        show_answer_trie(TrNode_child(ans_node), answer_str, 0, answer_arity, 0);
      }
#endif /* TABLING_DEBUG */
      /* load answers */
      *--Y = 0;  /* vars_arity */
      *--Y = 0;  /* heap_arity */
      PREG = (yamop *) TrNode_child(ans_node);
      PREFETCH_OP(PREG);
      GONext();
    }
    /* subgoal not completed */
    store_consumer_node(Y, sg_fr);
    TABLING_DEBUG_MESSAGE("          -> subgoal not completed (B %x sg_fr %x)",
                           (unsigned int)B, (unsigned int)sg_fr);
#ifdef OPTYAP_ERRORS
    if (GLOBAL_execution_mode == PARALLEL_MODE) {
      choiceptr aux_cp;
      aux_cp = B;
      while (YOUNGER_CP(aux_cp, LOCAL_top_cp_on_stack))
        aux_cp = aux_cp->cp_b;
      if (aux_cp->cp_or_fr != DepFr_top_or_fr(LOCAL_top_dep_fr))
        OPTYAP_ERROR_MESSAGE("Error on DepFr_top_or_fr (table_try_me_single)");
    }
#endif /* OPTYAP_ERRORS */
    /* execute the table_answer_return instruction */
    goto answer_return;
  ENDPBOp();



  PBOp(table_try_me, ld)
    sg_node_ptr sg_node;
    sg_fr_ptr sg_fr;
    CELL *Yaddr;

    TABLING_INSTR_DEBUG_MESSAGE("table_try_me", PREG->u.ld.p, XREGS+1);
    Yaddr = Y;
    check_trail();
    sg_node = subgoal_search(PREG->u.ld.s, TabEnt_subgoal_trie(PREG->u.ld.te), &Yaddr);
    Y = Yaddr;
    /* the child field of the returned node points to the subgoal frame */
    LOCK(TrNode_lock(sg_node));
    if (TrNode_child(sg_node) == NULL) {
      /* new subgoal */
      store_new_subgoal(Y, PREG->u.ld.s, PREG->u.ld.d, sg_fr, sg_node);
      TABLING_DEBUG_MESSAGE("          -> new subgoal (B %x sg_fr %x)",
                            (unsigned int)B, (unsigned int)sg_fr);
      TrNode_child(sg_node) = (sg_node_ptr)sg_fr;
      UNLOCK(TrNode_lock(sg_node));
      PREG = NEXTOP(PREG, ld);
      PREFETCH_OP(PREG);
      allocate_environment(Y);
      GONext();
    }
    /* subgoal not new */
    UNLOCK(TrNode_lock(sg_node));
    sg_fr = (sg_fr_ptr) TrNode_child(sg_node);
    if (SgFr_completed(sg_fr)) {
      ans_node_ptr ans_node;
      /* subgoal completed */
      TABLING_DEBUG_MESSAGE("          -> subgoal completed (sg_fr %x)",
                            (unsigned int)sg_fr);
      ans_node = SgFr_answer_trie(sg_fr);
      if (IS_ANSWER_LEAF_NODE(ans_node)) {
        /* yes answer --> procceed */
        PREG = (yamop *)CPREG;
        PREFETCH_OP(PREG);
        Y = ENV;
        GONext();
      } else if (TrNode_child(ans_node) == NULL) {
        /* no answers --> fail */
        goto fail;
      }
#ifdef TABLING_DEBUG
      {
        char answer_str[1000];
        int answer_arity[1000];
        TABLING_DEBUG_MESSAGE("          -> show answer trie");
        answer_arity[0] = 0;
        show_answer_trie(TrNode_child(ans_node), answer_str, 0, answer_arity, 0);
      }
#endif /* TABLING_DEBUG */
      /* load answers */
      *--Y = 0;  /* vars_arity */
      *--Y = 0;  /* heap_arity */
      PREG = (yamop *) TrNode_child(ans_node);
      PREFETCH_OP(PREG);
      GONext();
    }
    /* subgoal not completed */
    store_consumer_node(Y, sg_fr);
    TABLING_DEBUG_MESSAGE("          -> subgoal not completed (B %x sg_fr %x)",
                          (unsigned int)B, (unsigned int)sg_fr);
#ifdef OPTYAP_ERRORS
    if (GLOBAL_execution_mode == PARALLEL_MODE) {
      choiceptr aux_cp;
      aux_cp = B;
      while (YOUNGER_CP(aux_cp, LOCAL_top_cp_on_stack))
        aux_cp = aux_cp->cp_b;
      if (aux_cp->cp_or_fr != DepFr_top_or_fr(LOCAL_top_dep_fr))
        OPTYAP_ERROR_MESSAGE("Error on DepFr_top_or_fr (table_try_me)");
    }
#endif /* OPTYAP_ERRORS */
    /* execute the table_answer_return instruction */
    goto answer_return;
  ENDPBOp();



  Op(table_retry_me, ld)
    restore_generator_node(B, PREG->u.ld.s, PREG->u.ld.d);
    TABLING_INSTR_DEBUG_MESSAGE("table_retry_me", PREG->u.ld.p, XREGS+1);
    Y = (CELL *) PROTECT_FROZEN_B(B);
    set_cut(Y, B->cp_b);
    allocate_environment(Y);
    TABLING_DEBUG_MESSAGE("          -> (B %x sg_fr %x)",
                          (unsigned int)B, (unsigned int)GEN_CP(B)->gcp_sg_fr);
    PREG = NEXTOP(PREG,ld);
    GONext();
  ENDOp();



  Op(table_trust_me, ld)
    restore_generator_node(B, PREG->u.ld.s, COMPLETION);
    TABLING_INSTR_DEBUG_MESSAGE("table_trust_me", PREG->u.ld.p, XREGS+1);
    Y = (CELL *) PROTECT_FROZEN_B(B);
    set_cut(Y, B->cp_b);
    allocate_environment(Y);
    TABLING_DEBUG_MESSAGE("          -> (B %x sg_fr %x)",
                          (unsigned int)B, (unsigned int)GEN_CP(B)->gcp_sg_fr);
    PREG = NEXTOP(PREG,ld);
    GONext();
  ENDOp();



  BOp(table_answer_return, ld)
    TABLING_DEBUG_MESSAGE("TAB INSTR -> table_answer_return");
    TABLING_DEBUG_MESSAGE("          -> (B %x)",
                          (unsigned int)B);
#ifdef YAPOR
    if (SCH_shared_cp()) {
      UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    }
#endif /* YAPOR */
    goto answer_return;
  ENDBOp();



answer_return:
    INIT_PREFETCH()
    dep_fr_ptr dep_fr;
    sg_fr_ptr sg_fr;
    ans_node_ptr ans_node;

    TABLING_DEBUG_MESSAGE("TAB STRAT -> answer_return");
    dep_fr = CONS_CP(B)->ccp_dep_fr;
    sg_fr = DepFr_sg_fr(dep_fr);
    LOCK(DepFr_lock(dep_fr));
    ans_node = DepFr_last_ans(dep_fr);
    if (ans_node != SgFr_last_answer(sg_fr)) {
      /* new unconsumed answers */
      TABLING_DEBUG_MESSAGE("          -> consuming new answer (B %x)",
                            (unsigned int)B);
      /* update last consumed answer pointer */
      if (ans_node == NULL)
        ans_node = DepFr_last_ans(dep_fr) = SgFr_first_answer(sg_fr);
      else
        ans_node = DepFr_last_ans(dep_fr) = TrNode_child(ans_node);
      UNLOCK(DepFr_lock(dep_fr));
      /* restore consumer choice point */
      H = HBREG = PROTECT_FROZEN_H(B);
      CPREG = B->cp_cp;
      ENV = B->cp_env;
      /* The following procceed invalidates the efect of `set_cut(Y, B->cp_b);' */
      /* load a new answer */
      load_answer_trie(ans_node, B);
      /* procceed */
      PREG = (yamop *)CPREG;
      PREFETCH_OP(PREG);
      Y = ENV;
      GONext();
    } else {
      /* no unconsumed answers */
      UNLOCK(DepFr_lock(dep_fr));
      if (DepFr_backchain_cp(dep_fr) == NULL) {
        /* normal backtrack */
        TABLING_DEBUG_MESSAGE("          -> no new answers -> normal backtracking (B %x)",
                              (unsigned int)B);
#ifdef YAPOR
        if (SCH_shared_cp()) {
          SCHEDULER_GET_WORK();
	}
#endif /* YAPOR */
        B = B->cp_b;
        goto fail;
      } else {
        /* chain backtrack */
#ifdef YAPOR
        or_fr_ptr start_or_fr;
        or_fr_ptr end_or_fr;
#endif /* YAPOR */
        choiceptr top_chain_cp;
        choiceptr chain_cp;
        TABLING_DEBUG_MESSAGE("         -> no new answers -> chain backtracking (B %x)",
                              (unsigned int)B);
        /* find chain choice point to backtrack */
        top_chain_cp = DepFr_backchain_cp(dep_fr);
        if (YOUNGER_CP(DepFr_leader_cp(LOCAL_top_dep_fr), top_chain_cp))
          chain_cp = DepFr_leader_cp(LOCAL_top_dep_fr);
        else 
          chain_cp = top_chain_cp;
#ifdef YAPOR
        if (EQUAL_OR_YOUNGER_CP(LOCAL_top_cp_on_stack, chain_cp)) {
          end_or_fr = chain_cp->cp_or_fr; /* chain_cp is shared */
	} else {
          end_or_fr = LOCAL_top_or_fr; /* chain_cp is private */
	}
#endif /* YAPOR */
        dep_fr = DepFr_next(dep_fr);
	while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(dep_fr)), chain_cp)) {
          if (DepFr_last_ans(dep_fr) != SgFr_last_answer(DepFr_sg_fr(dep_fr))) {
            chain_cp = NORM_CP(DepFr_cons_cp(dep_fr));
#ifdef YAPOR
            end_or_fr = DepFr_top_or_fr(dep_fr);
#endif /* YAPOR */
	    /* update backchain cp */
            LOCK(DepFr_lock(dep_fr));
#ifdef YAPOR
            if (YOUNGER_CP(DepFr_backchain_cp(dep_fr), top_chain_cp))
#endif /* YAPOR */
              DepFr_backchain_cp(dep_fr) = top_chain_cp;
            UNLOCK(DepFr_lock(dep_fr));
            break;
	  }
          dep_fr = DepFr_next(dep_fr);
	}
#ifdef TABLING_ERRORS
        if (EQUAL_OR_YOUNGER_CP(top_chain_cp, B))
          TABLING_ERROR_MESSAGE("EQUAL_OR_YOUNGER_CP(top_chain_cp, B) (answer_return)");
        else if (EQUAL_OR_YOUNGER_CP(chain_cp, B))
          TABLING_ERROR_MESSAGE("EQUAL_OR_YOUNGER_CP(chain_cp, B) (answer_return)");
#endif /* TABLING_ERRORS */
#ifdef OPTYAP_ERRORS
        if (GLOBAL_execution_mode == PARALLEL_MODE) {
          if (YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack)) {
            OPTYAP_ERROR_MESSAGE("YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack) (answer_return)");
	  } else if (YOUNGER_CP(OrFr_node(end_or_fr), LOCAL_top_cp_on_stack)) {
            OPTYAP_ERROR_MESSAGE("YOUNGER_CP(OrFr_node(end_or_fr), LOCAL_top_cp_on_stack) (answer_return)");
	  }
	}
#endif /* OPTYAP_ERRORS */
#ifdef YAPOR
        /* update shared nodes */
        start_or_fr = LOCAL_top_or_fr;
        if (start_or_fr != end_or_fr) {
          LOCAL_top_or_fr = end_or_fr;
          LOCAL_top_cp = OrFr_node(end_or_fr);
          do {
            while (YOUNGER_CP(OrFr_node(start_or_fr), OrFr_node(end_or_fr))) {
              LOCK_OR_FRAME(start_or_fr);
              BITMAP_insert(OrFr_hidden_members(start_or_fr), worker_id);
              BITMAP_delete(OrFr_members(start_or_fr), worker_id);
              UNLOCK_OR_FRAME(start_or_fr);
              start_or_fr = OrFr_next(start_or_fr);
  	    }
            while (YOUNGER_CP(OrFr_node(end_or_fr), OrFr_node(start_or_fr))) {
              LOCK_OR_FRAME(end_or_fr);
              BITMAP_insert(OrFr_members(end_or_fr), worker_id);
              BITMAP_delete(OrFr_hidden_members(end_or_fr), worker_id);
              UNLOCK_OR_FRAME(end_or_fr);
              end_or_fr = OrFr_next(end_or_fr);
	    }
	  } while (start_or_fr != end_or_fr);
        }
#endif /* YAPOR */
#ifdef OPTYAP_ERRORS
        if (GLOBAL_execution_mode == PARALLEL_MODE) {
          if (YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack)) {
            OPTYAP_ERROR_MESSAGE("YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack) (answer_return)");
	  } else {
            choiceptr aux_cp;
            aux_cp = chain_cp;
            while (aux_cp != LOCAL_top_cp) {
              if (YOUNGER_CP(LOCAL_top_cp, aux_cp)) {
                OPTYAP_ERROR_MESSAGE("LOCAL_top_cp not in branch (answer_return)");
                break;
              }
              if (EQUAL_OR_YOUNGER_CP(LOCAL_top_cp_on_stack, aux_cp)) {
                OPTYAP_ERROR_MESSAGE("shared frozen segments in branch (answer_return)");
                break;
              }
              aux_cp = aux_cp->cp_b;
            }
	  }
	}
#endif /* OPTYAP_ERRORS */
        /* restore bindings */
        restore_bindings(B->cp_tr, chain_cp->cp_tr);
        /* update registers TR and B */
        TR = TR_FZ;
        TRAIL_REF(chain_cp->cp_tr);
        B = chain_cp;
        /* backtrack to chain choice point */
        if (chain_cp == DepFr_leader_cp(LOCAL_top_dep_fr))
          goto completion;
	PREG = B->cp_ap;
        PREFETCH_OP(PREG);
	SP = SP0;
	GONext();
      }
    }
  }



  PBOp(table_new_answer, s)
    int arity = PREG->u.s.s;
    gen_cp_ptr gcp = GEN_CP(Y[E_B]);
    sg_fr_ptr sg_fr = gcp->gcp_sg_fr;
    ans_node_ptr ans_node;
    TABLING_INSTR_DEBUG_MESSAGE("table_new_answer", PREG, (CELL *)(gcp + 1) + arity);
    /* possible optimization: when the number of substitution variables 
       is zero, an answer is sufficient to perform an early completion */
    ans_node = answer_search(SgFr_answer_trie(sg_fr), (CELL *)(gcp + 1) + arity);
    LOCK(TrNode_lock(ans_node));
    if (! IS_ANSWER_LEAF_NODE(ans_node)) {
      /* new answer */
      TABLING_DEBUG_MESSAGE("          -> new answer (sg_fr %x)",
                            (unsigned int)sg_fr);
      TAG_AS_ANSWER_LEAF_NODE(ans_node);
      UNLOCK(TrNode_lock(ans_node));
      LOCK(SgFr_lock(sg_fr));
      if (SgFr_first_answer(sg_fr) == NULL)
        SgFr_first_answer(sg_fr) = ans_node;
      else 
        TrNode_child(SgFr_last_answer(sg_fr)) = ans_node;
      SgFr_last_answer(sg_fr) = ans_node;
      UNLOCK(SgFr_lock(sg_fr));
      /* deallocate and procceed */
      PREG = CPREG = (yamop *) Y[E_CP];
      PREFETCH_OP(PREG);
      SREG = Y;
      ENV = Y = (CELL *)Y[E_E];
      GONext();
    }
    TABLING_DEBUG_MESSAGE("          -> repeated answer (sg_fr %x)",
                          (unsigned int)sg_fr);
    UNLOCK(TrNode_lock(ans_node));
    goto fail;
  ENDPBOp();



  BOp(table_completion, ld)
    TABLING_DEBUG_MESSAGE("TAB INSTR -> table_completion");
#ifdef YAPOR
    if (SCH_shared_cp()) {
      SCH_new_alternative(PREG, NULL);
    } else
#endif /* YAPOR */
      B->cp_ap = NULL;   
    if (EQUAL_OR_YOUNGER_CP(B_FZ, B) && B != DepFr_leader_cp(LOCAL_top_dep_fr)) {
      /* not leader on that node */
      TABLING_DEBUG_MESSAGE("          -> not leader node (B %x sg_fr %x)", 
                            (unsigned int)B, (unsigned int)GEN_CP(B)->gcp_sg_fr);
#ifdef YAPOR
      if (SCH_shared_cp()) {
        SCHEDULER_GET_WORK();
      }
#endif /* YAPOR */
      B = B->cp_b;
      goto fail;
    }
    /* leader on that node */
    TABLING_DEBUG_MESSAGE("          -> leader node (B %x sg_fr %x)",
                          (unsigned int)B, (unsigned int)GEN_CP(B)->gcp_sg_fr);
    goto completion;
  ENDBOp();



completion:
    INIT_PREFETCH()
    dep_fr_ptr dep_fr;

    TABLING_DEBUG_MESSAGE("TAB STRAT -> completion");
    dep_fr = LOCAL_top_dep_fr;
#ifdef YAPOR
    if (SCH_shared_cp()) {
      LOCK_OR_FRAME(LOCAL_top_or_fr);
    }
#endif /* YAPOR */
    /* check for dependency frames with unconsumed answers */
    while (YOUNGER_CP(NORM_CP(DepFr_cons_cp(dep_fr)), B)) {
      if (DepFr_last_ans(dep_fr) != SgFr_last_answer(DepFr_sg_fr(dep_fr))) {    
        /* dependency frame with unconsumed answers */
        cons_cp_ptr consumer_cp;
#ifdef YAPOR
        if (SCH_shared_cp()) {
          UNLOCK_OR_FRAME(LOCAL_top_or_fr);
        }
#endif /* YAPOR */
        consumer_cp = DepFr_cons_cp(dep_fr);
        TABLING_DEBUG_MESSAGE("          -> consumer node with answers (B %x consumer %x)",
                              (unsigned int)B, (unsigned int)consumer_cp);
#ifdef OPTYAP_ERRORS
        if (GLOBAL_execution_mode == PARALLEL_MODE) {
          if (YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack)) {
            OPTYAP_ERROR_MESSAGE("YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack) (completion)");
          } else {
            choiceptr aux_cp;
            aux_cp = NORM_CP(consumer_cp);
            while (YOUNGER_CP(aux_cp, LOCAL_top_cp_on_stack))
              aux_cp = aux_cp->cp_b;
            if (aux_cp->cp_or_fr != DepFr_top_or_fr(dep_fr))
              OPTYAP_ERROR_MESSAGE("Error on DepFr_top_or_fr (completion)");
	  }
	}
#endif /* OPTYAP_ERRORS */
#ifdef YAPOR
        /* update shared nodes */
        if (YOUNGER_CP(LOCAL_top_cp_on_stack, LOCAL_top_cp)) {
          or_fr_ptr or_frame;
          or_frame = DepFr_top_or_fr(dep_fr); /* or_frame = DepFr_top_or_fr(consumer_cp->ccp_dep_fr); */
          LOCAL_top_cp = OrFr_node(or_frame);
          /* As B is a leader node, all younger nodes have B */
          /* (and consequently LOCAL_top_or_fr) in their branches */
          while (or_frame != LOCAL_top_or_fr) {
            /* put me member in all nodes between DepFr_top_or_fr and LOCAL_top_or_fr */ 
            LOCK_OR_FRAME(or_frame);
            BITMAP_insert(OrFr_members(or_frame), worker_id);
            BITMAP_delete(OrFr_hidden_members(or_frame), worker_id);
            UNLOCK_OR_FRAME(or_frame);
            or_frame = OrFr_next(or_frame);
          }
          LOCAL_top_or_fr = LOCAL_top_cp->cp_or_fr;
        }
#endif /* YAPOR */
#ifdef OPTYAP_ERRORS
        if (GLOBAL_execution_mode == PARALLEL_MODE) {
          if (YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack)) {
            OPTYAP_ERROR_MESSAGE("YOUNGER_CP(LOCAL_top_cp, LOCAL_top_cp_on_stack) (completion)");
          } else {
            choiceptr aux_cp;
            aux_cp = NORM_CP(consumer_cp);
            while (aux_cp != LOCAL_top_cp) {
              if (YOUNGER_CP(LOCAL_top_cp, aux_cp)) {
                OPTYAP_ERROR_MESSAGE("LOCAL_top_cp not in branch (completion)");
                break;
              }
              if (EQUAL_OR_YOUNGER_CP(LOCAL_top_cp_on_stack, aux_cp)) {
                OPTYAP_ERROR_MESSAGE("shared frozen segments in branch (completion)");
                break;
              }
              aux_cp = aux_cp->cp_b;
            }
          }
	}
#endif /* OPTYAP_ERRORS */
        /* update back chain */
        LOCK(DepFr_lock(dep_fr));
        if (B->cp_ap) {
#ifdef YAPOR
          if (YOUNGER_CP(DepFr_backchain_cp(dep_fr), B))
#endif /* YAPOR */
            DepFr_backchain_cp(dep_fr) = B;
	} else {
#ifdef YAPOR
          if (YOUNGER_CP(DepFr_backchain_cp(dep_fr), B->cp_b))
#endif /* YAPOR */
            DepFr_backchain_cp(dep_fr) = B->cp_b;
	}
        UNLOCK(DepFr_lock(dep_fr));
        /* rebind variables */
#ifdef TABLING_ERRORS
        if (EQUAL_OR_YOUNGER_CP(B, NORM_CP(consumer_cp)))
          TABLING_ERROR_MESSAGE("EQUAL_OR_YOUNGER_CP(B, NORM_CP(consumer_cp)) (completion)");
        if (B->cp_tr > consumer_cp->ccp_tr)
          TABLING_ERROR_MESSAGE("B->cp_tr > consumer_cp->ccp_tr (completion)");
#endif /* TABLING_ERRORS */
        rebind_variables(B->cp_tr, consumer_cp->ccp_tr);
        /* update registers TR and B */
        TR = TR_FZ;
        if (TR != consumer_cp->ccp_tr)
           TRAIL_REF(consumer_cp->ccp_tr);
        B = NORM_CP(consumer_cp);
        goto answer_return;
      }
      dep_fr = DepFr_next(dep_fr);
    }
#ifdef YAPOR
    if (SCH_shared_cp()) {
      TABLING_DEBUG_MESSAGE("          -> public node (B %x)",
                            (unsigned int)B);
      /* check for workers (present or hidden) */
      if (! BITMAP_empty(OrFr_hidden_members(LOCAL_top_or_fr)) || 
          ! BITMAP_alone(OrFr_members(LOCAL_top_or_fr), worker_id)) {
        UNLOCK_OR_FRAME(LOCAL_top_or_fr);
        suspend_completion_branch();
      } else {
        susp_fr_ptr resume_fr;
        /* check for suspension frames with unconsumed answers */


printf("***** W%d: collect_suspension_frames_with_unconsumed_answers\n", worker_id);


        collect_suspension_frames_with_unconsumed_answers();
        UNLOCK_OR_FRAME(LOCAL_top_or_fr);
        /* check for collected suspension frames */
        if ((resume_fr = LOCAL_top_collected_susp_fr) != NULL) {


printf("***** W%d: suspend completion branch (2)\n", worker_id);


          /* update LOCAL_top_collected_susp_fr */
          LOCAL_top_collected_susp_fr = SuspFr_next(LOCAL_top_collected_susp_fr);
          suspend_completion_branch();
          /* restore bindings */
          restore_bindings(B->cp_tr, SuspFr_top_cp(resume_fr)->cp_tr);
          /* resume top collected suspension branch */
          resume_suspension_frame(resume_fr);
          /* update registers TR and B */
          TR = TR_FZ;
          TRAIL_REF(LOCAL_top_cp->cp_tr);
          B = LOCAL_top_cp;
          goto completion;
        }


printf("***** W%d: PUBLIC COMPLETION\n", worker_id);


        /* complete all (public completion) */
        public_completion();
      }
      /* backtrack */
      TR = B->cp_tr;
      PREG = B->cp_ap;  /* GETWORK instruction */
      PREFETCH_OP(PREG);
      GONext();
    }
#endif /* YAPOR */
    /* complete all (private completion) */
    TABLING_DEBUG_MESSAGE("          -> private node (B %x)",
                          (unsigned int)B);
    private_completion();
    /* backtrack */
    B = B->cp_b;
    goto fail;
  }
