/*************************************************************************
*									 *
*	 YAP Prolog 	%W% %G% 					 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		TermExt.h						 *
* mods:									 *
* comments:	Extensions to standard terms for YAP			 *
* version:      $Id: Yap.h.m4,v 1.1 1997/06/02 16:32:00 vsc Exp $	 *
*************************************************************************/

#define ExtFromFunctor(f)  (((CELL)(f)-(CELL)EXT_BASE)/sizeof(FunctorEntry))

/* defined extensions */
typedef enum {
  empty_ext,		/* default op, this should never be called */
#ifdef COROUTINING
  susp_ext,              /* support for delayable goals */
  attvars_ext,           /* support for attributed variables */
#endif
#ifdef YAP_ARRAYS
  array_acc_ext,         /* support for transparent array access */
#endif /* YAP_ARRAYS */
  /* add your own extensions here */
  /* keep this one */
  last_ext
} exts;

/* make sure that these data structures are the first thing to be allocated
   in the heap when we start the system */
typedef struct {
  FunctorEntry funcs[last_ext+3];
  AtomEntry    AtNil;
  char         AtNilChars[8];
  AtomEntry    AtDot;
  char         AtDotChars[8];
} special_functors;

#define EXT_BASE  ((special_functors *)HEAP_INIT_BASE)

#if USE_OFFSETS
#define   AtomNil ((Atom)(&(((special_functors *)(NULL))->AtNil)))
#define   AtomDot ((Atom)(&(((special_functors *)(NULL))->AtDot)))
#define   TermNil MkAtomTerm(AtomNil)
#define   TermDot MkAtomTerm(AtomDot)
#else
#define   AtomNil AbsAtom(&(EXT_BASE->AtNil))
#define   AtomDot AbsAtom(&(EXT_BASE->AtDot))
#define   TermNil MkAtomTerm(AtomNil)
#define   TermDot MkAtomTerm(AtomDot)
#endif

#ifdef  YAP_ARRAYS
#define   FunctorArrayAccess ((Functor)(&(EXT_BASE->funcs[array_acc_ext])))
#endif  /* YAP_ARRAYS */
#ifdef  COROUTINING
#define   FunctorDelay       ((Functor)(&(EXT_BASE->funcs[susp_ext])))
#define   FunctorAttVar      ((Functor)(&(EXT_BASE->funcs[attvars_ext])))
#endif  /* COROUTINING */
#define   FunctorDBRef    ((Functor)(&(EXT_BASE->funcs[last_ext])))
#define   FunctorLongInt  ((Functor)(&(EXT_BASE->funcs[last_ext+1])))
#define   FunctorDouble   ((Functor)(&(EXT_BASE->funcs[last_ext+2])))

#define SF_STORE  ((special_functors *)AtomBase)

/* These are the 3 possible contexts when we call bind op */
#define SBIND	0	/* Bind it to a nonvar term */
#define SISPAIR 1	/* check whether it is a pair */
#define SISAPPL 2	/* check whether it is an appl */

typedef struct {
  /* what to do when someone tries to deref our term */
  Term (*deref_op)(Term);
  /* what to do when we want to know the state of a term */
  Term* (*sderef_op)(Term);
  /* what to do when someone tries to bind our term to someone else
     in some  predefined context */
  Term (*bind_op)(int, Term, Term);
  /* what to do if someone wants to copy our constraint */
  Term (*copy_term_op)(Term, CELL ***);
  /* op called when untrailing. It may be called from UNBIND or from GC. */
  CELL  *(*untrail_op)(int, CELL *);
  /* op called to do marking in GC */
  void (*mark_op)(CELL *);
  /* op called to do marking from TRAIL entries, if they are used */
  CELL *(*mark_from_trail_op)(CELL *);
  /* say here whether you need to be informed of engine-made bindings */
  int binding_transparent;
} ext_op;

/* Also have a look at extensions_to_present_answer(Level) if you want to
   display residuals at the end of the computation, or for debugging */

/* The ops for untrail op are: */
#define OP_UNBIND 0		/* called when you are backtracking */
#define OP_GCMARK 2             /* called when you are in the marking phase of GC */
/* At least for now, I assume that CUT will ignore any
   extensions. This means that any trail data for extensions will *not
   look like variables*!!!!!
*/

/* array with the ops for your favourite extensions */
extern ext_op attas[last_ext];


/* extern Functor FunctorDouble; */
Inline(IsFloatTerm, int, Term, t, IsApplTerm(t) && FunctorOfTerm(t) == FunctorDouble)

/* extern Functor FunctorLongInt; */
     /* I assume the tag has at most 8 bits */
Inline(MkLongIntTerm, Term, Int, i, (H[0] = (CELL)FunctorLongInt,H[1] = MkIntTerm(i/256),H[2]=MkIntTerm(i%256),H+=3,AbsAppl(H-3)))
Destructor(Term, LongIntOf, Int, t, IntOfTerm(RepAppl(t)[1])*256+IntOfTerm(RepAppl(t)[2]))
Inline(IsLongIntTerm, int, Term, t, IsApplTerm(t) && FunctorOfTerm(t) == FunctorLongInt)

/* extern Functor FunctorLongInt; */
Inline(IsLargeNumTerm, int, Term, t, IsApplTerm(t) && ((FunctorOfTerm(t) == FunctorLongInt) || (FunctorOfTerm(t) == FunctorDouble)))

Inline(IsNumTerm, int, Term, t, (IsIntTerm(t) || IsLargeNumTerm(t)))

Inline(IsAtomicTerm, Int, Term, t, IsAtomOrIntTerm(t) || IsLargeNumTerm(t))

Inline(IsExtensionFunctor, Int, Functor, f,  f <= FunctorDouble)
Inline(IsBlobFunctor, Int, Functor, f,  (f <= FunctorDouble && f >= FunctorDBRef))
Inline(IsPrimitiveTerm, Int, Term, t, (IsAtomOrIntTerm(t) || (IsApplTerm(t) && IsBlobFunctor(FunctorOfTerm(t)))))

#ifdef TERM_EXTENSIONS

Inline(IsAttachFunc, Int, Functor, f,  ((f) < FunctorDBRef))

Inline(IsAttachedTerm, Int, Term, t, (IsApplTerm(t) && IsAttachFunc(FunctorOfTerm(t))) )

#else

Inline(IsAttachFunc, Int, Functor, f, FALSE)

Inline(IsAttachedTerm, Int, Term, t, FALSE)

#endif

