/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		adtdefs.c						 *
* Last rev:								 *
* mods:									 *
* comments:	abstract machine definitions				 *
*									 *
*************************************************************************/
#ifdef SCCS
static char SccsId[] = "%W% %G%";

#endif

#define ADTDEFS_C

#include "Yap.h"
#include "Yatom.h"
#include "Heap.h"
#include <stdio.h>

static Prop
GetFunctorProp(Atom a, unsigned int arity)
{				/* look property list of atom a for kind  */
  FunctorEntry *pp = RepFunctorProp(RepAtom(a)->PropOfAE);

  while (!EndOfPAEntr(pp) &&
	 (!IsFunctorProperty(pp->KindOfPE) ||
	  pp->ArityOfFE != arity))
    pp = RepFunctorProp(pp->NextOfPE);
  return (AbsFunctorProp(pp));
}

/* vsc: We must guarantee that IsVarTerm(functor) returns true! */
Functor
MkFunctor(Atom ap, unsigned int arity)
{

  FunctorEntry *p;
  Prop p0 = GetFunctorProp(ap, arity);

  if (p0)
    return ((Functor) RepProp(p0));
  YAPEnterCriticalSection();
  p = (FunctorEntry *) AllocAtomSpace(sizeof(*p));
  p->KindOfPE = FunctorProperty;
  p->NameOfFE = ap;
  p->ArityOfFE = arity;
  p->NextOfPE = RepAtom(ap)->PropOfAE;
  RepAtom(ap)->PropOfAE = AbsProp((PropEntry *) p);
  YAPLeaveCriticalSection();
  return ((Functor) p);
}

/* vsc: We must guarantee that IsVarTerm(functor) returns true! */
void
MkFunctorWithAddress(Atom ap, unsigned int arity, FunctorEntry *p)
{

  YAPEnterCriticalSection();
  p->KindOfPE = FunctorProperty;
  p->NameOfFE = ap;
  p->ArityOfFE = arity;
  p->NextOfPE = RepAtom(ap)->PropOfAE;
  RepAtom(ap)->PropOfAE = AbsProp((PropEntry *) p);
  YAPLeaveCriticalSection();
}

#define HashFunction(c,h) { h=0; while(*c) h += *c++; h%=MaxHash; }

inline static Atom
SearchInInvisible(char *atom)
{
  AtomEntry *chain = RepAtom(*INVISIBLECHAIN);

  while (!EndOfPAEntr(chain) && strcmp(chain->StrOfAE, atom) != 0)
    chain = RepAtom(chain->NextOfAE);
  if (chain == NIL)
    return (0);
  return (AbsAtom(chain));
}

Atom
LookupAtom(char *atom)
{				/* lookup atom in atom table            */
  register CELL hash;
  register unsigned char *p;
  Atom a;
  AtomEntry *ae;

  if (LookupMode == FullLookup) {
    Atom t;

    if ((t = SearchInInvisible(atom)) != 0)
      return (t);
  }
  /* compute hash */
  p = (unsigned char *)atom;
  HashFunction(p, hash);
  a = HashChain[hash];
  /* search atom in chain */
  while (a != NIL) {
    ae = RepAtom(a);
    if (strcmp(ae->StrOfAE, atom) == 0)
      return (a);
    a = ae->NextOfAE;
  }
  /* add new atom to start of chain */
  YAPEnterCriticalSection();
  ae = (AtomEntry *) AllocAtomSpace((sizeof *ae) + strlen(atom));
  a = AbsAtom(ae);
  ae->NextOfAE = HashChain[hash];
  HashChain[hash] = a;
  ae->PropOfAE = NIL;
  if (ae->StrOfAE != atom)
    strcpy(ae->StrOfAE, atom);
  YAPLeaveCriticalSection();
  return (a);
}

void
LookupAtomWithAddress(char *atom, AtomEntry *ae)
{				/* lookup atom in atom table            */
  register CELL hash;
  register unsigned char *p;
  Atom a;

  /* compute hash */
  p = (unsigned char *)atom;
  HashFunction(p, hash);
  a = HashChain[hash];
  /* search atom in chain */
  while (a != NIL) {
    ae = RepAtom(a);
    if (strcmp(ae->StrOfAE, atom) == 0)
      return;
    a = ae->NextOfAE;
  }
  /* add new atom to start of chain */
  YAPEnterCriticalSection();
  ae->NextOfAE = HashChain[hash];
  a = AbsAtom(ae);
  HashChain[hash] = a;
  ae->PropOfAE = NIL;
  strcpy(ae->StrOfAE, atom);
  YAPLeaveCriticalSection();
}

void
ReleaseAtom(Atom atom)
{				/* Releases an atom from the hash chain */
  register Int hash;
  register unsigned char *p;
  AtomEntry *inChain;
  AtomEntry *ap = RepAtom(atom);
  char *name = ap->StrOfAE;

  /* compute hash */
  p = (unsigned char *)name;
  HashFunction(p, hash);
  /*printf("For %s we are releasing %x at %d\n",name,atom,hash);*/
  /* if the first in the chain */
  if (HashChain[hash] == atom) {
    YAPEnterCriticalSection();
    HashChain[hash] = ap->NextOfAE;
    YAPLeaveCriticalSection();
    return;
  }
  /* else */
  inChain = RepAtom(HashChain[hash]);
  while (inChain->NextOfAE != atom)
    inChain = RepAtom(inChain->NextOfAE);
  YAPEnterCriticalSection();
  inChain->NextOfAE = ap->NextOfAE;
  YAPLeaveCriticalSection();
}

Prop
GetAProp(Atom a, PropFlags kind)
{				/* look property list of atom a for kind  */
  PropEntry *pp = RepProp((RepAtom(a))->PropOfAE);

  while (!EndOfPAEntr(pp) && pp->KindOfPE != kind)
    pp = RepProp(pp->NextOfPE);
  return (AbsProp(pp));
}

Prop
GetPredProp(Atom ap, unsigned int arity)
     /* get predicate entry for ap/arity;               */
{
  Prop p0;
  PredEntry *p = RepPredProp(p0 = RepAtom(ap)->PropOfAE);

  while (p0 && (p->KindOfPE != PEProp || p->ArityOfPE != arity ||
		(p->ModuleOfPred && p->ModuleOfPred != CurrentModule)))
    p = RepPredProp(p0 = p->NextOfPE);
  return (p0);
}

/* get expression entry for at/arity;               */
Prop
GetExpProp(Atom at, unsigned int arity)
{
  Prop p0;
  ExpEntry *p = RepExpProp(p0 = RepAtom(at)->PropOfAE);

  while (p0 && (p->KindOfPE != ExpProperty || p->ArityOfEE != arity))
    p = RepExpProp(p0 = p->NextOfPE);
  return (p0);
}

Prop
PredProp(Atom ap, unsigned int arity)
     /* get predicate entry for ap/arity; create it if neccessary.              */
{
  Prop p0;
  PredEntry *p = RepPredProp(p0 = RepAtom(ap)->PropOfAE);

  while (p0 && (p->KindOfPE != 0 || p->ArityOfPE != arity ||
		(p->ModuleOfPred && p->ModuleOfPred != CurrentModule)))
    p = RepPredProp(p0 = p->NextOfPE);

  if (p0 != NIL)
    return (p0);
  YAPEnterCriticalSection();
  p = (PredEntry *) AllocAtomSpace(sizeof(*p));
#ifdef YAPOR
  INIT_LOCK(p->lock);
  INIT_LOCK(p->indexado);
#endif	/* YAPOR */
  p->KindOfPE = PEProp;
  p->ArityOfPE = arity;
  p->FirstClause = p->LastClause = NIL;
  p->PredFlags = 0L;
  p->StateOfPred = 0;
  p->OwnerFile = AtomNil;
  p->OpcodeOfPred = UNDEF_OPCODE;
  p->TrueCodeOfPred = p->CodeOfPred = (CODEADDR)(&(p->OpcodeOfPred)); 
  p->ModuleOfPred = CurrentModule;
  p->StatisticsForPred.NOfEntries = 0;
  p->StatisticsForPred.NOfHeadSuccesses = 0;
  p->StatisticsForPred.NOfRetries = 0;
  if (arity == 0)
    p->FunctorOfPred = (Functor) ap;
  else {
    p->FunctorOfPred = MkFunctor(ap, arity);
  }
  p->NextOfPE = RepAtom(ap)->PropOfAE;
#ifdef TABLING
  p->TableOfPred = NULL;
#endif /* TABLING */
  RepAtom(ap)->PropOfAE = p0 = AbsPredProp(p);
  YAPLeaveCriticalSection();
  return (p0);
}

Term
GetValue(Atom a)
{
  Prop p0 = GetAProp(a, ValProperty);

  if (p0 == 0)
    return (TermNil);
  return (RepValProp(p0)->ValueOfVE);
}

void
PutValue(Atom a, Term v)
{
  Prop p0 = GetAProp(a, ValProperty);
  ValEntry *p;

  YAPEnterCriticalSection();
  if (p0 != NIL) {
    p = RepValProp(p0);
  } else {
    p = (ValEntry *) AllocAtomSpace(sizeof(ValEntry));
    p->NextOfPE = RepAtom(a)->PropOfAE;
    RepAtom(a)->PropOfAE = AbsValProp(p);
    p->KindOfPE = ValProperty;
  }
  if (IsFloatTerm(v)) {
    /* store a float in code space, so that we can access the property */
    union {
      Float f;
      CELL ar[sizeof(Float) / sizeof(CELL)];
    } un;
    CELL *pt, *iptr;
    unsigned int i;

    un.f = FloatOfTerm(v);
    if (p0 != NIL && IsApplTerm(p->ValueOfVE))
      pt = RepAppl(p->ValueOfVE);
    else {
      pt = (CELL *) AllocAtomSpace(sizeof(CELL)*(1 + 2*sizeof(Float)/sizeof(CELL)));
    }

    pt[0] = (CELL)FunctorDouble;
    iptr = pt+1;
    for (i = 0; i < sizeof(Float) / sizeof(CELL); i++) {
      *iptr++ = MkIntTerm(un.ar[i]/256);
      *iptr++ = MkIntTerm(un.ar[i]%256);
    }
    p->ValueOfVE = AbsAppl(pt);
  } else if (IsLongIntTerm(v)) {
    CELL *pt;
    Int val = LongIntOfTerm(v);
    if (p0 != NIL && IsApplTerm(p->ValueOfVE)) { 
      pt = RepAppl(p->ValueOfVE);
    } else {
      pt = (CELL *) AllocAtomSpace(3 * sizeof(CELL));
    }
    pt[0] = (CELL)FunctorLongInt;
    pt[1] = MkIntTerm(val/256);
    pt[2] = MkIntTerm(val%256);
    p->ValueOfVE = AbsAppl(pt);
  } else {
    if (p0 != NIL && IsApplTerm(p->ValueOfVE)) {
      /* recover space */
      FreeCodeSpace((char *) (RepAppl(p->ValueOfVE)));
    }
    p->ValueOfVE = v;
  }
  YAPLeaveCriticalSection();
}

Float
FloatOfTerm(Term t)
{
#if 0 && defined(__GNUC__) && SIZEOF_DOUBLE == 8 && SIZEOF_INT_P == 4
  CELL *ptr = RepAppl(t)+1;
  long long x = ((long long)IntOfTerm(ptr[0])<<40) |
    ((long long)IntOfTerm(ptr[1])<<32) |
    ((long long)IntOfTerm(ptr[2])<<8) |
    ((long long)IntOfTerm(ptr[3]));
  return((Float)x);
#else
  CELL *iptr = RepAppl(t) + 1;
  unsigned int i;
  union {
    Float f;
    CELL ar[sizeof(Float) / sizeof(CELL)];
  } un;

  for (i = 0; i < sizeof(Float) / sizeof(CELL); i++) {
    un.ar[i] = IntOfTerm(iptr[0])*256+IntOfTerm(iptr[1]);
    iptr+=2;
  }
  /* avoid a possible SEGV because we are returning a field from an union */
  return (un.f);
#endif
}

Term
MkFloatTerm(Float flt)
{
#if 0 && defined(__GNUC__) && SIZEOF_DOUBLE == 8 && SIZEOF_INT_P == 4
  CELL *ptr = H;
  long long x = (long long)flt;
  H[4] = MkIntTerm((Int)x&0xff);
  H[3] = MkIntTerm((Int)(x>>8)&0xffffff);
  H[2] = MkIntTerm((Int)(x>>32)&0xff);
  H[1] = MkIntTerm((Int)(x>>40)&0xffffff);
  H[0] = (CELL)FunctorDouble;
  H += 5;
  return(AbsAppl(ptr));
#else
  register CELL *iptr = H;
  union {
    Float f;
    CELL ar[sizeof(Float) / sizeof(CELL)];
  } un;
  CELL out = AbsAppl(iptr);
  unsigned int i;

  un.f = flt;
  *iptr++ = (CELL) FunctorDouble;
  for (i = 0; i < sizeof(Float) / sizeof(CELL); i++) {
    *iptr++ = MkIntTerm(un.ar[i]/256);
    *iptr++ = MkIntTerm(un.ar[i]%256);
  }
  H = iptr;
  return (out);
#endif

}

Term
StringToList(char *s)
{
  register Term t;
  register unsigned char *cp = (unsigned char *)s + strlen(s);

  t = MkAtomTerm(AtomNil);
  while (cp > (unsigned char *)s) {
    t = MkPairTerm(MkIntTerm(*--cp), t);
  }
  return (t);
}

Term
StringToListOfAtoms(char *s)
{
  register Term t;
  char so[2];
  register unsigned char *cp = (unsigned char *)s + strlen(s);

  so[1] = '\0';
  t = MkAtomTerm(AtomNil);
  while (cp > (unsigned char *)s) {
    so[0] = *--cp;
    t = MkPairTerm(MkAtomTerm(LookupAtom(so)), t);
  }
  return (t);
}

Term
ArrayToList(register Term *tp, int nof)
{
  register Term *pt = tp + nof;
  register Term t;

  t = MkAtomTerm(AtomNil);
  while (pt > tp) {
    Term tm = *--pt;
#if SBA
    if (tm == 0)
      t = MkPairTerm((CELL)pt, t);
    else
#endif
      t = MkPairTerm(tm, t);
  }
  return (t);
}

int
GetName(char *s, Term t)
{
  register Term Head;
  register Int i;

  if (IsVarTerm(t) || !IsPairTerm(t))
    return (FALSE);
  while (IsPairTerm(t)) {
    Head = HeadOfTerm(t);
    if (!IsNumTerm(Head))
      return (FALSE);
    i = IntOfTerm(Head);
    if (i < 0 || i > 255)
      return (FALSE);
    *s++ = i;
    t = TailOfTerm(t);
  }
  *s = '\0';
  return (TRUE);
}

#ifdef SFUNC

Term
MkSFTerm(Functor f, int n, Term *a, empty_value)
{
  Term t, p = AbsAppl(H);
  int i;

  *H++ = f;
  RESET_VARIABLE(H);
  ++H;
  for (i = 1; i <= n; ++i) {
    t = Derefa(a++);
    if (t != empty_value) {
      *H++ = i;
      *H++ = t;
    }
  }
  *H++ = 0;
  return (p);
}

CELL *
ArgsOfSFTerm(Term t)
{
  CELL *p = RepAppl(t) + 1;

  while (*p != (CELL) p)
    p = CellPtr(*p) + 1;
  return (p + 1);
}

#endif
