/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		exec.c							 *
* Last rev:	8/2/88							 *
* mods:									 *
* comments:	Execute Prolog code					 *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "@(#)cdmgr.c	1.1 05/02/98";
#endif

#include "Yap.h"
#include "clause.h"
#include "yapio.h"
#include "eval.h"
#include "tracer.h"
#ifdef YAPOR
#include "or.macros.h"
#endif	/* YAPOR */


STATIC_PROTO(Int  CallProlog, (Prop, unsigned int, Int));
STATIC_PROTO(Int  p_save_cp, (void));
STATIC_PROTO(Int  p_execute, (void));
STATIC_PROTO(Int  p_execute0, (void));
STATIC_PROTO(Int  p_at_execute, (void));

static Int 
CallProlog(Prop pe, unsigned int arity, Int position)
{
  CELL            flags;
  PredEntry      *pen = RepPredProp(pe);

  if (pe == NIL)
    flags = 0;
  else
    flags = pen->PredFlags;
  if (flags & UserCPredFlag) {
    Int(*p) (void) = (Int(*) (void)) pen->CodeOfPred;
    Int out;
    
    save_machine_regs();
    out = c_execute(p);
    restore_machine_regs();
    return(out);
  } else if (flags & CPredFlag) {
    Int(*p) (void) = (Int(*) (void)) pen->CodeOfPred;
    return ((*p) ());
  } else if (flags & BasicPredFlag) {
    if (pen->OpcodeOfPred != UNDEF_OPCODE) {
      Int(*p) (void) = (Int(*) (void)) pen->CodeOfPred;
      return (((*p) ()) != FALSE);
    }
  }
  if ((flags & (CompiledPredFlag | DynamicPredFlag)) ||
      pen->OpcodeOfPred == UNDEF_OPCODE) {
    if (pen->OpcodeOfPred == INDEX_OPCODE) {
      IPred((CODEADDR)pen);
    }
#ifdef DEPTH_LIMIT
      if (DEPTH <= MkIntTerm(1)) {/* I assume Module==0 is primitives */
	if (pen->ModuleOfPred) {
	  if (DEPTH == MkIntTerm(0))
	    return(FALSE);
	  else DEPTH = RESET_DEPTH();
	}
      } else if (pen->ModuleOfPred)
	DEPTH -= MkIntConstant(2);
#endif	/* DEPTH_LIMIT */
#ifdef LOW_LEVEL_TRACER
      if (do_low_level_trace)
	low_level_trace(enter_pred,pen,XREGS+1);
#endif	/* LOW_LEVEL_TRACE */
    if (position == -1 ||
	pen->OpcodeOfPred == UNDEF_OPCODE) {
      if (pen->PredFlags & ProfiledPredFlag)
	pen->StatisticsForPred.NOfEntries++;
      ENV = YENV;
      YENV = ASP;
      YENV[E_CB] = (CELL) B;
      CP = P;
      P = (yamop *)(pen->CodeOfPred);
      return (Unsigned(&(pen->StateOfPred)));
    } else {
      CODEADDR        q;
      ENV = YENV;
      YENV = ASP;
      YENV[E_CB] = (CELL)(B->cp_b);
      CP = P;
      q = pen->FirstClause;
      if (pen->PredFlags & ProfiledPredFlag) {
	if (position == 1)
	  pen->StatisticsForPred.NOfEntries++;
	else
	  pen->StatisticsForPred.NOfRetries++;
      }
      if (flags & DynamicPredFlag) {
	CLAUSECODE->arity = pen->ArityOfPE;
	CLAUSECODE->func = pen->FunctorOfPred;
	while (position > 1) {
	  while (ClauseCodeToClause(q)->ClFlags & ErasedMask)
	    q = NextClause(q);
	  position--;
	  q = NextClause(q);
	}
	while (ClauseCodeToClause(q)->ClFlags & ErasedMask)
	  q = NextClause(q);
	if (!(ClauseCodeToClause(q)->ClFlags & InUseMask)) {
	  OPREG     *opp = &(ClauseCodeToClause(q)->ClFlags);
	  TRAIL_REF(opp);
	  *opp |= InUseMask;
	}
	CLAUSECODE->clause = (CODEADDR)NEXTOP((yamop *)(q),ld);
	P = (yamop *)CLAUSECODE->clause;
	return((CELL)(&(CLAUSECODE->clause)));
      } else {
	for (; position > 1; position--)
	  q = NextClause(q);
	P = NEXTOP((yamop *)(q),ld);
	return (Unsigned(&(pen->StateOfPred)));
      }
    }
  }
  return (FALSE);
}

static Int
p_save_cp(void)
{
  Term t = Deref(ARG1);
  if (!IsVarTerm(t)) return(FALSE);
  Bind((CELL *)t,MkIntTerm(LCL0-(CELL *)B));
  return(TRUE);
}

static int SpecialCallFunctor(Functor f) {
  return(f == FunctorComma || f == FunctorOr || f == FunctorArrow ||
	 f == FunctorVBar || f == FunctorNot || f == FunctorAltNot);
}

static Int
p_execute(void)
{				/* '$execute'(Goal)	 */
  Term            t = Deref(ARG1);
  unsigned int    arity;
  Prop            pe;
  Atom            a;

 restart_exec:
  if (IsVarTerm(t)) {
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
      a = AtomMetaCall;
      ARG2 = TermNil;
      arity = 2;
    } else {
      Error(INSTANTIATION_ERROR,t,"call/1");    
      return(FALSE);
    }
  } else if (IsNumTerm(t) || IsRefTerm(t)) {
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
      a = AtomMetaCall;
      ARG2 = TermNil;
      arity = 2;
    } else {
      Error(TYPE_ERROR_CALLABLE,t,"call/1");    
      return(FALSE);
    }
  } else if (yap_flags[SPY_CREEP_FLAG]) {
    a = NameOfFunctor(FunctorSpiedMetaCall);
    arity = 1;
  } else if (PredGoalExpansion->OpcodeOfPred != UNDEF_OPCODE) {
    a = AtomMetaCall;
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
      ARG2 = TermNil;
      arity = 2;
    } else
      arity = 1;
  } else if (IsAtomTerm(t)) {
    a = AtomOfTerm(t);
    arity = 0;
    if (a == AtomTrue || a == AtomOtherwise)
      return(TRUE);
    else if (a == AtomFail || a == AtomFalse)
      return(FALSE);
    else if (a == AtomRepeat || a == AtomCut) {
      a = AtomMetaCall;
      if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
	ARG2 = TermNil;
	arity = 2;
      } else
	arity = 1;
    }
  } else if (IsApplTerm(t)) {
    register Functor f = FunctorOfTerm(t);
    register unsigned int    i;
    register CELL *pt;

    if (IsBlobFunctor(f))
      return(FALSE);
    
    if (SpecialCallFunctor(f)) {
      a = AtomMetaCall;
      if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
	ARG2 = TermNil;
	arity = 2;
      } else
	arity = 1;
    } else if (f == FunctorModule) {
      Term mod = ArgOfTerm(1, t);
      if (mod == ModuleName[CurrentModule]) {
	/* we can skip this operation */
	t = ArgOfTerm(2, t);
	goto restart_exec;
      } else {
	a = AtomMetaCall;
	if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
	  ARG2 = TermNil;
	  arity = 2;
	} else
	  arity = 1;
      }
    } else {
      arity = ArityOfFunctor(f);
      a = NameOfFunctor(f);
      /* I cannot use the standard macro here because
	 otherwise I would dereference the argument and
	 might skip a svar */
      pt = RepAppl(t)+1;
      for (i = 1; i <= arity; ++i) {
#if SBA
	Term d0 = *pt++;
	if (d0 == 0)
	  XREGS[i] = (CELL)(pt-1);
	else
	  XREGS[i] = d0;
#else
	XREGS[i] = *pt++;
#endif
      }
    }
  } else if (IsPairTerm(t)) {
    a = AtomMetaCall;
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
      ARG2 = TermNil;
      arity = 2;
    } else
      arity = 1;
  } else
    return (FALSE);	/* for the moment */
  /*	N = arity; */
  /* call may not define new system predicates!! */
  if (CurrentModule)
    pe = PredProp(a, arity);
  else {
    pe = GetPredProp(a, arity);
    if (pe == NIL) {
      ARG1 = t;
      if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
	arity = 2;
	ARG2 = TermNil;
      } else
	arity = 1;
      pe = PredProp(AtomMetaCall, arity);
    }
  }
  if (RepPredProp(pe)->PredFlags & MetaPredFlag) {
    /* reset argument */
    ARG1 = t;
    if (yap_flags[LANGUAGE_MODE_FLAG] == 1) {
      ARG2 = TermNil;
      arity = 2;
    } else
      arity = 1;
    pe = PredProp(AtomMetaCall, arity);
  }
  return (CallProlog(pe, arity, (Int) (-1)));
}

static Int
p_execute_within(void)
{				/* '$execute_within'(Goal,CutPt,OrigGoal)	 */
  Term            t = Deref(ARG1);
  unsigned int    arity;
  Prop            pe;
  Atom            a;
  Term cut_pt = ARG2, OrigGoal = ARG3;

 restart_exec:
  if (IsVarTerm(t)) {
    Error(INSTANTIATION_ERROR,OrigGoal,"call/1");    
    return(FALSE);
  } else if (IsNumTerm(t) || IsRefTerm(t)) {
    Error(TYPE_ERROR_CALLABLE,OrigGoal,"call/1");    
    return(FALSE);
  }
  if (yap_flags[SPY_CREEP_FLAG]) {
    Term ts[3];
    ts[0] = t;
    ts[1] = cut_pt;
    ts[2] = OrigGoal;
    t = MkApplTerm(FunctorSpiedMetaCall, 1, ts);
  }
  if (PredGoalExpansion->OpcodeOfPred != UNDEF_OPCODE) {
    a = AtomMetaCall;
    arity = 3;
  } else if (IsAtomTerm(t)) {
    a = AtomOfTerm(t);
    arity = 0;
    if (a == AtomTrue || a == AtomOtherwise)
      return(TRUE);
    else if (a == AtomFail || a == AtomFalse)
      return(FALSE);
    else if (a == AtomRepeat || a == AtomCut) {
      a = AtomMetaCall;
      arity = 3;
    }
  } else if (IsApplTerm(t)) {
    register Functor f = FunctorOfTerm(t);
    register unsigned int    i;
    register CELL *pt;

    if (IsBlobFunctor(f))
      return(FALSE);
    
    if (SpecialCallFunctor(f)) {
      a = AtomMetaCall;
      arity = 3;
    } else if (f == FunctorModule) {
      Term mod = ArgOfTerm(1, t);
      if (mod == ModuleName[CurrentModule]) {
	/* we can skip this operation */
	t = ArgOfTerm(2, t);
	goto restart_exec;
      } else {
	a = AtomMetaCall;
	arity = 3;
      }
    } else {
      arity = ArityOfFunctor(f);
      a = NameOfFunctor(f);
      /* I cannot use the standard macro here because
	 otherwise I would dereference the argument and
	 might skip a svar */
      pt = RepAppl(t)+1;
      for (i = 1; i <= arity; ++i) {
#if SBA
	Term d0 = *pt++;
	if (d0 == 0)
	  XREGS[i] = (CELL)(pt-1);
	else
	  XREGS[i] = d0;
#else
	XREGS[i] = *pt++;
#endif
      }
    }
  } else if (IsPairTerm(t)) {
    a = AtomMetaCall;
    arity = 3;
  } else
    return (FALSE);	/* for the moment */
  /*	N = arity; */
  /* call may not define new system predicates!! */
  if (CurrentModule)
    pe = PredProp(a, arity);
  else {
    pe = GetPredProp(a, arity);
    if (pe == NIL) {
      ARG1 = t;
      pe = PredProp(AtomMetaCall, 1);
      arity = 3;
    }
  }
  if (RepPredProp(pe)->PredFlags & MetaPredFlag) {
    /* reset argument */
    ARG1 = t;
    pe = PredProp(AtomMetaCall, 1);
    arity = 3;
  }
  return (CallProlog(pe, arity, (Int) (-1)));
}

static Int
p_execute0(void)
{				/* '$execute'(Goal)	 */
  Term            t = Deref(ARG1);
  unsigned int    arity;
  Prop            pe;
  Atom            a;

  if (IsAtomTerm(t))
    arity = 0, a = AtomOfTerm(t);
  else if (IsApplTerm(t)) {
    register Functor f = FunctorOfTerm(t);
    register unsigned int    i;
    register CELL *pt;

    if (IsExtensionFunctor(f))
      return(FALSE);
    arity = ArityOfFunctor(f);
    a = NameOfFunctor(f);
    /* I cannot use the standard macro here because
       otherwise I would dereference the argument and
       might skip a svar */
    pt = RepAppl(t)+1;
    for (i = 1; i <= arity; ++i)
#if SBA
      {
	Term d0 = *pt++;
	if (d0 == 0)
	  XREGS[i] = (CELL)(pt-1);
	else
	  XREGS[i] = d0;
      }
#else
      XREGS[i] = *pt++;
#endif
  } else
    return (FALSE);	/* for the moment */
  /*	N = arity; */
  /* call may not define new system predicates!! */
  if (CurrentModule)
    pe = PredProp(a, arity);
  else {
    pe = GetPredProp(a, arity);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, arity, (Int) (-1)));
}

static Int
p_execute_0(void)
{				/* '$execute_0'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  if (CurrentModule)
    pe = PredProp(a, 0);
  else {
    pe = GetPredProp(a, 0);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 0, (Int) (-1)));

}

static Int
p_execute_1(void)
{				/* '$execute_0'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  if (CurrentModule)
    pe = PredProp(a, 1);
  else {
    pe = GetPredProp(a, 1);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 1, (Int) (-1)));
}

static Int
p_execute_2(void)
{				/* '$execute_2'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  if (CurrentModule)
    pe = PredProp(a, 2);
  else {
    pe = GetPredProp(a, 2);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 2, (Int) (-1)));
}

static Int
p_execute_3(void)
{				/* '$execute_3'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  ARG3 = ARG4;
  if (CurrentModule)
    pe = PredProp(a, 3);
  else {
    pe = GetPredProp(a, 3);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 3, (Int) (-1)));
}

static Int
p_execute_4(void)
{				/* '$execute_4'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  ARG3 = ARG4;
  ARG4 = ARG5;
  if (CurrentModule)
    pe = PredProp(a, 4);
  else {
    pe = GetPredProp(a, 4);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 4, (Int) (-1)));
}

static Int
p_execute_5(void)
{				/* '$execute_5'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  ARG3 = ARG4;
  ARG4 = ARG5;
  ARG5 = ARG6;
  if (CurrentModule)
    pe = PredProp(a, 5);
  else {
    pe = GetPredProp(a, 5);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 5, (Int) (-1)));
}

static Int
p_execute_6(void)
{				/* '$execute_6'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  ARG3 = ARG4;
  ARG4 = ARG5;
  ARG5 = ARG6;
  ARG6 = ARG7;
  if (CurrentModule)
    pe = PredProp(a, 6);
  else {
    pe = GetPredProp(a, 6);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 6, (Int) (-1)));
}

static Int
p_execute_7(void)
{				/* '$execute_7'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  ARG3 = ARG4;
  ARG4 = ARG5;
  ARG5 = ARG6;
  ARG6 = ARG7;
  ARG7 = ARG8;
  if (CurrentModule)
    pe = PredProp(a, 7);
  else {
    pe = GetPredProp(a, 7);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 7, (Int) (-1)));
}

static Int
p_execute_8(void)
{				/* '$execute_8'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  ARG3 = ARG4;
  ARG4 = ARG5;
  ARG5 = ARG6;
  ARG6 = ARG7;
  ARG7 = ARG8;
  ARG8 = ARG9;
  if (CurrentModule)
    pe = PredProp(a, 8);
  else {
    pe = GetPredProp(a, 8);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 8, (Int) (-1)));
}

static Int
p_execute_9(void)
{				/* '$execute_9'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  ARG3 = ARG4;
  ARG4 = ARG5;
  ARG5 = ARG6;
  ARG6 = ARG7;
  ARG7 = ARG8;
  ARG8 = ARG9;
  ARG9 = ARG10;
  if (CurrentModule)
    pe = PredProp(a, 9);
  else {
    pe = GetPredProp(a, 9);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 9, (Int) (-1)));
}

static Int
p_execute_10(void)
{				/* '$execute_10'(Goal)	 */
  Term            t = Deref(ARG1);
  Prop            pe;
  Atom            a;

  a = AtomOfTerm(t);
  ARG1 = ARG2;
  ARG2 = ARG3;
  ARG3 = ARG4;
  ARG4 = ARG5;
  ARG5 = ARG6;
  ARG6 = ARG7;
  ARG7 = ARG8;
  ARG8 = ARG9;
  ARG9 = ARG10;
  ARG10 = ARG11;
  if (CurrentModule)
    pe = PredProp(a, 10);
  else {
    pe = GetPredProp(a, 10);
    if (pe == NIL)
      return(FALSE);
  }
  return (CallProlog(pe, 10, (Int) (-1)));
}

#ifdef DEPTH_LIMIT
static Int
p_execute_depth_limit(void) {
  Term d = Deref(ARG2);
  if (IsVarTerm(d)) {
    Error(INSTANTIATION_ERROR,d,"depth_bound_call/2");    
  } else if (!IsIntTerm(d)) {
    Error(TYPE_ERROR_INTEGER, d, "depth_bound_call/2");
    return(FALSE);
  }
  DEPTH = MkIntTerm(IntOfTerm(d)*2);
  return(p_execute());
}
#endif

static Int
p_pred_goal_expansion_on(void) {
  /* a goal needs expansion if we have goal_expansion defined or
     if the goal is a meta-call */
  return (PredGoalExpansion->OpcodeOfPred != UNDEF_OPCODE);
}

static Int 
p_at_execute(void)
{				/* '$execute'(Goal,ClauseNumber) */
  Term            t = Deref(ARG1), t2 = Deref(ARG2);
  unsigned int arity;
  Prop            pe;
  Atom            a;

  if (IsAtomTerm(t))
    arity = 0, a = AtomOfTerm(t);
  else if (IsApplTerm(t)) {
    register Functor f = FunctorOfTerm(t);
    register unsigned int    i;
    register CELL *pt;

    if (IsBlobFunctor(f))
      return(FALSE);
    arity = ArityOfFunctor(f);
    a = NameOfFunctor(f);
    /* I cannot use the standard macro here because
       otherwise I would dereference the argument and
       might skip a svar */
    pt = RepAppl(t)+1;
    for (i = 1; i <= arity; ++i)
#if SBA
      {
	Term d0 = *pt++;
	if (d0 == 0)
	  XREGS[i] = (CELL)(pt-1);
	else
	  XREGS[i] = d0;
      }
#else
      XREGS[i] = *pt++;
#endif
  } else
    return (FALSE);	/* for the moment */
  if (IsVarTerm(t2) || !IsIntTerm(t2))
    return (FALSE);
  /* N = arity; */
  /* call may not define new system predicates!! */
  if (CurrentModule)
    pe = PredProp(a, arity);
  else
    pe = GetPredProp(a, arity);
  return (CallProlog(pe, arity, IntOfTerm(t2)));
}

int
exec_absmi(int top)
{
  int lval;
  if (top && (lval = sigsetjmp (RestartEnv, 1)) != 0) {
    if (lval == 1) { /* restart */
      int depth;
      /* otherwise, SetDBForThrow will fail entering critical mode */
      PrologMode = UserMode;
      /* find out where to cut to */
      depth = SetDBForThrow(MkAtomTerm(LookupAtom("abort")));
      if (depth == 0) {
	Error(SYSTEM_ERROR, TermNil, "database entry for throw corrupted");
      }
      /* make the abstract machine jump where we want them to jump to */

#if defined(SBA) && defined(FROZEN_REGS)
      B = (choiceptr)depth;
#else
      B = (choiceptr)(LCL0-depth);
#endif
      yap_flags[SPY_CREEP_FLAG] = 0;
      CreepFlag = MinStackGap*(stack_overflows+1);
#if defined(__GNUC__) && defined(hppa)
     /* siglongjmp resets the TR hardware register */
      restore_TR();
#endif
    }
    if (lval == 2) { /* arithmetic exception */
      /* must be done here, otherwise siglongjmp will clobber all the registers */
      Error(YAP_matherror,current_eval_term,NULL);
      /* reset the registers so that we don't have trash in abstract machine */
      eval_flt = 0.0;
      eval_int = 0;
      set_fpu_exceptions(yap_flags[LANGUAGE_MODE_FLAG] == 1);
    }
    P = (yamop *)FAILCODE;
  }
  return(absmi(0));
}

static int
do_goal(CODEADDR CodeAdr, int arity, CELL *pt, int args_to_save, int top)
{
  choiceptr saved_b = B;

  /* create an initial pseudo environment so that when garbage
     collection is going up in the environment chain it doesn't get
     confused */
  YENV = ASP;
  YENV[E_CP] = (CELL)P;
  YENV[E_CB] = (CELL)B;
  YENV[E_E]  = (CELL)ENV;
#ifdef  DEPTH_LIMIT
  YENV[E_DEPTH] = DEPTH;
#endif
  ENV = YENV;
  ASP -= EnvSizeInCells;
  /* and now create a pseudo choicepoint for much the same reasons */
  CP = (yamop *) YESCODE;
  /* this is an hack to save the arguments */
  {
    CELL *pt1 = XREGS+1+args_to_save;

    while (pt1 > XREGS+1)
      *--ASP = *--pt1;
  }
  *--ASP = MkIntTerm(args_to_save+1);
  { 
    int i;
    for (i = 0; i < arity; i++) {
      XREGS[i+1] = *pt++;
    }
  }
  B = (choiceptr)ASP;
  B--;
  B->cp_h     = H;
  B->cp_tr    = TR;
  B->cp_cp    = CP;
  B->cp_ap    = (yamop *) NOCODE;
  B->cp_env   = ENV;
  B->cp_b     = saved_b;
#ifdef DEPTH_LIMIT
  B->cp_depth = DEPTH;
#endif /* DEPTH_LIMIT */
  if (top) {
#if COROUTINING
    WokenGoals = NewTimedVar(TermNil);
#ifdef MULTI_ASSIGNMENT_VARIABLES
    MutableList = NewTimedVar(TermNil);
    AttsMutableList = NewTimedVar(TermNil);
#endif
#endif
  }
  YENV = ASP = (CELL *)B;
  HB = H;
  YENV[E_CB] = Unsigned (B);
  P = (yamop *) CodeAdr;
  S = CellPtr (&(RepPredProp (PredProp (AtomCall, 1))->StateOfPred));	/* A1 mishaps */
  TopB = B;
  return(exec_absmi(top));
}


Int
execute_goal(Term t, int nargs)
{
  Int             out;
  CODEADDR        CodeAdr;
  choiceptr  saved_b;
  yamop *saved_p, *saved_cp;
  int arity;
  Atom a;
  Prop pe;
  PredEntry *ppe;
  CELL *pt;
  /* preserve the current restart environment */
  /* visualc*/
  /* just keep the difference because of possible garbage collections */
  Int MyOldTopB = LCL0-(CELL *)TopB;
  Int MyOldDelayedB = LCL0-(CELL *)DelayedB;
  choiceptr OldTopB, OldDelayedB;
  DelayedB = NULL;
  /* forget we have a DelayedB to do, this is not a problem for the embedded computation to solve */

  saved_b = B;
  saved_p = P;
  saved_cp = CP;
  
  if (IsAtomTerm(t)) {
    arity = 0;
    a = AtomOfTerm(t);
    pt = NULL;
  } else if (IsApplTerm(t)) {
    Functor f = FunctorOfTerm(t);

    if (IsBlobFunctor(f)) {
      Error(TYPE_ERROR_CALLABLE,t,"call/1");
      return(FALSE);
    }
    arity = ArityOfFunctor(f);
    a = NameOfFunctor(f);
    /* I cannot use the standard macro here because
       otherwise I would dereference the argument and
       might skip a svar */
    pt = RepAppl(t)+1;
  } else {
    Error(TYPE_ERROR_CALLABLE,t,"call/1");
    return(FALSE);
  }
  pe = GetPredProp(a, arity);
  ppe = RepPredProp(pe);
  if (pe == NIL ||
      ppe->OpcodeOfPred == UNDEF_OPCODE ||
      ppe->PredFlags & (UserCPredFlag|CPredFlag|BasicPredFlag) )
    {
      pe = PredProp(AtomMetaCall, 1);
      ppe = RepPredProp(pe);
      arity = 1;
      pt = & t;
    }
  CodeAdr = RepPredProp (PredProp (a, arity))->CodeOfPred;
  out = do_goal(CodeAdr, arity, pt, nargs, FALSE);

  if (out == 1) {
    choiceptr old_B;
    /* we succeeded, let's prune */
    /* restore the old environment */
    /* get to previous environment */
#ifdef YAPOR
    if (! CUT_shared((choiceptr)(ENV[E_CB])))
      B = LOCAL_top_cp;
#else
    B = (choiceptr)(ENV[E_CB]);
#endif /* YAPOR */
    /* find out where we have the old arguments */
    old_B = ((choiceptr)(ENV-(EnvSizeInCells+nargs+1)))-1;
    {
      int i;
      for (i=1; i<= nargs; i++) {
#ifdef __GNUC__
	XREGS[i] = old_B->cp_args[i];
#else
	XREGS[i] = old_B->cp_last.cp_args[i];
#endif
      }
    }
    CP   = saved_cp;
    P    = saved_p;
    ASP  = ENV;
    ENV  = (CELL *)(ENV[E_E]);
#ifdef DEPTH_LIMIT
    DEPTH= ENV[E_DEPTH];
#endif
    /* we have failed, and usually we would backtrack to this B,
       trouble is, we may also have a delayed cut to do */
    OldTopB = (choiceptr)(LCL0-MyOldTopB);
    OldDelayedB = (choiceptr)(LCL0-MyOldDelayedB);
    if (DelayedB != NULL) {
      if (YOUNGER_CP(DelayedB,B)) {
	/* we have a delayed cut to do */
	if (YOUNGER_CP(OldTopB,DelayedB)) {
	  /* and this delayed cut is to before the c-code that actually called us */
	  B = OldTopB;
	  /* did we have a cut which was cutting more than our current cut? */
	  if (OldDelayedB != NULL && YOUNGER_CP(DelayedB,OldDelayedB))
	    DelayedB = OldDelayedB;
	} else {
	  /* just cut back to where we should cut */
	  B = DelayedB;
	  DelayedB = OldDelayedB;
	}
      }
    } else {
      DelayedB = OldDelayedB;
    }
    TopB = OldTopB;
    HB   = B->cp_h;
    YENV = ENV;
    return(TRUE);
  } else if (out == 0) {
    ASP  = B->cp_env;
    {
      int i;
      for (i=1; i<= nargs; i++) {
#ifdef __GNUC__
	XREGS[i] = B->cp_args[i];
#else
	XREGS[i] = B->cp_last.cp_args[i];
#endif
      }
    }
    P    = saved_p;
    CP   = saved_cp;
    H    = B->cp_h;
#ifdef DEPTH_LIMIT
    DEPTH= B->cp_depth;
#endif
    YENV = ENV  = (CELL *)((B->cp_env)[E_E]);
    B    = B->cp_b;
    /* we have failed, and usually we would backtrack to this B,
       trouble is, we may also have a delayed cut to do */
    OldTopB = (choiceptr)(LCL0-MyOldTopB);
    OldDelayedB = (choiceptr)(LCL0-MyOldDelayedB);
    if (DelayedB != NULL) {
      if (YOUNGER_CP(DelayedB,B)) {
	/* we have a delayed cut to do */
	if (YOUNGER_CP(OldTopB,DelayedB)) {
	  /* and this delayed cut is to before the c-code that actually called us */
	  B = OldTopB;
	  /* did we have a cut which was cutting more than our current cut? */
	  if (OldDelayedB != NULL && YOUNGER_CP(DelayedB,OldDelayedB))
	    DelayedB = OldDelayedB;
	} else {
	  /* just cut back to where we should cut */
	  B = DelayedB;
	  DelayedB = OldDelayedB;
	}
      }
    } else {
      DelayedB = OldDelayedB;
    }
    TopB = OldTopB;
    HB   = B->cp_h;
    return(FALSE);
  } else {
    Error(SYSTEM_ERROR,TermNil,"emulator crashed");
    return(FALSE);
  }
}

int
YapRunTopGoal(Term t)
{
  CODEADDR        CodeAdr;
  int arity;
  Atom a;
  Prop pe;
  PredEntry *ppe;
  CELL *pt;

  if (IsAtomTerm(t)) {
    arity = 0;
    a = AtomOfTerm(t);
    pt = NULL;
  } else if (IsApplTerm(t)) {
    Functor f = FunctorOfTerm(t);

    if (IsBlobFunctor(f)) {
      Error(TYPE_ERROR_CALLABLE,t,"call/1");
      return(FALSE);
    }
    arity = ArityOfFunctor(f);
    a = NameOfFunctor(f);
    /* I cannot use the standard macro here because
       otherwise I would dereference the argument and
       might skip a svar */
    pt = RepAppl(t)+1;
  } else {
    Error(TYPE_ERROR_CALLABLE,t,"call/1");
    return(FALSE);
  }
  pe = GetPredProp(a, arity);
  ppe = RepPredProp(pe);
  if (pe == NIL ||
      ppe->OpcodeOfPred == UNDEF_OPCODE ||
      ppe->PredFlags & (UserCPredFlag|CPredFlag|BasicPredFlag) )
    {
      /* we must always start the emulator with Prolog code */
      return(FALSE);
    }
  CodeAdr = ppe->CodeOfPred;
  return(do_goal(CodeAdr, arity, pt, 0, TRUE));
}

int
YapRestartTopGoal(void)
{
  P = (yamop *)FAILCODE;
  return(exec_absmi(TRUE));
}

int
YapReset(void)
{
  /* first, backtrack to the root */
  while (B->cp_b != NULL)
    B = B->cp_b;
  P = (yamop *)FAILCODE;
  if (exec_absmi(0) != 0)
    return(FALSE);
  /* reinitialise the engine */
  InitYaamRegs();
  return(TRUE);
}

void 
InitExecFs(void)
{
  InitCPred("$execute", 1, p_execute, 0);
  InitCPred("$execute_within", 3, p_execute_within, 0);
  InitCPred("$execute", 2, p_at_execute, 0);
  InitCPred("$call_with_args", 1, p_execute_0, 0);
  InitCPred("$call_with_args", 2, p_execute_1, 0);
  InitCPred("$call_with_args", 3, p_execute_2, 0);
  InitCPred("$call_with_args", 4, p_execute_3, 0);
  InitCPred("$call_with_args", 5, p_execute_4, 0);
  InitCPred("$call_with_args", 6, p_execute_5, 0);
  InitCPred("$call_with_args", 7, p_execute_6, 0);
  InitCPred("$call_with_args", 8, p_execute_7, 0);
  InitCPred("$call_with_args", 9, p_execute_8, 0);
  InitCPred("$call_with_args", 10, p_execute_9, 0);
  InitCPred("$call_with_args", 11, p_execute_10, 0);
#ifdef DEPTH_LIMIT
  InitCPred("$execute_under_depth_limit", 2, p_execute_depth_limit, 0);
#endif
  InitCPred("$execute0", 1, p_execute0, 0);
  InitCPred("$save_current_choice_point", 1, p_save_cp, 0);
  InitCPred("$pred_goal_expansion_on", 0, p_pred_goal_expansion_on, SafePredFlag);
}

