/*************************************************************************
*									 *
*	 YAP Prolog 							 *
*									 *
*	Yap Prolog was developed at NCCUP - Universidade do Porto	 *
*									 *
* Copyright L.Damas, V.S.Costa and Universidade do Porto 1985-1997	 *
*									 *
**************************************************************************
*									 *
* File:		heapgc.c						 *
* Last rev:								 *
* mods:									 *
* comments:	Global Stack garbage collector                           *
*									 *
*************************************************************************/
#ifdef SCCS
static char     SccsId[] = "%W% %G%";
#endif /* SCCS */

#include "absmi.h"
#include "yapio.h"


/* #define DB_SEARCH_METHOD 1 */

/* global variables for garbage collection */

#ifndef DEBUG
static
#endif
 int      gc_calls = 0;	/* number of times GC has been called */

static Int      tot_gc_time = 0; /* total time spent in GC */

static choiceptr current_B;

/* in a single gc */
Int      total_marked;	/* number of heap objects marked */

STATIC_PROTO(Int  p_inform_gc, (void));
STATIC_PROTO(Int  p_gc, (void));

#ifdef DEBUG

int
vsc_stop(void) {
}

#endif

#ifdef DEBUG
STATIC_PROTO(void trace_gc, (void));
STATIC_PROTO(int  count_cells_marked, (void));
#endif /* DEBUG */
STATIC_PROTO(void push_registers, (Int));
STATIC_PROTO(void marking_phase, (tr_fr_ptr, CELL *, yamop *));
STATIC_PROTO(void compaction_phase, (tr_fr_ptr, CELL *, yamop *));
STATIC_PROTO(void pop_registers, (Int));
#if DB_SEARCH_METHOD
#else
STATIC_PROTO(void store_ref_in_dbtable, (DBRef));
STATIC_PROTO(DBRef find_ref_in_dbtable, (DBRef));
#endif
STATIC_PROTO(void init_dbtable, (tr_fr_ptr));
STATIC_PROTO(void mark_db_fixed, (CELL *));
STATIC_PROTO(void mark_regs, (tr_fr_ptr));
STATIC_PROTO(void mark_external_reference, (CELL *));
STATIC_PROTO(void mark_trail, (tr_fr_ptr, tr_fr_ptr, CELL *));
STATIC_PROTO(void mark_environments, (CELL *, OPREG, CELL *));
STATIC_PROTO(void mark_choicepoints, (choiceptr, tr_fr_ptr));
STATIC_PROTO(void into_relocation_chain, (CELL *, CELL *));
STATIC_PROTO(void sweep_trail, (choiceptr, tr_fr_ptr));
STATIC_PROTO(void sweep_environments, (CELL *, OPREG, CELL *));
STATIC_PROTO(void sweep_choicepoints, (choiceptr));
STATIC_PROTO(choiceptr update_B_H, (choiceptr, CELL *, CELL *, CELL *));
STATIC_PROTO(void compact_heap, (void));
STATIC_PROTO(void update_relocation_chain, (CELL *, CELL *));

static Int      tot_gc_recovered = 0; /* number of heap objects in all garbage collections */

#include "heapgc.h"

static int discard_trail_entries = 0;

#ifdef DEBUG

static void
trace_gc(void)
{
  CELL *dest;

  for (dest = H0; dest < H; ++dest) {
    if (MARKED(*dest)) {
#if SHORT_INTS
      YP_fprintf(YP_stderr, "[GC] bad cell %lx at %p\n", *dest, dest);
#else
      YP_fprintf(YP_stderr, "[GC] bad cell %x at %p\n",	 *dest, dest);
#endif
    }
  }
}
#endif



/* find all accessible objects on the heap and squeeze out all the rest */



/* push the active registers onto the trail for inclusion during gc */

static void 
push_registers(Int num_regs)
{
	int             i;

#ifdef COROUTINING
#ifdef MULTI_ASSIGNMENT_VARIABLES
	TrailTerm(TR++) = WokenGoals;
	TrailTerm(TR++) = MutableList;
	TrailTerm(TR++) = AttsMutableList;
#endif
#endif
	for (i = 1; i <= num_regs; i++)
		TrailTerm(TR++) = (CELL) XREGS[i];
}



/* pop the corrected register values from the trail and update the registers */

static void 
pop_registers(Int num_regs)
{
	int             i;
	CELL_PTR ptr = (CELL *)TR;

#ifdef COROUTINING
#ifdef MULTI_ASSIGNMENT_VARIABLES
	WokenGoals = *ptr++;
	MutableList = *ptr++;
	AttsMutableList = *ptr++;
#endif
#endif
	for (i = 1; i <= num_regs; i++)
		XREGS[i] = *ptr++;
}

#ifdef DEBUG
static int 
count_cells_marked(void)
{
  CELL *current;
  int found_marked = 0;

  for (current = H - 1; current >= H0; current--) {
    if (MARKED(*current)) {
      found_marked++;
    }
  }
  return(found_marked);
}
#endif


#if DB_SEARCH_METHOD

/* In this method we store first pointers to the DB, and when we find an entry
   in the trail we go and see if anyone is pointing at it. We try to be efficient
   space-wise, by removing entries to code that does not need them
*/

typedef struct DB_entry {
  CELL *addr;
  struct DB_entry *next;
} *dbentry;

#define MAX_DB_ENTRIES 1024

static dbentry db_vec, free_entries;

static dbentry dbtable[MAX_DB_ENTRIES];

static int dbscale;

static CODEADDR MyHpBase;

static int
dbslot(CELL *ptr)
{
  return((Addr(ptr)-Addr(MyHpBase))/dbscale);
}

static dbentry
new_dbvec_entry(void)
{
  dbentry val;
  if (free_entries != NULL) {
    val = free_entries;
    free_entries = free_entries->next;
    return(val);
  }
  val = db_vec;
  db_vec++;
  return(val);
}
  
static void
add_new_dbuse(CELL *ptr)
{
  int off = dbslot(ptr);
  dbentry prev = NULL;
  dbentry curr = dbtable[off];

  while (curr != NULL && curr->addr < ptr) {
    prev = curr;
    curr = curr->next;
  }
  if (curr == NULL) {
    dbentry db_entry = new_dbvec_entry();
    db_entry->addr = ptr;
    db_entry->next = NULL;
    dbtable[off] = db_entry;
    return;
  } else if (curr->addr > ptr) {
    dbentry db_entry = new_dbvec_entry();
    db_entry->addr = ptr;
    db_entry->next = curr;
    if (prev == NULL) {
      dbtable[off] = db_entry;
    } else {
      prev->next = db_entry;
    }
  }
}

static void 
mark_db_fixed(CELL *trail_ptr) {
  add_new_dbuse(trail_ptr);
}


/* there are two cases: either off0 == offf or offf > off0 */
static int
found_dbentries(CELL *beg, CELL *end)
{
  int off0 = dbslot(beg);
  int offf = dbslot(end);
  int off = off0;
  int found = FALSE;
  dbentry prev = NULL;
  dbentry curr = dbtable[off];

  while (curr != NULL && curr->addr < beg) {
    prev = curr;
    curr = curr->next;
  }
  do {
    while (curr != NULL) {
      if (curr->addr > end) {
	return(found);
      } else {
	dbentry nexte = curr->next;
	found = TRUE;
	if (prev == NULL)
	  dbtable[off] = nexte;
	else
	  prev->next = nexte;
	curr->next = free_entries;
	free_entries = curr;
	curr = nexte;
      }
    }
    if (off < offf) {
      off++;
      prev = NULL;
      curr = dbtable[off];
    } else
      return(found);
  } while (TRUE);
}



#else

/* straightforward binary tree scheme that, given a key, finds a
   matching dbref */  

typedef struct db_entry {
  DBRef val;
  struct db_entry *left;
  CELL *lim;
  struct db_entry *right;
} *dbentry;

static dbentry db_vec, db_vec0;


/* init the table */
static void
store_ref_in_dbtable(DBRef entry)
{
  dbentry parent = db_vec0;
  dbentry new = db_vec;

  if ((ADDR)new > TrailTop-1024)
    growtrail(64 * 1024L);
  new->val = entry;
  new->lim = entry->Contents+entry->NOfCells;
  new->left = new->right = NULL;
  if (db_vec == db_vec0) {
    db_vec++;
    return;
  }
  db_vec++;
  parent = db_vec0;
 beg:
  if (entry < parent->val) {
    if (parent->right == NULL) {
      parent->right = new;
    } else {
      parent = parent->right;
      goto beg;
    }
  } else {
    if (parent->left == NULL) {
      parent->left = new;
    } else {
      parent = parent->left;
      goto beg;
    }
  }
}

/* find an element in the dbentries table */
static DBRef
find_ref_in_dbtable(DBRef entry)
{
  dbentry current = db_vec0;

  while (current != NULL) {
    if (current->val < entry && current->lim > (CELL *)entry)
      return(current->val);
    if (entry < current->val)
      current = current->right;
    else
      current = current->left;
  }
  return(NULL);
}

static void 
mark_db_fixed(CELL *ptr) {
  DBRef el;

  el = find_ref_in_dbtable((DBRef)ptr);
  if (el != NULL)
    el->Flags |= GcFoundMask;
}

#endif

static void 
init_dbtable(tr_fr_ptr trail_ptr) {
#if DB_SEARCH_METHOD
  MyHpBase = HeapBase;
  db_vec = (dbentry)TR;
  free_entries = NULL;
  dbscale = ((Addr(HeapTop)-Addr(MyHpBase))+MAX_DB_ENTRIES-1)/MAX_DB_ENTRIES;
#else
  db_vec0 = db_vec = (dbentry)TR;
  while (trail_ptr > (tr_fr_ptr)TrailBase) {
    register CELL trail_cell;
    
    trail_ptr--;
    
    trail_cell = TrailTerm(trail_ptr);

    if (!IsVarTerm(trail_cell) && IsPairTerm(trail_cell)) {
      CELL *pt0 = RepPair(trail_cell);
      if (pt0 < H0) {
	/* DB pointer */ 
	CELL flags = Flags((CELL)pt0);
	/* for the moment, if all references to the term in the stacks
	   are only pointers, reset the flag */
	if (FlagOn(DBClMask, flags) && !FlagOn(LogUpdMask, flags)) {
	  if (FlagOn(DBNoVars, flags)) {
	    CODEADDR entry = ((CODEADDR)pt0 - (CELL) &(((DBRef) NIL)->Flags));
	    store_ref_in_dbtable((DBRef)entry);
	  }
	}
      }
    }
  }
#endif
}



/* mark a heap object and all heap objects accessible from it */

void 
mark_variable(CELL_PTR current)
{
  CELL_PTR        next;
  register CELL	ccur;
  unsigned int    arity;
  unsigned int		i;

 begin:
  ccur = *current;
  if (MARKED(ccur))
    return;
  MARK(current);
  total_marked++;
  next = GET_NEXT(ccur);

  if (IsVarTerm(ccur)) {
    if (ONHEAP(next)) {
      /* do variable shunting between variables in the global */
      if (current >= HB && current < (CELL *)current_B && next >= HB) {
	CELL cnext = *next;
	if (MARKED(cnext)) {
	  cnext = UNMARK_CELL(cnext);
	}
	if (IsVarTerm(cnext)) {
	  /* a chain of variables over HB, we can shunt */ 
	  if ((CELL *)cnext >= HB) {
	    *current = MARK_CELL(cnext);
	    current = (CELL *)cnext;
	  } else {
	    current = next;
	  }
	} else if (IsAtomicTerm(cnext)) {
	  /* I don't know if this binding is conditional yet
	     (I could possibly consult the trail??) */
	  current = next;
	} else if (IsPairTerm(cnext)) {
	  if (RepPair(cnext) >= HB) {
	    *current = MARK_CELL(cnext);
	    /* should I go on and mark the term? */
	    if (MARKED(*next)) {
	      return;
	    }
	    /* force marking */
	    *current = cnext;
	    total_marked--;
	  }  else {
	    current = next;
	  } 
	} else if (IsApplTerm(cnext)) {
	  if (RepAppl(cnext) >= HB) {
	    *current = MARK_CELL(cnext);
	    /* should I go on and mark the term? */
	    if (MARKED(*next)) {
	      return;
	    }
	    /* force marking */
	    *current = cnext;
	    total_marked--;
	  } else {
	    current = next;
	  }
	}
      } else
	/* what I'd do without variable shunting */
	{
	  current = next;
	}
      goto begin;
    }
#ifdef DEBUG
    else if (next < (CELL *)HeapBase)
      YP_fprintf(YP_stderr, "ooops while marking %x, %x at %p\n", ccur, current, next);
#endif
    return;
  } else if (IsPairTerm(ccur)) {
    if (ONHEAP(next)) {
      mark_variable(next);
      current = next + 1;
      goto begin;
    } else if (ONCODE(next)) {
      mark_db_fixed(RepPair(ccur));
    }
    return;
  } else if (IsApplTerm(ccur)) {
    register CELL cnext = *next;
    
    if (ONCODE(next)) {
      if ((Functor)cnext == FunctorDBRef) {
	DBRef tref = DBRefOfTerm(ccur);
	/* make sure the reference is marked as in use */
	tref->Flags |= GcFoundMask;
      } else {
	mark_db_fixed(next);
      }
      return;
    }
    if ( MARKED(cnext) || !ONHEAP(next) )
      return;
    
    if (next < H0) return;
#if TERM_EXTENSIONS
    if (IsAttachFunc((Functor)cnext)) {
      attas[ExtFromFunctor((Functor)cnext)].mark_op(next);
      return;
    }
#endif
    arity = ArityOfFunctor((Functor)(cnext));
    MARK(next);
    ++total_marked;
    for (i = 1; i < arity; ++i)
      mark_variable(next + i);
    current = next + arity;
    goto begin;
  }
}

static void
mark_external_reference(CELL *ptr) {
  CELL reg = *ptr;

  if (IsVarTerm(reg)) {
    if (ONHEAP(reg)) {
      mark_variable(ptr);
      total_marked--;
    }
  } else if (IsApplTerm(reg)) {
   CELL *next = RepAppl(reg);

   if (ONHEAP(next)) {
      mark_variable(ptr);
      total_marked--;
   } else if (ONCODE(next)) {
     if ((Functor)(*next) == FunctorDBRef) {
       DBRef tref = DBRefOfTerm(reg);
       /* make sure the reference is marked as in use */
       tref->Flags |= GcFoundMask;
     } else {
       mark_db_fixed(next);
     }
   }
  } else if (IsPairTerm(reg)) {
   CELL *next = RepPair(reg);
   
   if (ONHEAP(next)) {
      mark_variable(ptr);
      total_marked--;
   }
  }
}

/*
 * mark all heap objects accessible from the trail (which includes the active
 * general purpose registers) 
 */

static void 
mark_regs(tr_fr_ptr old_TR)
{
  tr_fr_ptr        trail_ptr;

	
  /* first, whatever we dumped on the trail. Easier just to do
     the registers separately?  */
  for (trail_ptr = old_TR; trail_ptr < TR; trail_ptr++)
    mark_external_reference(&TrailTerm(trail_ptr));
}


/* mark all heap objects accessible from a chain of environments */

static void 
mark_environments(CELL_PTR gc_ENV, OPREG size, CELL *pvbmap)
{
  CELL_PTR        saved_var;

  while (gc_ENV != NULL) {	/* no more environments */
    Int bmap = 0;
    int currv = 0;

#ifdef DEBUG
    if (size <  0 || size > 512)
      YP_fprintf(YP_stderr,"Oops, env size %d\n", size);
#endif
    /* for each saved variable */
    if (size > EnvSizeInCells) {
      int tsize = size - EnvSizeInCells;

      currv = sizeof(CELL)*8-tsize%(sizeof(CELL)*8);
      pvbmap += tsize/(sizeof(CELL)*8);
      bmap = *pvbmap;
      bmap = (Int)(((CELL)bmap) << currv);
    }
	  
    for (saved_var = gc_ENV - size; saved_var < gc_ENV - EnvSizeInCells; saved_var++) {
      if (currv == sizeof(CELL)*8) {
	pvbmap--;
	bmap = *pvbmap;
	currv = 0;
      }
      /* we may have already been here */
      if (bmap < 0 && !MARKED(*saved_var)) {
	  mark_external_reference(saved_var);
      }
      bmap <<= 1;
      currv++;
    }
    /* have we met this environment before?? */
    /* we use the B field in the environment to tell whether we have
       been here before or not.

       We do it at the end because we don't want to lose any variables
       that would have been trimmed at the first environment visit.
    */
    if (MARKED(gc_ENV[E_CB]))
      return;
    MARK(gc_ENV+E_CB);

    size = EnvSize((CELL_PTR) (gc_ENV[E_CP]));	/* size = EnvSize(CP) */
    pvbmap = EnvBMap((CELL_PTR) (gc_ENV[E_CP]));
    gc_ENV = (CELL_PTR) gc_ENV[E_E];	/* link to prev
					 * environment */
  }
}

/* 
   Cleaning the trail should be quick and simple, right? Well, not
   really :-(. The problem is that the trail includes a dumping ground
   of the WAM registers and of extra choice-point fields, which need
   to be cleaned from somewhere.

   And cleaning the trail itself is not easy. The problem is that we
   may not have cleaned the trail after cuts. If we naively followed
   these pointers, we could have direct references to the global
   stack!  A solution is to verify whether we are poiting at a
   legitimate trail entry. Unfortunately this requires some extra work
   following choice-points.

*/

static void
mark_trail(tr_fr_ptr trail_ptr, tr_fr_ptr trail_base, CELL *gc_H)
{
  while (trail_ptr > trail_base) {
    register CELL trail_cell;
    
    trail_ptr--;
    
    trail_cell = TrailTerm(trail_ptr);

    if (IsVarTerm(trail_cell)) {
      CELL *hp = (CELL *)trail_cell;
      /* if a variable older than the current CP has not been marked yet,
	 than its new binding is not accessible and we can reset it. Note
	 we must use gc_H to avoid trouble with dangling variables
	 in the heap */
      if (hp < gc_H && hp >= H0 && !MARKED(*hp)) {
	/* reset to be a variable */
	RESET_VARIABLE(hp);
	discard_trail_entries++;
	RESET_VARIABLE(trail_ptr);
      } else if (hp < (CELL *)HeapTop) {
	/* I decided to allow pointers from the Heap back into the trail.
	   The point of doing so is to have dynamic arrays */
	mark_external_reference(hp);
      }
    }
    else if (IsPairTerm(trail_cell)) {
      CELL *pt0 = RepPair(trail_cell);
      if (pt0 < H0) {
#if DB_SEARCH_METHOD
	/* DB pointer */ 
	CELL flags = Flags((CELL)pt0);
	/* for the moment, if all references to the term in the stacks
	   are only pointers, reset the flag */
	if (FlagOn(DBClMask, flags) && !FlagOn(LogUpdMask, flags)) {
	  if (FlagOn(DBNoVars, flags)) {
	    DBRef entry = (DBRef) ((CODEADDR)pt0 - (CELL) &(((DBRef) NIL)->Flags));
	    if (found_dbentries(entry->Contents,
				 entry->Contents+entry->NOfCells))
	      entry->Flags |= GcFoundMask;
	  }
	}
#endif
      }
#if  MULTI_ASSIGNMENT_VARIABLES
      else {
	if (HEAP_PTR(trail_cell)) {
	  /* fool the gc into thinking this is a variable */
	  TrailTerm(trail_ptr) = (CELL)RepPair(trail_cell);
	  mark_external_reference(&(TrailTerm(trail_ptr)));
	  /* reset the gc to believe the original tag */
	  TrailTerm(trail_ptr) = AbsPair((CELL *)TrailTerm(trail_ptr));
	}
	trail_ptr --;
	trail_cell = TrailTerm(trail_ptr);
	if (HEAP_PTR(trail_cell)) {
	  mark_external_reference(&TrailTerm(trail_ptr));
	}
      }
#endif
    }
  }
}

/*
 * mark all heap objects accessible from each choicepoint & its chain of
 * environments 
 */

static void 
mark_choicepoints(register choiceptr gc_B, tr_fr_ptr saved_TR)
{

  while (gc_B != NULL) {
    op_numbers opnum;
    register OPCODE op;
    yamop *rtp = gc_B->cp_ap;
    current_B = gc_B;
    HB = gc_B->cp_h;

    op = rtp->opc;
    opnum = op_from_opcode(op);
    mark_trail(saved_TR, gc_B->cp_tr, gc_B->cp_h);
    saved_TR = gc_B->cp_tr;
  restart_cp:
    if (opnum == _or_else || opnum == _or_last) {
    /* ; choice point */
    

      mark_environments((CELL_PTR) (gc_B->cp_a1),
#ifdef YAPOR
			-gc_B->cp_cp->u.ldl.s / ((OPREG)sizeof(CELL)),
			EnvBMapOffset((CELL *)(gc_B->cp_cp->u.ldl.bl))
#else
			-gc_B->cp_cp->u.sla.s / ((OPREG)sizeof(CELL)),
			EnvBMapOffset((CELL *)(gc_B->cp_cp->u.sla.l2))
#endif
			);
    } else {
      /* choicepoint with arguments */
      register CELL_PTR        saved_reg;
      OPREG nargs;
	  
      if (opnum == _Nstop)
	mark_environments((CELL_PTR) gc_B->cp_env,
			  EnvSizeInCells,
			  NULL);
      else
	mark_environments((CELL_PTR) gc_B->cp_env,
			  EnvSize((CELL_PTR) (gc_B->cp_cp)),
			  EnvBMap((CELL_PTR) (gc_B->cp_cp)));
      /* extended choice point */
      switch (opnum) {
      case _Nstop:
	if (gc_B->cp_b != NULL) {

	  nargs = IntOfTerm(gc_B->cp_a1);
	  break;
	} else {
	  /* this is the last choice point, the work is done  ;-) */
	  return;
	}
      case _switch_last:
      case _switch_l_list:
	nargs = rtp->u.slll.s;
	break;
      case _retry_c:
      case _retry_userc:
	if (gc_B->cp_ap == RETRY_C_RECORDED_CODE 
	    || gc_B->cp_ap == RETRY_C_RECORDED_K_CODE 
	    || gc_B->cp_ap == RETRY_C_DRECORDED_CODE 
	    || gc_B->cp_ap == RETRY_C_RECORDEDP_CODE) {
	  /* we have a reference from the choice-point stack to a term */
	  choiceptr old_b = B;
	  DBRef ref;
	  B = gc_B;
	  ref = (DBRef)EXTRA_CBACK_ARG(3,1);
	  if (IsVarTerm((CELL)ref))
	    ref->Flags |= GcFoundMask;
	  else {
	    if (ONCODE((CELL)ref)) {
	      mark_db_fixed(RepAppl((CELL)ref));
	    }
	  }
	  B = old_b;
	}
	nargs = rtp->u.lds.s+rtp->u.lds.extra;
	break;
      case _trust_logical_pred:
      case _retry_profiled:
	rtp = NEXTOP(rtp,l); 
	op = rtp->opc;
	opnum = op_from_opcode(op);
	goto restart_cp;
#ifdef DEBUG
      case _retry_me:
      case _trust_me:
      case _profiled_retry_me:
      case _profiled_trust_me:
      case _retry_me0:
      case _trust_me0:
      case _retry_me1:
      case _trust_me1:
      case _retry_me2:
      case _trust_me2:
      case _retry_me3:
      case _trust_me3:
      case _retry_me4:
      case _trust_me4:
      case _retry_and_mark:
      case _profiled_retry_and_mark:
      case _retry:
      case _trust_in:
      case _trust:
      case _retry_first:
      case _trust_first_in:
      case _trust_first:
      case _retry_tail:
      case _trust_tail_in:
      case _trust_tail:
      case _retry_head:
      case _trust_head_in:
      case _trust_head:
	nargs = rtp->u.ld.s;
	break;
      default:
	YP_fprintf(YP_stderr, "OOps in GC: Unexpected opcode: %d\n", opnum);
	nargs = 0;
#else
      default:
	nargs = rtp->u.ld.s;
#endif
      }

      /* for each saved register */
      for (saved_reg = &gc_B->cp_a1;
	   /* assumes we can count registers in CP this
	      way */
	   saved_reg < &gc_B->cp_a1 + nargs;
	   saved_reg++) {
	mark_external_reference(saved_reg);
      }
    }	    
    gc_B = gc_B->cp_b;
  }
}




/*
 * insert a cell which points to a heap object into relocation chain of that
 * object 
 */

static void 
into_relocation_chain(CELL_PTR current, CELL_PTR next)
{
#ifdef TAGS_FAST_OPS
	register CELL ccur = *current, cnext = *next;

	if (IsVarTerm(ccur)) {
	  *current = ( MARKED(ccur) ? MARK_CELL(UNMARKED(cnext)) :
		       UNMARKED(cnext) );
	  *next = (MARKED(cnext) ? MBIT : 0) | RBIT | (Int) current;
	} else if (IsPairTerm(ccur)) {
	  *current = ( MARKED(ccur) ? MARK_CELL(UNMARKED(cnext)) :
		       UNMARKED(cnext) );
	  *next = AbsPair((CELL *)
			  ((MARKED(cnext) ? MBIT : 0) | RBIT | (Int) current));
	} else if (IsApplTerm(ccur)) {
	  *current = ( MARKED(ccur) ? MARK_CELL(UNMARKED(cnext)) :
		       UNMARKED(cnext) );
	  *next = AbsAppl((CELL *)
			  ((MARKED(cnext) ? MBIT : 0) | RBIT | (Int) current));
	} else {
	  YP_fprintf(YP_stderr," OH MY GOD !!!!!!!!!!!!\n");
	}
#else
	CELL             current_tag;

	current_tag = TAG(*current);
	*current = (*current & MBIT) | (*next & ~MBIT);
#if INVERT_RBIT
	*next = ((*next & MBIT) | (CELL) current | current_tag) & ~RBIT;
#else
	*next = (*next & MBIT) | RBIT | (CELL) current | current_tag;
#endif
#endif
}


/* insert trail cells which point to heap objects into relocation chains */

static void 
sweep_trail(choiceptr gc_B, tr_fr_ptr old_TR)
{
	tr_fr_ptr     trail_ptr;
	CELL *cp_H = gc_B->cp_h;
	Int OldHeapUsed = HeapUsed;
#ifdef DEBUG
	Int hp_entrs = 0, hp_erased = 0, hp_not_in_use = 0,
	  hp_in_use_erased = 0, code_entries = 0;
#endif


#if DB_SEARCH_METHOD
#if DEBUG
	{
	  int i;
	  for (i=0; i<MAX_DB_ENTRIES; i++)
	    if (dbtable[i] != NULL)
	      fprintf(YP_stderr, "oops at entry %d: %p\n", i, dbtable[i]->addr);
	}
#endif
#endif
	/* first, whatever we dumped on the trail. Easier just to do
	   the registers separately?  */
	for (trail_ptr = old_TR; trail_ptr < TR; trail_ptr++)
	  if (HEAP_PTR(TrailTerm(trail_ptr))) {
	    UNMARK(&TrailTerm(trail_ptr));
	    into_relocation_chain(&TrailTerm(trail_ptr), GET_NEXT(TrailTerm(trail_ptr)));
	  }


	/* next, follows the real trail entries */
	trail_ptr = old_TR;
	while (trail_ptr > (tr_fr_ptr)TrailBase) {
	  register CELL trail_cell;

	  trail_ptr--;

	  trail_cell = TrailTerm(trail_ptr);

	  if (gc_B && trail_ptr < gc_B->cp_tr) {
	    do {
	      gc_B = gc_B->cp_b;
	    } while (gc_B && trail_ptr < gc_B->cp_tr);
	    cp_H = gc_B->cp_h;
	  }

	  if (IsVarTerm(trail_cell)) {
	      /* we need to check whether this is a honest to god trail entry */
	    if ((CELL *)trail_cell < cp_H && MARKED(*(CELL *)trail_cell) && (CELL *)trail_cell > H0) {
	      into_relocation_chain(&TrailTerm(trail_ptr), GET_NEXT(trail_cell));
	    } else if ((CELL *)trail_cell < (CELL *)HeapTop) {
	      /* we may have pointers from the heap back into the cell */
	      UNMARK(CellPtr(trail_cell));
	      into_relocation_chain(CellPtr(trail_cell), GET_NEXT(*(CELL *)trail_cell));
	    } else {
	      /* clean the trail, avoid dangling pointers! */
	      if ((CELL *)trail_cell < (CELL *)gc_B) {
		RESET_VARIABLE(&TrailTerm(trail_ptr));
		discard_trail_entries++;
	      }
	    }
	  } else if (IsPairTerm(trail_cell)) {
#if  MULTI_ASSIGNMENT_VARIABLES
	    if (MARKED(trail_cell)) {
	      UNMARK(&TrailTerm(trail_ptr));
	      into_relocation_chain(&TrailTerm(trail_ptr), GET_NEXT(trail_cell));
	      trail_cell = TrailTerm(trail_ptr-1);
	      trail_ptr--;
	      if (MARKED(trail_cell)) {
		trail_cell = TrailTerm(trail_ptr);
		UNMARK(&TrailTerm(trail_ptr));
		into_relocation_chain(&TrailTerm(trail_ptr), GET_NEXT(trail_cell));
	      }
	    } else
#endif
	      if (RepPair(trail_cell) < H0) {
		CELL *pt0 = RepPair(trail_cell);
		CELL flags = Flags((CELL)pt0);
#ifdef DEBUG
		if (FlagOn(DBClMask, flags) && !FlagOn(LogUpdMask, flags)) {
		  hp_entrs++;
		  if (!FlagOn(GcFoundMask, flags)) {
		    hp_not_in_use++;
		    if (FlagOn(ErasedMask, flags)) {
		      hp_erased++;
		    }
		  } else {
		    if (FlagOn(ErasedMask, flags)) {
		      hp_in_use_erased++;
		    }		
		  }
		} else {
		  code_entries++;
		}
#endif
		if (!FlagOn(GcFoundMask, flags)) {
		  if (FlagOn(DBClMask, flags)  && !FlagOn(LogUpdMask, flags)) {
		    Flags((CELL)pt0) = ResetFlag(InUseMask, flags);
		    if (FlagOn(ErasedMask, flags)) {
		      ErDBE((DBRef) ((CELL)pt0 - (CELL) &(((DBRef) NIL)->Flags)));
		    }
		    RESET_VARIABLE(trail_ptr);
		    discard_trail_entries++;
		  }
		} else {
		  Flags((CELL)pt0) = ResetFlag(GcFoundMask, flags);
		}
	      } else {
		/* skip one position */
		trail_ptr--;
	      }
	    }
	  }
    if (is_gc_verbose()) {
      YP_fprintf(YP_stderr,
#if SHORT_INTS
		 "[GC]       Trail: discarded %d (%ld%%) cells out of %ld\n",
#else
		 "[GC]       Trail: discarded %d (%d%%) cells out of %d\n",
#endif
		 discard_trail_entries,
		 discard_trail_entries*100/(old_TR-(tr_fr_ptr)TrailBase),
		 old_TR-(tr_fr_ptr)TrailBase);
#ifdef DEBUG
      if (hp_entrs > 0)
	YP_fprintf(YP_stderr,
		   "[GC]       Trail: unmarked %d dbentries (%d%%) out of %d\n",
		   hp_not_in_use,
		   hp_not_in_use*100/hp_entrs,
		   hp_entrs);
      if (hp_in_use_erased > 0 && hp_erased > 0)
	YP_fprintf(YP_stderr,
		   "[GC]       Trail: deleted %d dbentries (%d%%) out of %d\n",
		   hp_erased,
		   hp_erased/((hp_erased+hp_in_use_erased)/100),
		   (hp_erased+hp_in_use_erased));
#endif
      YP_fprintf(YP_stderr,
#if SHORT_INTS
		 "[GC]       Heap: recovered %ld bytes (%ld%%) out of %ld\n",
#else
		 "[GC]       Heap: recovered %d bytes (%d%%) out of %d\n",
#endif
		 OldHeapUsed-HeapUsed,
		 (OldHeapUsed-HeapUsed)/(OldHeapUsed/100),
		 OldHeapUsed);
    }
}


/*
 * insert cells of a chain of environments which point to heap objects into
 * relocation chains 
 */

static void 
sweep_environments(CELL_PTR gc_ENV, OPREG size, CELL *pvbmap)
{
  CELL_PTR        saved_var;

  while (gc_ENV != NULL) {	/* no more environments */
    Int bmap = 0;
    int currv = 0;


    /* for each saved variable */

    if (size > EnvSizeInCells) {
      int tsize = size - EnvSizeInCells;
      
      currv = sizeof(CELL)*8-tsize%(sizeof(CELL)*8);
      pvbmap += tsize/(sizeof(CELL)*8);
      bmap = *pvbmap;
      bmap = (Int)(((CELL)bmap) << currv);
    }
    for (saved_var = gc_ENV - size; saved_var < gc_ENV - EnvSizeInCells; saved_var++) {
      
      if (currv == sizeof(CELL)*8) {
	pvbmap--;
	bmap = *pvbmap;
	currv = 0;
      }
      if (bmap < 0) {
	CELL env_cell = *saved_var;
	if (HEAP_PTR(env_cell) && MARKED(env_cell)) {
	  UNMARK(saved_var);
	  into_relocation_chain(saved_var, GET_NEXT(env_cell));
	}
      }
      bmap <<= 1;
      currv++;
    }
    /* have we met this environment before?? */
    /* we use the B field in the environment to tell whether we have
       been here before or not 
    */
    if (!MARKED(gc_ENV[E_CB]))
      return;
    UNMARK(gc_ENV+E_CB);

    size = EnvSize((CELL_PTR) (gc_ENV[E_CP]));	/* size = EnvSize(CP) */
    pvbmap = EnvBMap((CELL_PTR) (gc_ENV[E_CP]));
    gc_ENV = (CELL_PTR) gc_ENV[E_E];	/* link to prev
					 * environment */
  }
}

/*
 * insert cells of each choicepoint & its chain of environments which point
 * to heap objects into relocation chains 
 */
static void 
sweep_choicepoints(choiceptr gc_B)
{

  while(gc_B != NULL) {
    yamop *rtp = gc_B->cp_ap;
    register OPCODE op = rtp->opc;
    op_numbers opnum = op_from_opcode(op);

  restart_cp:
    /*
     * YP_fprintf(YP_stderr,"sweeping cps: %x, %x, %x\n",
     * *gc_B,CP_Extra(gc_B),CP_Nargs(gc_B)); 
     */
    /* any choice point */
    switch (opnum) {
    case _Nstop:
      /* end of the road, say bye bye! */
      sweep_environments(gc_B->cp_env,
			 EnvSizeInCells,
			 NULL);
      if (gc_B->cp_b != NULL) {
	register CELL_PTR saved_reg;
	
	/* for each saved register */
	for (saved_reg = &gc_B->cp_a1;
	     saved_reg < &gc_B->cp_a1 + IntOfTerm(gc_B->cp_a1);
	     saved_reg++) {
	  if (HEAP_PTR(*saved_reg)) {
	    UNMARK(saved_reg);
	    into_relocation_chain(saved_reg, GET_NEXT(*saved_reg));
	  }
	}
	break;
      } else
	return;
    case _or_else:
    case _or_last:

      sweep_environments((CELL_PTR)(gc_B->cp_a1),
#ifdef YAPOR
			 -gc_B->cp_cp->u.ldl.s / ((OPREG)sizeof(CELL)),
			 EnvBMapOffset((CELL *)(gc_B->cp_cp->u.ldl.bl))
#else
			 -gc_B->cp_cp->u.sla.s / ((OPREG)sizeof(CELL)),
			 EnvBMapOffset((CELL *)(gc_B->cp_cp->u.sla.l2))
#endif
			 );
      break;
    case _trust_logical_pred:
    case _retry_profiled:
      rtp = NEXTOP(rtp,l);
      op = rtp->opc;
      opnum = op_from_opcode(op);
      goto restart_cp;
    case _retry_c:
    case _retry_userc:
      {
	register CELL_PTR saved_reg;
	
	/* for each extra saved register */
	for (saved_reg = &(gc_B->cp_a1)+rtp->u.lds.s;
	     saved_reg < &(gc_B->cp_a1)+rtp->u.lds.s+rtp->u.lds.extra;
	     saved_reg++)
	  if (HEAP_PTR(*saved_reg)) {
	    UNMARK(saved_reg);
	    into_relocation_chain(saved_reg, GET_NEXT(*saved_reg));
	  }
      }
    default:
      {
	register CELL_PTR saved_reg;

	sweep_environments(gc_B->cp_env,
			   EnvSize((CELL_PTR) (gc_B->cp_cp)),
			   EnvBMap((CELL_PTR) (gc_B->cp_cp)));

	/* for each saved register */
	for (saved_reg = &gc_B->cp_a1;
	     saved_reg < &gc_B->cp_a1 + rtp->u.ld.s;
	     saved_reg++)
	    if (HEAP_PTR(*saved_reg)) {
	      UNMARK(saved_reg);
	      into_relocation_chain(saved_reg, GET_NEXT(*saved_reg));
	    }
      }
    }

    /* link to prev choicepoint */
    gc_B = gc_B->cp_b;
  }
}




/* update a relocation chain to point all its cells to new location of object */
static void
update_relocation_chain(CELL_PTR current, CELL_PTR dest)
{
  CELL_PTR        next;
  CELL            ccur = *current;

#ifdef TAGS_FAST_OPS
  while (RMARKED(ccur)) {
    register CELL cnext;

    next = GET_NEXT(ccur);
    cnext = *next;
    
    if (IsVarTerm(ccur)) {
      ccur = *current = (MARKED_VAR(ccur) ?
			 ENSURE_MARKED(cnext) : 
			 UNMARKED(cnext) );
      *next = (MARKED(cnext) ? MBIT : 0) | (Int) dest;
    } else if (IsPairTerm(ccur)) {
      ccur = *current = (MARKED_COMP(ccur) ?
			 ENSURE_MARKED(cnext) :
			 UNMARKED(cnext) );
      *next = AbsPair((CELL *)
		      ((MARKED(cnext) ? MBIT : 0) |
		       (Int) dest));
    } else if (IsApplTerm(ccur)) {
      ccur = *current = (MARKED_COMP(ccur) ?
			 ENSURE_MARKED(cnext) : 
			 UNMARKED(cnext) );
      *next = AbsAppl((CELL *)
		      ((MARKED(cnext) ? MBIT : 0) |
		       (Int) dest));	  
    }
#ifdef DEBUG
    else {
      Abort("[GC] ATOMIC in relocation chain");
    }
#endif
  }
#else /* TAGS_FAST_OPS */
  while (RMARKED(ccur)) {
    CELL             current_tag;
    next = GET_NEXT(ccur);
    current_tag = TAG(ccur);
    ccur = *current = (ccur & MBIT) | (*next & ~MBIT);
#if INVERT_RBIT
    *next = (*next & MBIT) | (CELL) dest | current_tag | RBIT;
#else
    *next = (*next & MBIT) | (CELL) dest | current_tag;
#endif
  }
#endif /* TAGS_FAST_OPS */
}

static inline choiceptr
  update_B_H( choiceptr gc_B, CELL *current, CELL *dest, CELL *odest) {
  /* also make the value of H in a choicepoint
     coherent with the new global
     */
  while (gc_B && current <= gc_B->cp_h) {
    if (gc_B->cp_h == current) {
      gc_B->cp_h = dest;
    } else {
      gc_B->cp_h = odest;
    }
    gc_B = gc_B->cp_b;
  }
  return(gc_B);
}

/*
 * move marked objects on the heap upwards over unmarked objects, and reset
 * all pointers to point to new locations 
 */
static void 
compact_heap(void)
{
  CELL_PTR        dest, current, next;
#ifdef DEBUG
  Int             found_marked = 0;
#endif /* DEBUG */
  choiceptr        gc_B = B;
  int in_garbage = 0;
 

  /*
   * upward phase - scan heap from high to low, setting marked upward
   * ptrs to point to what will be the new locations of the
   * objects pointed to  
   */

  dest = (CELL_PTR) H0 + total_marked - 1;
  for (current = H - 1; current >= H0; current--) {
    if (MARKED(*current)) {
      gc_B = update_B_H(gc_B, current, dest, dest+1);
      if (in_garbage > 0) {
	current[1] = in_garbage;
	in_garbage = 0;
      }
#ifdef DEBUG
      found_marked++;
#endif /* DEBUG */
      update_relocation_chain(current, dest);
      if (HEAP_PTR(*current)) {
	next = GET_NEXT(*current);
	if (next < current)	/* push into reloc.
				 * chain */
	  into_relocation_chain(current, next);
	else if (current == next)	/* cell pointing to
					 * itself */
	  *current = (*current & MBIT) | (CELL) dest;	/* no tag */
      }
      dest--;
    } else {
      in_garbage++;
    }
  }
  if (in_garbage)
    H0[0] = in_garbage;

#ifdef DEBUG
  if (total_marked != found_marked)
#if SHORT_INTS
    YP_fprintf(YP_stderr,"[GC] Upward (%d): %ld total against %ld found\n", gc_calls, total_marked, found_marked);
#else
  YP_fprintf(YP_stderr,"[GC] Upward (%d): %d total against %d found\n", gc_calls, total_marked, found_marked);
#endif
  found_marked = 0;
#endif


  /*
   * downward phase - scan heap from low to high, moving marked objects
   * to their new locations & setting downward pointers to pt to new
   * locations 
   */

  dest = (CELL_PTR) H0;
  for (current = H0; current < H; current++) {
    CELL ccur = *current;
    if (MARKED(ccur)) {
#ifdef DEBUG
      found_marked++;
#endif
      update_relocation_chain(current, dest);
      ccur = *current;
      next = GET_NEXT(ccur);
      if (HEAP_PTR(ccur) &&	/* move current cell &
				 * push */
	  next > current) {	/* into relocation chain  */
	*dest = ccur;
	into_relocation_chain(dest, next);
	UNMARK(dest);
      } else {
	/* just move current cell */
	*dest = ccur = UNMARK_CELL(ccur);
      }
      /* next cell, please */
      dest++;
    } else {
      current += (ccur-1);
    }
  }
#ifdef DEBUG
  if (total_marked != found_marked)
#if SHORT_INTS
    YP_fprintf(YP_stderr,"[GC] Downward (%d): %ld total against %ld found\n", gc_calls, total_marked, found_marked);
#else
  YP_fprintf(YP_stderr,"[GC] Downward (%d): %d total against %d found\n", gc_calls, total_marked, found_marked);
#endif
#endif

  H = dest;		/* reset H */
  HB = B->cp_h;

#ifdef DEBUG
  trace_gc();
#endif /* DEBUG */

}



/*
 * mark all objects on the heap that are accessible from active registers,
 * the trail, environments, and choicepoints 
 */

static void 
marking_phase(tr_fr_ptr old_TR, CELL *current_env, yamop *curp)
{
  current_B = B;
  init_dbtable(old_TR);
  /* active environments */
  mark_environments(current_env, EnvSize(curp), EnvBMap((CELL *)curp));
  mark_choicepoints(B, old_TR);	/* choicepoints, and environs  */
  /* has to be done in the end because of trouble with corouting:
     the corouting registers have direct pointers to coroutining
     data-structures, they must be marked by the time we get there. */
  mark_regs(old_TR);		/* active registers & trail */
  /* reachable from choicepoints */
}


/*
 * move marked heap objects upwards over unmarked objects, and reset all
 * pointers to point to new locations 
 */

static void 
compaction_phase(tr_fr_ptr old_TR, CELL *current_env, yamop *curp)
{
	sweep_environments(current_env, EnvSize(curp), EnvBMap((CELL *)curp));
	sweep_choicepoints(B);
	sweep_trail(B, old_TR);
	compact_heap();
}

int
is_gc_verbose(void)
{
  return(GetValue(AtomGcVerbose) != TermNil);
}

static Int
p_inform_gc(void)
{
  Term tn = MkIntegerTerm(tot_gc_time);
  Term tt = MkIntTerm(gc_calls);
  Term ts = MkIntTerm((total_marked*sizeof(CELL)));
 
  return(unify(tn, ARG2) && unify(tt, ARG1) && unify(ts, ARG3));

}

int 
gc(Int predarity, CELL *current_env, yamop *nextop)
{
  Int           gc_margin = 128;
  Int		heap_cells = H-H0;
  int           gc_on = 0;
  int           gc_trace = 0;
  int		gc_verbose = is_gc_verbose();
  Term          Tgc_margin;
  tr_fr_ptr     old_TR;
  Int		m_time, c_time, time_start, gc_time;
  Int           effectiveness = 0;

#ifdef FIXED_STACKS
  abort_optyap("garbage collection");
#endif /* YAPOR */
#ifdef DEBUG
  trace_gc();
#endif
  if (GetValue(AtomGc) != TermNil)
    gc_on = 1;
  if (IsIntTerm(Tgc_margin = GetValue(AtomGcMargin)))
    gc_margin = IntOfTerm(Tgc_margin);
  else {
    if (gc_calls < 8) 
      gc_margin <<= gc_calls;
    else
      gc_margin <<= 8;
  }
  if (gc_margin < 0 || gc_margin > 4000)
    gc_margin = (LCL0 - H0) >> 9;
  gc_margin = gc_margin << 8;
  if (GetValue(AtomGcTrace) != TermNil)
    gc_trace = 1;
  if (gc_on) {
    gc_calls++;
    if (gc_trace) {
      YP_fprintf(YP_stderr, "[gc]\n");
    } else if (gc_verbose) {
      YP_fprintf(YP_stderr, "[GC] Start of garbage collection %d:\n", gc_calls);
#if SHORT_INTS
      YP_fprintf(YP_stderr, "[GC]       Global: %8ld cells (%p-%p)\n", heap_cells,H0,H);
      YP_fprintf(YP_stderr, "[GC]       Local:%8ld cells (%p-%p)\n", LCL0-ASP,LCL0,ASP);
      YP_fprintf(YP_stderr, "[GC]       Trail:%8ld cells (%p-%p)\n",
		 TR-(tr_fr_ptr)TrailBase,TrailBase,TR);
#else
      YP_fprintf(YP_stderr, "[GC]       Global: %8d cells (%p-%p)\n", heap_cells,H0,H);
      YP_fprintf(YP_stderr, "[GC]       Local:%8d cells (%p-%p)\n", LCL0-ASP,LCL0,ASP);
      YP_fprintf(YP_stderr, "[GC]       Trail:%8d cells (%p-%p)\n",
		 TR-(tr_fr_ptr)TrailBase,TrailBase,TR);
#endif
    }
    time_start = cputime();
    total_marked = 0;
    discard_trail_entries = 0;
    /* get the number of active registers */
    YAPEnterCriticalSection();
    old_TR = TR;
    push_registers(predarity);
    marking_phase(old_TR, current_env, nextop);
    m_time = cputime();
    gc_time = m_time-time_start;
    effectiveness = ((heap_cells-total_marked)*100)/heap_cells;
    if (gc_verbose) {
#if SHORT_INTS
      YP_fprintf(YP_stderr, "[GC]   Mark: Recovered %ld cells of %ld (%ld%%) in %g sec\n", heap_cells-total_marked, heap_cells, effectiveness, (double)(m_time-time_start)/1000);
#else
      YP_fprintf(YP_stderr, "[GC]   Mark: Recovered %d cells of %d (%d%%) in %g sec\n", heap_cells-total_marked, heap_cells, effectiveness, (double)(m_time-time_start)/1000);
#endif
    }
    time_start = m_time;
    compaction_phase(old_TR, current_env, nextop);
    TR = old_TR;
    pop_registers(predarity);
    c_time = cputime();
    YAPLeaveCriticalSection();
    if (gc_verbose) {
      YP_fprintf(YP_stderr, "[GC]   Compress: took %g sec\n", (double)(c_time-time_start)/1000);
    }
    gc_time += (c_time-time_start);
    tot_gc_time += gc_time;
    tot_gc_recovered += (H-H0)-total_marked;
    if (gc_verbose) {
      YP_fprintf(YP_stderr, "[GC] GC %d took %g sec, total of %g sec doing GC so far.\n", gc_calls, (double)gc_time/1000, (double)tot_gc_time/1000);
#if SHORT_INTS
      YP_fprintf(YP_stderr, "[GC]   Left %ld cells free in stacks.\n",
		 ASP-H);
#else
      YP_fprintf(YP_stderr, "[GC]   Left %d cells free in stacks.\n",
		 ASP-H);
#endif
    }
  }
  /* expand the stak if effectiveness is less than 20 % */
  if (ASP - H < gc_margin || !gc_on || effectiveness < 20) {
    if (ASP-H > gc_margin)
      gc_margin = (ASP-H)+MinStackGap*(stack_overflows+1);
    else
      gc_margin = 8 * (gc_margin - (ASP - H));
    gc_margin = ((gc_margin >> 16) + 1) << 16;
    if (gc_margin < MinStackGap)
      gc_margin = MinStackGap;
    while (gc_margin >= MinStackGap && !growstack(gc_margin))
      gc_margin = gc_margin/2;
    return(gc_margin >= MinStackGap);
  }
  /*
   * debug for(save_total=1; save_total<=N; ++save_total)
   * plwrite(XREGS[save_total],DebugPutc,0); 
   */
  return ( TRUE );
}


Int total_gc_time(void)
{
  return(tot_gc_time);
}

static Int
p_gc(void)
{
#ifndef FIXED_STACKS
  gc(0, ENV, CP);
#endif  /* FIXED_STACKS */
  return(TRUE);
}

void 
init_gc(void)
{
  InitCPred("$gc", 0, p_gc, 0);
  InitCPred("$inform_gc", 3, p_inform_gc, 0);
}
