/* -------------------------------- **
**      Scheduler instructions      **
** -------------------------------- */

  PBOp(getwork_first_time,e)
    /* wait for a new parallel goal */
    while (BITMAP_same(GLOBAL_bm_present_workers, GLOBAL_bm_finished_workers));
    make_root_choice_point();  
    PUT_ME_IN_EXECUTING(worker_id);
    /* wait until everyone else is executing! */
    while (! BITMAP_same(GLOBAL_bm_present_workers, GLOBAL_bm_executing_workers));
    SCHEDULER_GET_WORK();
  shared_end:
    PUT_ME_IN_FINISHED(worker_id);
    /* wait until everyone else is finished! */
    while (! BITMAP_same(GLOBAL_bm_present_workers, GLOBAL_bm_finished_workers));
    PUT_ME_OUT_EXECUTING(worker_id);
    if (worker_id == 0) {
      yapor_off();
      free_root_choice_point();
      /* wait until no one is executing */
      while (! BITMAP_empty(GLOBAL_bm_executing_workers));
      goto fail;
    } else {
      PREG = GETWORK_FIRST_TIME;
      PREFETCH_OP(PREG);
      GONext();
    }
  ENDPBOp();



  PBOp(getwork,ld)
    LOCK_OR_FRAME(LOCAL_top_or_fr);
#ifdef TABLING
    if (DepFr_leader_cp(LOCAL_top_dep_fr) == LOCAL_top_cp) {
      /* leader node */
      if (DepFr_leader_dep_is_on_stack(LOCAL_top_dep_fr) && OrFr_alternative(LOCAL_top_or_fr)) {
        /* the leader node depends on the current top node (this means that the current top 
           node is a generator node ???) and the current top node contains unexploited alternatives */
        PREG = OrFr_alternative(LOCAL_top_or_fr);
        PREFETCH_OP(PREG);
        GONext();
      }
      /* the leader node does not depends on the current top node or
         the current top node does not contains unexploited alternatives */
      UNLOCK_OR_FRAME(LOCAL_top_or_fr);
      goto completion;
    }
    /* not a leader node */
#endif /* TABLING */
    if (OrFr_alternative(LOCAL_top_or_fr)) {
      PREG = OrFr_alternative(LOCAL_top_or_fr);
      PREFETCH_OP(PREG);
      GONext();
    }
    UNLOCK_OR_FRAME(LOCAL_top_or_fr);
    SCHEDULER_GET_WORK();
  ENDPBOp();



  /* The idea is to check whether all workers in the node are idle.
     If they are we can go ahead, otherwise we should call the scheduler. */
  PBOp(getwork_seq,ld)
    bitmap idle_on_cp;

    BITMAP_copy(idle_on_cp, GLOBAL_bm_idle_workers);
    BITMAP_insert(idle_on_cp, worker_id);
    LOCK_OR_FRAME(LOCAL_top_or_fr);
    if (OrFr_alternative(LOCAL_top_or_fr) == NULL || 
        ! BITMAP_subset(idle_on_cp, OrFr_members(LOCAL_top_or_fr))) {
      UNLOCK_OR_FRAME(LOCAL_top_or_fr);
      SCHEDULER_GET_WORK();
    }
    PREG = OrFr_alternative(LOCAL_top_or_fr);
    PREFETCH_OP(PREG);
    GONext();
  ENDPBOp();



  PBOp(sync,ld)
    if (! CUT_wait_until_leftmost()) 
      goto shared_fail;
    PREG = NEXTOP(PREG, ld);
    PREFETCH_OP(PREG);
    GONext();
  ENDPBOp();
